/* ****************************************** */
/* Emu71: HP-71B software emulator for PC     */
/*                                            */
/* monit.c   monitor module for Emu71         */
/*                                            */
/* Version 2.4, J-F Garnier 1996-2007         */
/* ****************************************** */


#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <conio.h>
#include <time.h>

#include "trans.h"
#include "saturn.h"

/* variables for breakpoints */
static unsigned char lastopcode=0;
extern long lastaddr;

extern int desas(long addr, char *codes, char *ligne); /* desas.c */





/* ****************************************** */
/* hextoint()                                 */
/*                                            */
/* convert a hex nibble into a int            */
/* ****************************************** */
int hextoint(char c)
{
  if (c>'9') c -= 7;
  c -= 48;
  return(c);
}




/* ****************************************** */
/* inttohex()                                 */
/*                                            */
/* convert a int nibble (0-15) into a hex char*/
/* ****************************************** */
char inttohex(int c)
{
  c &= 15;
  if (c>9) c += 7;
  return(c+48);
}




/* ****************************************** */
/* lit()                                      */
/*                                            */
/* read n nibbles at address adr (20 bits)    */
/* ****************************************** */
void lit(long adr, int n, char *buf)
{
  char far *pt;
  int i, f_low;
  int adrh;
  int segment, offset;

  adrh=(int) (adr>>16);
  adrh &= 0x0f;
  segment=tab_seg[adrh];
  offset=(int) (adr>>1);
  offset &= 0x7fff;
  if (segment==0) {
    for (i=0;i<n;i++) *buf++ = '0';
  }
  else {
    pt=MK_FP(segment,offset);
    f_low= !(adr & 1);
    for (i=1;i<=n;i++) {
      if (f_low) *buf++ = inttohex(*pt);
      else *buf++ = inttohex((*pt++)>>4);
      f_low = !f_low;
    }
  }
  *buf=0;
}





void alpha(char *dest,char *hex,int n)
{
  int i, c;

  for (i=0;i<n;i++) {
    c = hextoint(*hex++);
    c += hextoint(*hex++)<<4;
    if ((c<32)||(c>127)) c = '.';
    *dest++ =c;
  }
}






/* display desassembled opcodes */
long dasm1(long adr)
{
  int j, n;
  char codes[24];
  char ligne[80];

  lit(adr,22,codes);
  n=desas(adr,codes,ligne);
  printf("%05lX ",adr);
  for (j=0;j<n;j++) printf("%c",codes[j]);
  for (;j<18;j++) printf(" ");
  printf(" %s\n",ligne);
  adr += n;
  return(adr);
}






void ch_reg(regsat reg, char *s)
{
  sprintf(s,"%08lx%08lx",reg[1], reg[0]);
}




void affiche(int full)
{
  char s[20];
  int i;

  ch_reg(regA,s); printf(" A=%s ",s);
  ch_reg(regB,s); printf(" B=%s ",s);
  ch_reg(regC,s); printf(" C=%s ",s);
  ch_reg(regD,s); printf(" D=%s",s);
  printf("\n");


  if (full) {
    ch_reg(regR0,s); printf("R0=%s ",s);
    ch_reg(regR1,s); printf("R1=%s ",s);
    ch_reg(regR2,s); printf("R2=%s ",s);
    ch_reg(regR3,s); printf("R3=%s",s);
    printf("\n");
    ch_reg(regR4,s); printf("R4=%s ",s);
    ch_reg(regR5,s); printf("R5=%s ",s);
    ch_reg(regR6,s); printf("R6=%s ",s);
    ch_reg(regR7,s); printf("R7=%s",s);
    printf("\n");
  }

  printf("D0=%05lx D1=%05lx PC=%05lx P=%x",
	 regD0, regD1, regPC, (int) Pointer);
  if (Carry) printf(" CY=1"); else printf(" CY=0");
  if (flagdec&0x80) printf(" DEC"); else printf(" HEX");
  printf(" ST=%04x",status);
  printf(" IN=%04x OUT=%03x",input, output);
  if (hwstatus&1) printf(" XM");
  if (hwstatus&2) printf(" SB");
  if (hwstatus&4) printf(" SR");
  if (hwstatus&8) printf(" MP");
  printf("\n");

  printf("Stack:");
  for (i=0;i<8;i++)
    printf(" %05lx", (retstk[(retstkptr+i)&63] & 0xfffffL ) );
  printf("\n");

  dasm1(regPC);

}


void affiche_stack(void)
{
  int i, j;

  printf("Extended Stack:\n");
  for (j=0;j<64;j+=8) {
    for (i=0;i<8;i++)
      printf(" %05lx", retstk[(retstkptr+i+j)&63]&0xfffffL );
    printf("\n");
  }
}


void set_brkpt(long adr)
{
  unsigned char far *ptr;
  int adrh, segment, offset;

  if (adr>=0) {
    adrh=(int) (adr>>16);
    adrh &= 0x0f;
    segment=tab_seg[adrh];
    offset=(int) (adr>>1);
    offset &= 0x7fff;
    if (segment!=0) {
      lastaddr=adr;
      ptr=MK_FP(segment,offset);
      if ((adr&1)==0) {
	lastopcode=*ptr;
	*ptr=0;
      }
      else {
	lastopcode=(*ptr)>>4;
	*ptr &= 0x0f;
	ptr++;
	lastopcode |= (*ptr)<<4;
	*ptr &= 0xf0;
      }
    }
  }
}

void res_brkpt(void)
{
  unsigned char far *ptr;
  int adrh, segment, offset;

  if (lastaddr>=0) {
    adrh=(int) (lastaddr>>16);
    adrh &= 0x0f;
    segment=tab_seg[adrh];
    offset=(int) (lastaddr>>1);
    offset &= 0x7fff;
    if (segment!=0) {
      ptr=MK_FP(segment,offset);
      if ((lastaddr&1)==0) {
	*ptr=lastopcode;
      }
      else {
	*ptr &= 0x0f;
	*ptr |= lastopcode << 4;
	ptr++;
	*ptr &= 0xf0;
	*ptr |= lastopcode >> 4;
      }
    }
    lastaddr=-1;
  }
}







int monitor(int sst)
{
  int n;

  n=executeSaturn(sst,0);
  if (n<0)
    printf("invalid instruction, stop at %05lx\n",regPC);
  else if (n>0)
    n=execute_special(n);
  clock();
  return(n);
}


/* display buffer */
/*
 FIRSTC (2f47c) Buffer position of 1st char.
 CURSOR (2f47e) Buffer position of cur.
 DSPBFS (2f480) 96 char.
*/
void view_display(void)
{
  char ligne[82];
  unsigned char codes[64];

  lit(0x2f480L,64,codes);
  alpha(ligne,codes,32);
  ligne[32]=0;
  printf("%s",ligne);
  lit(0x2f4c0L,64,codes);
  alpha(ligne,codes,32);
  ligne[32]=0;
  printf("%s\n",ligne);
}





void help(void)
{
  printf("c             : check ROM\n");
  printf("d [ . | addr ]: disasm (.=current PC)\n");
  printf("g addr        : go, run until PC=addr\n");
  printf("j addr        : jump, set PC=addr\n");
  printf("k x           : push keycode x into buffer\n");
  printf("l             : load processor state and RAMs\n");
  printf("m [ . | addr] : memory dump\n");
  printf("q             : quit\n");
  printf("r [ addr ]    : run emulator [break at addr]\n");
  printf("s             : save processor state and RAMs\n");
  printf("t             : trace (single step)\n");
  printf("vd            : view display buffer\n");
  printf("vr            : view CPU registers\n");
  printf("vs            : view extended stack\n");
  printf("vt            : view timer registers\n");
  printf("x             : reset all: ROM, RAM, registers\n");

}




/* returns 1 if deep sleep, and -1 for command 'q' */
int commande(void)
{
  char cmd[82];
  char ligne[82];
  char *tok;
  unsigned char codes[64];
  long debut, fin;
  int  ret;
  long addr;
  int i;

  fin_ecran();
  printf("Emu71 Ready.\n");
  affiche(1);
  fin=0;
  do {
    ret=0;
    printf("Cmd: ");
    gets(cmd);
    tok=strtok(cmd," ");
    tok=strtok(NULL," ");
    if (tok!=NULL) {
      if (tok[0]=='.')
	debut=regPC;
      else if (tok[0]=='=') {
	addr=ch_symb(tok);
	if (addr!= -1)
	  debut=addr;
	else {
	  printf("%s not found\n",tok);
	  cmd[0]=0;
        }
      }
      else
	sscanf(tok,"%lx",&debut);
    }
    else
      debut=fin;

    switch (cmd[0]) {

      case 'c':
	checkrom();
	break;

      case 'd':
	addr=debut;
	for (i=0;i<20;i++) {
          if (ch_label(addr,ligne)) {
	    printf("=%s\n",ligne);
            i++;
          }
	  addr=dasm1(addr);
	}
	fin=addr;
	break;

      case 'g':
	if (tok!=NULL) set_brkpt(debut);
	init_timer();
	do {
	  ret=monitor(255);
	  if (kbhit()) {
	    if (getch()==27) ret=-2;
	  }
	} while (ret==0);
	res_brkpt();
	affiche(0);
	ret=0;
	break;

      case 'j':
	regPC=debut;
	affiche(0);
	break;

      case 'k':
	push_key(debut);
	break;

      case 'l':
	load_state();
	break;

      case 'm':
        debut &= 0xfffffL;
	fin=debut+512;
	for (addr=debut;addr<fin;addr+=32) {
	  lit(addr,32,codes);
	  printf("%05lx %s",addr,codes);
	  lit(addr,34,codes);
	  alpha(ligne,codes,16);
	  ligne[16]=0;
	  printf("  %s",ligne);
	  alpha(ligne,&codes[1],16);
	  ligne[16]=0;
	  printf("  %s\n",ligne);
	}
	fin=addr;
	break;

      case 'q':
	ret=-1;
	break;

      case 'r':
	if (tok!=NULL) set_brkpt(debut);
        prepa_ecran(0);
	ret=simulator();  /* returns 1 if deep sleep */
        fin_ecran();
	res_brkpt();
	if (ret!=1) {
          ret=0;
          affiche(1);
        }
	break;

      case 's':
	save_state();
	break;

      case 't':
	do {
	  init_timer();
	  monitor(1);
	  affiche(0);
	} while (getch()!=27);
	break;

      case 'v':
	switch (cmd[1]) {
	  case 'd':
	    view_display();
	    break;
	  case 'r':
	    affiche(1);
	    break;
	  case 's':
	    affiche_stack();
	    break;
	  case 't':
	    lit(0x2e1f8L,8,codes);
	    printf("%s",codes);
	    lit(0x2e2f8L,8,codes);
	    printf(" %s",codes);
	    lit(0x2e3f8L,8,codes);
	    printf(" %s\n",codes);
	    break;
	  default:
	    printf("?\n");
	}
	break;


      case 'x':
	init_register();
	loadrom();
	init_ram();
	break;

      case '?':
	help();
	break;

      default:
	printf("?\n");
    }
  } while (ret==0);

  return(ret);
}
