; ********************************************************************
;               mulation du processeur Saturn par le 8086
;                       J-F Garnier 1993, 2004
; --------------------------------------------------------------------
; Module saturn.asm : dfinitions des registres, moteur d'excution
; --------------------------------------------------------------------
; Version 0.0 16/03/93
;   prototype, faisabilit
; Version 1.0 04/11/95
; Version 2.0 07/09/01 extensions pour HP42
; Version 2.1 08/09/02 maj
; Version 3.0 03/10/02 12 bits jump table
; Version 3.1 10/07/07 integration in Emu71
;
; ********************************************************************




		.286


		.model large


		; mulation du processeur Saturn par le 8086
		; J-F Garnier
		; conception 16/03/93
		; Version 1 : support du 1LF2 (hp71)
		;         volutions prvues: 1LK7 (hp28), 1LR2 (hp48)
		;
		; Modif:
		; 04/09/93 : commentaires
		; Reprise 29/10/95: optimisations


		; usage des registres 8086
		; es:si, cl   : simulation du PC.
		; ax, bx, dx  : temporaires.
		; ch          : flags
		; di          : gnralement pointeur registre destination.
		; bp          : compteur d'instructions (_next)


		.data

		; ---------------------------------------
		; REGISTRES DU SATURN



		public _regA, _regB, _regC, _regD
		public _regR0,_regR1,_regR2,_regR3
		public _regR4,_regR5,_regR6,_regR7
		public _regD0,_regD1,_regPC
		public _Carry,_Pointer
		public _retstk,_lastaddr
                public _status,_hwstatus
		public _input,_output
		public _flagdec,_retstkptr
		public _tab_seg
		public _timer_to,_last_ticks
;		public _flagbrk    ; emu42

;                extrn _dsp_offset: byte

; structure des registres d'mulation:
; octet    nibbles
;  7         15,14
;  6         13,12
;  5         11,10
;  4         9, 8
;  3         7, 6
;  2         5, 4
;  1         3, 2
;  0         1, 0
; -1         x, 4      duplication du nibble 4 pour optimiser la vitesse :
; -2         x, x      les instructions sur le champ A utilisent
;                      les octets -1, 0 et 1
;                      (l'octet -2 est rserv pour la version 32 bits)


; ordre des quartets dans les registres
;  -2 -1  0  1  2 ...  7
;  xx x4 10 32 54 ... fe


; organisation mmoire:
;   ordre des quartets en mmoire:
;  octets     0  1  2  3  4 ...
;  quartets  10 32 54 76 98 ...
; pour lire un quartet  l'adresse n:
;   lire octet  adresse n/2                octet = 10
;   si n paire, quartet = octet & 15        quartet = 0
;   si n impaire, quartet = octet >> 4      quartet = 1
; pour lire un octet  l'adresse n:
;   lire mot  l'adresse n/2                mot = 3210
;   si n paire, octet = mot & 255           octet = 10
;   si n impaire, octet = (mot >> 4)&255    octet = 21


                align 2

tabopcode       dw 4096 dup(?)

_retstk         db 256 dup(?)         ; 64 level stack


reg             db 0, 0
_regA           db 10 dup (0)
_regB           db 10 dup (0)
_regC           db 10 dup (0)
_regD           db 10 dup (0)
_regR0          db 10 dup (0)
_regR1          db 10 dup (0)
_regR2          db 10 dup (0)
_regR3          db 10 dup (0)
_regR4          db 10 dup (0)
_regR5          db 10 dup (0)
_regR6          db 10 dup (0)
_regR7          db 10 dup (0)

		; control registers
_Carry          db 0
_Pointer        db 0
_regD0          db 0, 0, 0, 0
_regD1          db 0, 0, 0, 0
_regPC          db 0, 0, 0, 0
_oldretstk         db 16*4 dup(0)         ; pile  8 (16) niveaux
_lastaddr       db 0ffh, 0ffh, 0ffh, 0ffh
_status         dw 0    ; register ST 16 flags
_output         dw 0
_input          dw 0
_hwstatus       db 0    ; bits 3-0 : MP, SR, SB, XM
		db 0

		; internal registers
_flagdec        db 0       ; decimal mode flag
_retstkptr      db 0       ; retstk  pointer
_stackptr       dw 0
cpt_instr	dw 0
cpt_ticks       dw 0       ; compteur de ticks timer 71
_last_ticks     dw 0
_timer_to       db 0
;cpt_h		db 0
;_flagbrk        db 0
tabloaded       db 0

		; ---------------------------------------

_tab_seg        dw 16 dup(0)            ; table of memory segments (32k each)

;cpttimer        dw 0
;lasttic         dw 0
;cpttic          dw 0

; tab semi graphic character for text mode display
;tabsg           db 32, 223, 220, 219
;dspseg          dw 0

;lastbyte        db 0, 0




		.code


		public _executeSaturn

;                public _init_tab

;                public _updatedsp


FASTFLAGS       equ 1



; ==============================================================
; macros de lecture d'opcode
; ==============================================================


; macro de lecture d'un nibble de code
; al[3-0] = (PC); PC += 1;
; Note : al[7-4]=0
; PC est constitu de es:si et cl.
fetch_nib       macro
		local fetch_nib1
		mov  al,es:[si]
		xor  cl,4
		jne  fetch_nib1
		sar  al,4
		inc  si
fetch_nib1:     and  al,15
		endm

incr_pc         macro
		local incr_pc1
		xor  cl,4
		jne  incr_pc1
		inc  si
incr_pc1:
		endm


; macro de lecture d'un octet de code
; Note : ah dtruit
fetch_byte      macro
		mov  ax,es:[si]
		sar  ax,cl      ; al[3-0] = (PC); al[7-4]= (PC+1);
		inc  si         ; PC += 2;
		endm

; macro de lecture de 3 nibbles de code
; en ah[3-0], al
fetch_3nib      macro
		local fetch_3nib1
		mov  ax,es:[si]
		xor  cl,4
		jne  fetch_3nib1
		sar  ax,4
		inc  si
fetch_3nib1:    inc  si
		and  ah,15
		endm



; ***********************************************************
;               s/p & macros gnrales
; ***********************************************************



                IF FASTFLAGS

; flags:
; 7:decimal, 4:half-carry, 2:A-field, 1:pointer, 0:carry

		; met tat carry  0
clear_carry     macro
                and  ch,0eeh
		endm

		; met tat carry  1
set_carry       macro
                or  ch,011h
		endm

sto_carry       macro
                and  ch,0eeh
		or   ch,ah
		endm

test_carry      macro
                test ch,011h
                endm

test_decimal    macro
                or   ch,ch
                endm

set_decimal     macro
                or   ch,080h
                endm

clr_decimal     macro
                and  ch,07fh
                endm

test_pointer    macro
                test ch,2
                endm

set_pointer     macro
                or   ch,2
                endm

clr_pointer     macro
                and  ch,0fdh
                endm

set_majw        macro
;                or   ch,4
                endm

clr_majw        macro
;                and  ch,0fbh
                endm

;test_majw       macro
;                test ch,4
;                endm

                ELSE

		; met tat carry  0
clear_carry     macro
		mov  _Carry,0
		endm

		; met tat carry  0
set_carry       macro
		mov  _Carry,0ffh
		endm

sto_carry       macro
		mov  _Carry,ah
		endm

test_carry      macro
                test _Carry,0ffh
                endm

test_decimal    macro
                test  _flagdec,0ffh
                endm

set_decimal     macro
                mov   _flagdec,0ffh
                endm

clr_decimal     macro
                mov   _flagdec,0
                endm

test_pointer    macro
                test _Pointer,0fh
                endm

set_pointer0    macro
                endm

clr_pointer0    macro
                endm


                ENDIF




		; met  jour tat carry
update_carry    macro
		lahf
		and  ah,1
                sto_carry
		endm

		; met  jour tat half carry
update_halfcarry    macro
		lahf
		and  ah,010h
                sto_carry
		endm


		align 2

		; mettre a jour registre entier (champ W) src et dest
		; src et dest sont points par si et di
maj2_w:		mov   al,[si+2]
		mov   ah,[si-1]
		and   ax,0ff0h
		or    al,ah
		mov   [si+2],al
		; mettre a jour registre entier (champ W) dest
		; dest est point par di
maj_w:          mov   al,[di+2]
		mov   ah,[di-1]
		and   ax,0ff0h
		or    al,ah
		mov   [di+2],al
		ret

maj_ws          macro
                mov   al,[di+2]
		mov   ah,[di-1]
		and   ax,0ff0h
		or    al,ah
		mov   [di+2],al
                endm

maj2_ws         macro
                mov   al,[bx+2]
		mov   ah,[bx-1]
		and   ax,0ff0h
		or    al,ah
		mov   [bx+2],al
                maj_ws
                endm

		; mettre  jour pseudo champ A de dest point par di
maj_a           macro
		mov   al,[di+2]
		mov   [di-1],al
		endm

; macro d'execution d'un opcode

exec_opcode	macro INSTRUCT, DEST, CHAMP
		mov   di, offset _reg&DEST&
		jmp   &INSTRUCT&_&CHAMP&
		endm

exec_opcode2	macro INSTRUCT, DEST, SRC, CHAMP
		push  si
		mov   si, offset _reg&SRC&
		mov   di, offset _reg&DEST&
		jmp   &INSTRUCT&_&CHAMP&
		endm

exec_opcode1s	macro INSTRUCT, DEST, CHAMP
		mov   di, offset _reg&DEST&
		jmp   short &INSTRUCT&_&CHAMP&
		endm

exec_opcode2s	macro INSTRUCT, DEST, SRC, CHAMP
		mov   bx, offset _reg&SRC&
		mov   di, offset _reg&DEST&
		jmp   short &INSTRUCT&_&CHAMP&
		endm




; ***********************************************************
;               moteur d'execution
;  excute une instruction pointe par PC : es:si et cl
; ***********************************************************


		align 2
_nexti:         xor  cl,4
                jne  _next
       		inc  si
_next:          dec  bp
		je   fin
_next0:         fetch_byte
		js   fixnext    ; no more needed on HP42S?
_next1:         and  ah,0fh
                mov  bx,ax
		add  bx,bx
		jmp  [tabopcode+bx]


NEXT            macro
		jmp  _next
		endm

NEXTI           macro
		jmp  _nexti
		endm



fixnext:	; segment suivant
		xor  bh,bh
		mov  bl,_regPC+2
		inc  bl
		mov  _regPC+2,bl
		add  bx,bx
		mov  bx,[_tab_seg+bx]
                or   bx,bx
                je   resetpc
                mov  es,bx
		and  si,07fffh
		jmp  short _next1
resetpc:       	; case es=0, jump to 0
                mov  _regPC+2,0
                xor  si,si
                xor  cl,cl
                jmp  short invalid   ;  !!!
                jmp  short _next




fin:
; fin de 256 instructions.
; test fin des cpt_instr instructions
;		mov  al,cpt_h
;		or   al,al
;		je   fin1
;		dec  al
;		mov  cpt_h,al
;		jmp  _next0
;fin1:
                xor  ax,ax
                jmp  short retour



invalid:        mov  ax,-1
                jmp  short retour
retouri:        incr_pc
retour:         push ax
		; convert regPC, si, cl en adr quartet 20 bits ah-di
		mov  ah,_regPC+2
		and  ah,0fh
		mov  al,cl
		mov  di,si
		shr  al,3   ; carry = bit 0
		rcl  di,1
		adc  ah,0        ; 02/01/97
		mov  word ptr _regPC,di
		mov  _regPC+2,ah

		irp  DEST,<A,B,C,D>
		mov  di,offset _reg&DEST&
		call maj_w
		endm
                mov  al,0fh
		and  _regD0+2,al
		and  _regD1+2,al

                IF FASTFLAGS
                mov  _Carry,0
                mov  _flagdec,0
                test ch,011h
                je   ff1
                mov  _Carry,0ffh
ff1:            test ch,080h
                je   ff2
                mov  _flagdec,0ffh
ff2:
                ENDIF

                ; setup stack pointer
                mov  al,_retstkptr
                shr  al,2
                mov  _retstkptr,al

		pop  ax
		ret




; ***********************************************************
;               point d'entree
; prototype C: int executeSaturn(int cpt);
; ***********************************************************


; point d'entre:
_executeSaturn  proc C far nb_instr:word, nb_ticks:word
		push si
		push di
	        mov  ax,nb_instr
		mov  bx,nb_ticks
                push bp
		call @executeSaturn
                pop  bp
                pop  di
		pop  si
		ret
_executeSaturn  endp



;_executeShndwn  proc C far nb_ticks: word
;		push si
;		push di
;		mov  bx,nb_ticks
;		add  cpt_ticks,bx
;		je   execShdn1
;               dec  cpt_ticks
;		call decr_timer
;execShdn1:	pop  di
;		pop  si
;		ret
;_executeShndwn  endp


@executeSaturn: mov  cpt_instr,ax
;;		add  cpt_ticks,bx
;;		cmp  ax,1
;;		jne  @exec1
;;		mov  cpt_ticks,0
@exec1:
                mov  al,tabloaded
                cmp  al,71
                je   @exec2
                call _init_tab
                mov  tabloaded,71
@exec2:
                ; setup stack
                mov  al,_retstkptr
                shl  al,2
                mov  _retstkptr,al

                ; setup PC
                mov  bl,_regPC+2
		and  bx,0fh
		mov  si,word ptr _regPC
		call bld_pointer       ; construit pointeur es:si,cl
                ; update A fields
		irp  DEST,<A,B,C,D>
		mov  di,offset _reg&DEST&
		maj_a
		endm
                ; setup bp as opcode counter
		mov  ax,cpt_instr
                mov  bp,ax
;;                xor  al,al
;;		mov  cpt_h,al


                ; setup flags
                IF FASTFLAGS
                xor  ch,ch
                test _Carry,0ffh
                je   fff1
                or   ch,1
fff1:           test _flagdec,0ffh
                je   fff2
                or   ch,080h
fff2:           test _Pointer,0fh
                je   fff3
                or   ch,2
fff3:
                ENDIF

		jmp  _next0






; ==============================================================

; decodage des instructions
; (dfinition de la table tab_opcode)

;		include sat_dec.inc

; ==============================================================









; ==================================================================

; dfinitions des instructions arithmtiques:
; x=x+y fs
; x=x-y fs
; x=y-x fs
; x=x+1 fs
; x=x-1 fs
; x=-x  fs
; x=-x-1 fs

;		include sat_arit.inc

; ==================================================================







; ==================================================================
; dfinitions des instructions de comparaison:
; ?x=y  fs
; ?x#y  fs
; ?x>y  fs
; ?x>=y fs
; ?x<y  fs
; ?x<=y fs
; ?x=0  fs
; ?x#0  fs

;		include sat_cmp.inc

; ==================================================================





; ==================================================================
; dfinitions des instructions logiques:
; x=x&y fs
; x=x!y fs
; xsl   fs
; xsr   fs
; x=0   fs
; x=y   fs
; xyex  fs
; Rn=x
; x=Rn
; xRnex

;		include sat_logi.inc

; ==================================================================





; ==================================================================
; dfinitions des instructions de saut:
; goc
; gonc
; goto
; gosub
; golong
; govlng
; gosubl
; gosbvl
; rtnsxm
; rtn
; rtnsc
; rtncc
; rti
; pc=a
; pc=c
; a=pc
; c=pc
; apcex
; cpcex
; pc=(a)
; pc=(c)

;		include sat_jmp.inc

; ==================================================================




; ==================================================================
; dfinitions des instructions de control:
; sethex
; setdec

; rstk=c
; c=rstk
; lchex
; lahex

; clrst
; c=st
; st=c
; cstex
; st=0  n
; st=1  n
; ?st=0 n
; ?st=1 n
; clrhst
; ?xx=0

; p=p+1
; p=p-1
; p=    n
; ?p=   n
; ?p#   n
; c+p+1
; c=p   n
; p=c   n
; cpex  n

; out=c
; out=cs
; a=in
; c=in
; uncnfg
; config
; c=id
; shutdn
; reset
; buscb
; buscc
; buscd
; sreq
; inton
; intoff
; rsi

; a=bit0 n
; a=bit1 n
; ?abit=0 n
; ?abit=1 n
; c=bit0 n
; c=bit1 n
; ?cbit=0 n
; ?cbit=1 n

; dx=(2)
; dx=(4)
; dx=(5)
; dx=dx+ n
; dx=dx- n
; dx=a
; dx=c
; adxex
; cdxex
; adxxs
; cdxxs

; a=a+con fs
; a=a-con fs
;		include sat_ctrl.inc

; ==================================================================



; ==================================================================
; dfinitions des instructions d'accs mmoire:
; a=dat0 fs
; c=dat0 fs
; a=dat1 fs
; c=dat1 fs
; dat0=a fs
; dat0=c fs
; dat1=a fs
; dat1=c fs

;		include sat_mem.inc

; ==================================================================


;		end

; ********************************************************************
;               mulation du processeur Saturn par le 8086
;                       J-F Garnier 1993, 2002
; --------------------------------------------------------------------
; Module sat_dec.asm : dcodage des instructions, table tab_opcode
; --------------------------------------------------------------------
; Version 1   16/03/93
;             04/09/93 :
;            commentaires,
;            corrig decode9x, decodeAx,  decodeBx,
;            complt tab_opcode1x
; Version 2   28/12/94
; version 2.1 09/09/01 HP42S support
; Version 3.0 03/10/02 12 bits jump table
; ********************************************************************




; macros de dfinitions des tables d'opcode arithmtiques

TAB_ARITH1 	macro ADR
		dw  _add_A_B&ADR&, _add_B_C&ADR&, _add_C_A&ADR&, _add_D_C&ADR&
		dw  _add_A_A&ADR&, _add_B_B&ADR&, _add_C_C&ADR&, _add_D_D&ADR&
		dw  _add_B_A&ADR&, _add_C_B&ADR&, _add_A_C&ADR&, _add_C_D&ADR&
		dw  _dec_A&ADR&,   _dec_B&ADR&,   _dec_C&ADR&,   _dec_D&ADR&
		endm

TAB_ARITH2 	macro ADR
		dw  _clr_A&ADR&,   _clr_B&ADR&,   _clr_C&ADR&,   _clr_D&ADR&
		dw  _mov_A_B&ADR&, _mov_B_C&ADR&, _mov_C_A&ADR&, _mov_D_C&ADR&
		dw  _mov_B_A&ADR&, _mov_C_B&ADR&, _mov_A_C&ADR&, _mov_C_D&ADR&
		dw  _xch_A_B&ADR&, _xch_B_C&ADR&, _xch_A_C&ADR&, _xch_C_D&ADR&
		endm

TAB_ARITH3 	macro ADR
		dw  _sub_A_B&ADR&, _sub_B_C&ADR&, _sub_C_A&ADR&, _sub_D_C&ADR&
		dw  _inc_A&ADR&,   _inc_B&ADR&,   _inc_C&ADR&,   _inc_D&ADR&
		dw  _sub_B_A&ADR&, _sub_C_B&ADR&, _sub_A_C&ADR&, _sub_C_D&ADR&
		dw  _subr_A_B&ADR&,_subr_B_C&ADR&,_subr_C_A&ADR&,_subr_D_C&ADR&
		endm

TAB_ARITH4 	macro ADR
		dw  _sl_A&ADR&,  _sl_B&ADR&,  _sl_C&ADR&,  _sl_D&ADR&
		dw  _sr_A&ADR&,  _sr_B&ADR&,  _sr_C&ADR&,  _sr_D&ADR&
		dw  _neg_A&ADR&, _neg_B&ADR&, _neg_C&ADR&, _neg_D&ADR&
		dw  _not_A&ADR&, _not_B&ADR&, _not_C&ADR&, _not_D&ADR&
		endm

TAB_CMP1 	macro ADR
		dw  _eq_A_B&ADR&, _eq_B_C&ADR&, _eq_C_A&ADR&, _eq_D_C&ADR&
		dw  _ne_A_B&ADR&, _ne_B_C&ADR&, _ne_C_A&ADR&, _ne_D_C&ADR&
		dw  _eq_A&ADR&,   _eq_B&ADR&,   _eq_C&ADR&,   _eq_D&ADR&
		dw  _ne_A&ADR&,   _ne_B&ADR&,   _ne_C&ADR&,   _ne_D&ADR&
		endm

TAB_CMP2 	macro ADR
		dw  _gt_A_B&ADR&, _gt_B_C&ADR&, _gt_C_A&ADR&, _gt_D_C&ADR&
		dw  _gt_B_A&ADR&, _gt_C_B&ADR&, _gt_A_C&ADR&, _gt_C_D&ADR&
		dw  _ge_A_B&ADR&, _ge_B_C&ADR&, _ge_C_A&ADR&, _ge_D_C&ADR&
		dw  _ge_B_A&ADR&, _ge_C_B&ADR&, _ge_A_C&ADR&, _ge_C_D&ADR&
		endm


; ------------------------------------------------------
; table principale des opcodes du Saturn
; ------------------------------------------------------

		align 2





comment !
tabopcode       dw  _rtnsxm, _rtn,    _rtnsc,   _rtncc
		dw  _sethex, _setdec, _rstk_c,  _c_rstk
		dw  _clrst,  _c_st,   _st_c,    _cstex
		dw  _incr_p, _decr_p,  decode0E, _rti

		dw  4 dup (decode1x)
		dw  decode1x, decode15, _inc_D0,  _inc_D1
		dw  _dec_D0,  _ld_D0_2, _ld_D0_4, _ld_D0_5
		dw  _dec_D1,  _ld_D1_2, _ld_D1_4, _ld_D1_5

		dw  16 dup (_p_x)

                dw  16 dup (_lchex)

		dw  16 dup (_goc)

		dw  16 dup (_gonc)

		dw  16 dup (_goto)

		dw  16 dup (_gosub)

		dw  decode80,decode81,_clrhst,  _tsthst
		dw  _clr_st, _set_st, _tst_st0, _tst_st1
		dw  _tst_p0, _tst_p1, decode8A, decode8B
		dw  _golong, _govlng, _gosubl,  _gosbvl

		dw  16 dup (decode9x)

		dw  16 dup (decodeAx)

		dw  16 dup (decodeBx)

		TAB_ARITH1 _a

		TAB_ARITH2 _a

		TAB_ARITH3 _a

		TAB_ARITH4 _a

!

_init_tab:
                push si
                push di
                mov si, offset tabcode
                xor ax,ax
                mov bx,ax
                mov di,ax
init_tab1:
                mov cx,cs:[si]
                inc si
                inc si
                or  cx,cx
                je  init_tab9
                mov dx,cx
                ; cl = nb of entries, ch= nb of repeats
init_tab2:
                mov ch,dh
                mov ax,cs:[si]
                inc si
                inc si
init_tab3:
                add bx,bx
                mov [bx+tabopcode],ax
                inc di
                shr bx,1
                inc bh
                and bh,0fh
                jne init_tab4
                add bl,10h
                jnc init_tab4
                inc bl
init_tab4:      dec ch
                jne init_tab3
                dec cl
                jne init_tab2
                jmp short init_tab1
init_tab9:      mov ax,di
                pop di
                pop si
                ret
;_init_tab       endp



                align 2

                ;  #entries, #repeat
                ; **  means return in _nexti (3 nibs opcodes)
tabcode         db  16, 16    ; 0x
                dw  _rtnsxm, _rtn,    _rtnsc,   _rtncc
		dw  _sethex, _setdec, _rstk_c,  _c_rstk
		dw  _clrst,  _c_st,   _st_c,    _cstex
		dw  _incr_p, _decr_p,  decode0E, _rti

                db  80 , 1   ; 10x- 14x **
        	dw  _mov_R0_A, _mov_R1_A, _mov_R2_A, _mov_R3_A
		dw  _mov_R4_A, _mov_R5_A, _mov_R6_A, _mov_R7_A
		dw  _mov_R0_C, _mov_R1_C, _mov_R2_C, _mov_R3_C
		dw  _mov_R4_C, _mov_R5_C, _mov_R6_C, _mov_R7_C
                dw  _mov_A_R0, _mov_A_R1, _mov_A_R2, _mov_A_R3
		dw  _mov_A_R4, _mov_A_R5, _mov_A_R6, _mov_A_R7
		dw  _mov_C_R0, _mov_C_R1, _mov_C_R2, _mov_C_R3
		dw  _mov_C_R4, _mov_C_R5, _mov_C_R6, _mov_C_R7
		dw  _xch_A_R0, _xch_A_R1, _xch_A_R2, _xch_A_R3
		dw  _xch_A_R4, _xch_A_R5, _xch_A_R6, _xch_A_R7
		dw  _xch_C_R0, _xch_C_R1, _xch_C_R2, _xch_C_R3
		dw  _xch_C_R4, _xch_C_R5, _xch_C_R6, _xch_C_R7
		dw  _mov_D0_A, _mov_D1_A, _xch_A_D0, _xch_A_D1
		dw  _mov_D0_C, _mov_D1_C, _xch_C_D0, _xch_C_D1
		dw  _mov_D0_A_s, _mov_D1_A_s, _xch_A_D0_s, _xch_A_D1_s
		dw  _mov_D0_C_s, _mov_D1_C_s, _xch_C_D0_s, _xch_C_D1_s
               	dw  _sto_A_D0_a, _sto_A_D1_a, _lod_A_D0_a, _lod_A_D1_a
		dw  _sto_C_D0_a, _sto_C_D1_a, _lod_C_D0_a, _lod_C_D1_a
		dw  _sto_A_D0_b, _sto_A_D1_b, _lod_A_D0_b, _lod_A_D1_b
		dw  _sto_C_D0_b, _sto_C_D1_b, _lod_C_D0_b, _lod_C_D1_b
                db  1, 8    ; 150-157
                dw  decode15
                db  8, 1    ; 158-15F
		dw  _sto_A_D0_n, _sto_A_D1_n, _lod_A_D0_n, _lod_A_D1_n
		dw  _sto_C_D0_n, _sto_C_D1_n, _lod_C_D0_n, _lod_C_D1_n
                db  10, 16  ; 16x - 1Fx
		dw  _inc_D0,  _inc_D1
		dw  _dec_D0,  _ld_D0_2, _ld_D0_4, _ld_D0_5
		dw  _dec_D1,  _ld_D1_2, _ld_D1_4, _ld_D1_5

                db  1, 16       ; 2x
		dw  _p_0
                db  1, 240
                dw  _p_x
                db  6, 16      ; 30x to 35x
                dw  _lchex1, _lchex2, _lchex3, _lchex4, _lchex5, _lchex6
                db  1, 160     ; 3Xx
                dw  _lchex
                db  1, 1       ; 4xx
                dw  _rtnc
                db  1, 255
                dw  _goc
                db  1, 1       ; 5xx
                dw  _rtnnc
                db  1, 255
                dw  _gonc
                db  1, 0       ; 6xxx
                dw  _goto
                db  1, 0       ; 7xxx
                dw  _gosub

                db  32, 1      ; 80x - 81x **
        	dw  _mov_OUT_CS, _mov_OUT_C, _mov_A_IN, _mov_C_IN
		dw  _uncnfg, _config, _mov_C_ID, _shutdn
		dw  decode808, _add_C_P_1, _reset, _buscc
		dw  _mov_C_P_n, _mov_P_C_n, _sreq, _xch_C_P_n
                dw  _slc_A, _slc_B, _slc_C, _slc_D
		dw  _src_A, _src_B, _src_C, _src_D
		dw  decode818, decode819, decode81a, decode81b
		dw  _srb_A, _srb_B, _srb_C, _srb_D
                db  8, 16    ; 82x - 89x **
		dw  _clrhst,  _tsthst
		dw  _clr_st, _set_st, _tst_st0, _tst_st1
		dw  _tst_p0, _tst_p1
                db  32, 1    ;   8Ax - 8Bx **
		TAB_CMP1 _a
		TAB_CMP2 _a
                db  4, 16    ; 8Cx - 8Fx
		dw  _golong, _govlng, _gosubl,  _gosbvl

                db  0, 1       ; 9xx **
		TAB_CMP1 _p
		TAB_CMP1 _wp
		TAB_CMP1 _xs
		TAB_CMP1 _x
		TAB_CMP1 _s
		TAB_CMP1 _m
		TAB_CMP1 _b
		TAB_CMP1 _w
		TAB_CMP2 _p
		TAB_CMP2 _wp
		TAB_CMP2 _xs
		TAB_CMP2 _x
		TAB_CMP2 _s
		TAB_CMP2 _m
		TAB_CMP2 _b
		TAB_CMP2 _w

                db  0, 1      ; Axx  **
		TAB_ARITH1 _p
		TAB_ARITH1 _wp
		TAB_ARITH1 _xs
		TAB_ARITH1 _x
		TAB_ARITH1 _s
		TAB_ARITH1 _m
		TAB_ARITH1 _b
		TAB_ARITH1 _w
		TAB_ARITH2 _p
		TAB_ARITH2 _wp
		TAB_ARITH2 _xs
		TAB_ARITH2 _x
		TAB_ARITH2 _s
		TAB_ARITH2 _m
		TAB_ARITH2 _b
		TAB_ARITH2 _w

                db  0, 1     ; Bxx  **
		TAB_ARITH3 _p
		TAB_ARITH3 _wp
		TAB_ARITH3 _xs
		TAB_ARITH3 _x
		TAB_ARITH3 _s
		TAB_ARITH3 _m
		TAB_ARITH3 _b
		TAB_ARITH3 _w
		TAB_ARITH4 _p
		TAB_ARITH4 _wp
		TAB_ARITH4 _xs
		TAB_ARITH4 _x
		TAB_ARITH4 _s
		TAB_ARITH4 _m
		TAB_ARITH4 _b
		TAB_ARITH4 _w

                db  64, 16   ; Cx - Fx
		TAB_ARITH1 _a
                TAB_ARITH2 _a
                TAB_ARITH3 _a
                TAB_ARITH4 _a

                db  0, 0     ; end of table


; ---------------------------------------------------------------------


; dcodage des opcodes 0Exy



decode0E:	fetch_byte
		ror  al,4       ; pour faciliter dcodage
		xor  bx,bx
		mov  bl,al
		add  bx,bx
		jmp  cs:[tab_opcode0E+bx]



; macros de dfinitions d'entres de la table des opcodes 0Exy

TAB_0E 		macro ADR
		dw  _and_A_B&ADR&, _and_B_C&ADR&, _and_C_A&ADR&, _and_D_C&ADR&
		dw  _and_B_A&ADR&, _and_C_B&ADR&, _and_A_C&ADR&, _and_C_D&ADR&
		dw  _or_A_B&ADR&,  _or_B_C&ADR&,  _or_C_A&ADR&,  _or_D_C&ADR&
		dw  _or_B_A&ADR&,  _or_C_B&ADR&,  _or_A_C&ADR&,  _or_C_D&ADR&
		endm

TAB_INV8	macro
		rept 8
		dw  invalid
		endm
		endm


; table des opcodes 0Exy
		align 2
tab_opcode0E    label word
		TAB_0E _p
		TAB_0E _wp
		TAB_0E _xs
		TAB_0E _x
		TAB_0E _s
		TAB_0E _m
		TAB_0E _b
		TAB_0E _w
		rept 7
		TAB_INV8
		TAB_INV8
		endm
		TAB_0E _a



; ---------------------------------------------------------------------

; dcodage des opcodes 150y-157y

decode15:
		fetch_byte
		ror  al,4       ; pour faciliter dcodage
	        xor  bx,bx
		mov  bl,al      ; here, 0<bl<127
		add  bl,bl
		jmp  cs:[tab_opcode15+bx]

; macro de dfinitions d'entres de la table des opcodes 15xy

TAB_CHAMP 	macro INSTRUCT
		dw  &INSTRUCT&_p, &INSTRUCT&_wp, &INSTRUCT&_xs, &INSTRUCT&_x
		dw  &INSTRUCT&_s, &INSTRUCT&_m,  &INSTRUCT&_b,  &INSTRUCT&_w
		endm

; table des opcodes 15xy
		align 2
tab_opcode15    label word
		TAB_CHAMP _sto_A_D0
		TAB_INV8
		TAB_CHAMP _sto_A_D1
		TAB_INV8
		TAB_CHAMP _lod_A_D0
		TAB_INV8
		TAB_CHAMP _lod_A_D1
		TAB_INV8
		TAB_CHAMP _sto_C_D0

;		TAB_INV8
                rept 7
                dw  invalid
                endm
                dw  _stoil_C_D0_8  ; 154f write il for ilmailbox emulation

		TAB_CHAMP _sto_C_D1
		TAB_INV8
		TAB_CHAMP _lod_C_D0

;		TAB_INV8
                rept 6
                dw  invalid
                endm
                dw  _lodil_C_D0_7  ; 156e read il for ilmailbox emulation
                dw  invalid

		TAB_CHAMP _lod_C_D1
		TAB_INV8




; ---------------------------------------------------------------------


; dcodage des opcodes 808x :
decode808:
                fetch_byte
                shr  al,4
		xor  bx,bx
		mov  bl,al
		add  bx,bx
		jmp  cs:[tab_opcode808+bx]

                align 2
tab_opcode808   dw  _inton, _rsi, _lahex, _buscb
                dw  _clrb_A_n, _setb_A_n, _isclr_A_n, _isset_A_n
                dw  _clrb_C_n, _setb_C_n, _isclr_C_n, _isset_C_n
                dw  _movi_PC_A, _buscd, _movi_PC_C, _intoff


; ---------------------------------------------------------------------

; dcodage des opcodes 818-81b


decode818:
                incr_pc
;                test _flagbrk,0ffh
;                jne  inv81x
                fetch_3nib
                ; al[3-0]=field, al[5-4]={A,B,C,D} reg, al[7-6]=operation, ah[3-0]=n
                ; operation: 0: +CON(1-16), 2: -CON(1-16)
                ; (extension: 1: +CON(17-32), 3: -CON(17-32))
                jmp  _con_Reg_n
inv81x:         jmp  invalid

decode819:      ; srb reg fs
                incr_pc
;                test _flagbrk,0ffh
;                jne  inv81x
                fetch_byte
                ; al[3-0]=field, al[5-4]={A,B,C,D} reg, al[7-6]=operation
                ; operation: 0=SRB
                ; (extension: 1-3: ??)
                jmp _srb_Reg_n

decode81a:      ; mov or xchg [A,C] <-> [R0,R1,R2,R3,R4] fs
                jmp  invalid    ; TBD
                incr_pc
                fetch_3nib
                ; al[3-0]=field, al[7-4]=operation, ah[3]={A,C} reg, ah[2-0]=Rx reg
                ; operation: 0=mov Rx,{A,C} ; 1=mov {A,C},Rx ; 2=xchg {A,C},Rx
                ; (extension  4=ST+, 5=ST-, 6=RC+, 7=RC- )
                NEXT

decode81b:      ; 81bx new instructions on PC
;                test _flagbrk,0ffh
;                jne  inv81x
                fetch_byte
                shr  al,4
		xor  bx,bx
		mov  bl,al
		add  bx,bx
		jmp  cs:[tab_opcode81b+bx]

tab_opcode81b   dw  invalid, invalid, _mov_PC_A, _mov_PC_C
                dw  _mov_A_PC, _mov_C_PC, _xch_A_PC, _xch_C_PC
                dw  4 dup(invalid)
                dw  4 dup(invalid)



; fpu interface xtensions:
; FMULT    in: 0.1 <= C, A < 1 ;  out: 0.01 <= C < 1
; FDIV     in: 0.1 <= C, A < 1 ;  out: 0.01 <= C < 1
; FSQRT    in: 0.01 <= C < 1   ;  out: 0.1  <= C < 1
; FINV
; FPTAN    in: 0 < C < PI/4    ;  out:
; FPATAN
; FLOG10   in: 0.1 <= C < 1     ; out: 0 <= C < 1
; FPOW10   in: 0 <= C < 1      ; out: 1 <= C < 10

; ********************************************************************
;               mulation du processeur Saturn par le 8086
;                       J-F Garnier 1993, 95
; --------------------------------------------------------------------
; Module sat_arit.asm : mulation des instructions arithmtiques
; --------------------------------------------------------------------
; Version 1.0 16/03/93
;   prototype
; Version 1.1 01/11/95 : modif. traitement dec/hex, macros.
;                        corrections, documentation.
; ********************************************************************


; -------------------------------------------------------------------
; routines arithmtiques :
; x=x+y fs
; x=x-y fs
; x=y-x fs
; x=x+1 fs
; x=x-1 fs
; x=-x  fs
; x=-x-1 fs
;
; addition
; soustraction dans l'un ou l'autre
; incrementation
; decrementation
; negation
; complmentation
;
; mode dcimal ou hexa
; ajuste Carry
; -------------------------------------------------------------------


; ***************************************************************
; 		dfinitions des macros
; ***************************************************************


; -------------------------------------------------------------------
; build_arith2
; -------------------------------------------------------------------
; macro de traitement des instructions arithmtiques  2 arguments
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT = _add, _sub, _subr
; entres: si, di pointent sur les registres oprandes
;          si: source
;          di: destination
;          la sauvegarde de si (pointeur programme) est dans la pile
;          et doit tre restaur avant NEXT
; -------------------------------------------------------------------

build_arith2	macro INSTRUCT
		local lbl_p1, lbl_p2, lbl_p3, lbl_p8, lbl_p9
                local lbl_wp1, lbl_wp2
		local tab_arith2h_wp, tab_arith2d_wp
		local lbl_xs1, lbl_xs2
                local lbl_x1, lbl_x2
		local lbl_s1, lbl_s2
                local lbl_m1, lbl_m2
                local lbl_w1, lbl_w2

&INSTRUCT&_p:
		push  di
		call  maj2_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    lbl_p2
		; paire, quartet infrieur
		add   si,ax
		add   di,ax
		test_decimal
		js    lbl_p1
		call  &INSTRUCT&h_lonib
		jmp   short lbl_p9
lbl_p1:         call  &INSTRUCT&d_lonib       ; !
		jmp   short lbl_p9
lbl_p2:         ; impaire, quartet suprieur
		add   si,ax
		add   di,ax
		test_decimal
		js    lbl_p3
		call  &INSTRUCT&h_hinib
		jmp   short lbl_p8
lbl_p3:         call  &INSTRUCT&d_hinib       ; !
lbl_p8:		update_carry
lbl_p9:         pop   di
		maj_a
		pop   si
		NEXTI


&INSTRUCT&_wp:
		push  di
		call  maj2_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		test_decimal
                jns   lbl_wp1
                add   bx,32
lbl_wp1:	call  cs:[bx+tab_arith2h_wp]
                pop   di
		maj_a
		pop   si
		NEXTI
tab_arith2h_wp	dw    &INSTRUCT&h_1,  &INSTRUCT&h_2,  &INSTRUCT&h_3,  &INSTRUCT&h_4
		dw    &INSTRUCT&h_5,  &INSTRUCT&h_6,  &INSTRUCT&h_7,  &INSTRUCT&h_8
		dw    &INSTRUCT&h_9,  &INSTRUCT&h_10, &INSTRUCT&h_11, &INSTRUCT&h_12
		dw    &INSTRUCT&h_13, &INSTRUCT&h_14, &INSTRUCT&h_15, &INSTRUCT&h_16
tab_arith2d_wp	dw    &INSTRUCT&d_1,  &INSTRUCT&d_2,  &INSTRUCT&d_3,  &INSTRUCT&d_4
		dw    &INSTRUCT&d_5,  &INSTRUCT&d_6,  &INSTRUCT&d_7,  &INSTRUCT&d_8
		dw    &INSTRUCT&d_9,  &INSTRUCT&d_10, &INSTRUCT&d_11, &INSTRUCT&d_12
		dw    &INSTRUCT&d_13, &INSTRUCT&d_14, &INSTRUCT&d_15, &INSTRUCT&d_16


&INSTRUCT&_xs:
		inc   si
		inc   di
		test_decimal
		js    lbl_xs1
		call  &INSTRUCT&h_lonib
		pop   si
		NEXTI
lbl_xs1:	call  &INSTRUCT&d_lonib
		pop   si
		NEXTI

&INSTRUCT&_x:
		test_decimal
		js    lbl_x1
		call  &INSTRUCT&h_3
		pop   si
		NEXTI
lbl_x1:		call  &INSTRUCT&d_3
		pop   si
		NEXTI

comment !
&INSTRUCT&_s:   add   si,7
		add   di,7
		test_decimal
		js    lbl_s1
		call  &INSTRUCT&h_hinib
		jmp   short lbl_s2
lbl_s1:		call  &INSTRUCT&d_hinib
lbl_s2:		update_carry
		pop   si
		NEXTI
!

&INSTRUCT&_m:
		push  di
		call  maj2_w
		inc   si
		inc   di
		test_decimal
                js    lbl_m1
		call  &INSTRUCT&h_hinib
		inc   si
		inc   di
		call  &INSTRUCT&h_11
                jmp   short lbl_m2
lbl_m1:		call  &INSTRUCT&d_hinib
		inc   si
		inc   di
		call  &INSTRUCT&d_11
lbl_m2:         pop   di
		maj_a
		pop   si
		NEXTI

		endm







; ===============================================================


; **************************************************************
; 		sub
; dest = dest - src
; **************************************************************

; fonctions de soustraction hexa

subh_byte       macro
		mov  al,[si]
		sbb  [di],al
		inc  si
		inc  di
		endm

_subh_hinib:    mov  al,[si]
		and  al,0f0h
		sub  [di],al
		ret

_subh_16:       subh_byte
_subh_14:       subh_byte
_subh_12:       subh_byte
_subh_10:       subh_byte
_subh_8:        subh_byte
_subh_6:        subh_byte
_subh_4:        subh_byte
_subh_2:        subh_byte
		update_carry
		ret

_subh_15:       subh_byte
_subh_13:       subh_byte
_subh_11:       subh_byte
_subh_9:        subh_byte
_subh_7:        subh_byte
_subh_5:        subh_byte
_subh_3:        subh_byte
_subh_1:
_subh_lonib:    mov  al,[di]
		mov  bh,al
		sbb  al,[si]
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret

; fonctions de soustraction dcimal

subd_byte       macro
		mov  al,[di]
		sbb  al,[si]
		das
		mov  [di],al
		inc  si
		inc  di
		endm

_subd_hinib:	;
		mov  al,[di]
		mov  ah,[si]
		mov  dl,al      ; copie original destination
		and  ax,0f0f0h  ; masque
		sub  al,ah
		jnc  subd_hi1
		das
subd_hi1:	lahf
		and  dl,0fh
		or   al,dl
		sahf
		mov  [di],al
		ret

_subd_16:       subd_byte
_subd_14:       subd_byte
_subd_12:       subd_byte
_subd_10:       subd_byte
_subd_8:        subd_byte
_subd_6:        subd_byte
_subd_4:        subd_byte
_subd_2:        subd_byte
		update_carry
		ret

_subd_15:       subd_byte
_subd_13:       subd_byte
_subd_11:       subd_byte
_subd_9:        subd_byte
_subd_7:        subd_byte
_subd_5:        subd_byte
_subd_3:        subd_byte
_subd_1:
_subd_lonib:
		mov  al,[di]
		mov  bh,al
		sbb  al,[si]
		lahf
		and  ah,16
		je   subd_lo1
		sahf
		das
;		update_halfcarry
subd_lo1:	sto_carry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret




; ----------------------------------------


                align 2
_sub_A_B_a:     exec_opcode2s _sub A B a
_sub_B_C_a:     exec_opcode2s _sub B C a
_sub_C_A_a:     exec_opcode2s _sub C A a
_sub_D_C_a:     exec_opcode2s _sub D C a
_sub_B_A_a:     exec_opcode2s _sub B A a
_sub_C_B_a:     exec_opcode2s _sub C B a
_sub_A_C_a:     exec_opcode2s _sub A C a
_sub_C_D_a:     exec_opcode2s _sub C D a

_sub_a:
		test_decimal
		js   _subd_a
_subh_a:        mov  ax,[bx]
		sub  [di],ax
		mov  al,[bx-1]
		sbb  [di-1],al
		update_halfcarry
		NEXT
_subd_a:	mov  ax,[di]
		mov  dx,[bx]
		sub  al,dl
		das
		xchg al,ah
		sbb  al,dh
		das
		xchg al,ah
		mov  [di],ax
		mov  al,[di-1]
		sbb  al,[bx-1]
		das
		mov  [di-1],al
		update_halfcarry
		NEXT



                align 2
_sub_A_B_b:     exec_opcode2s _sub A B b
_sub_B_C_b:     exec_opcode2s _sub B C b
_sub_C_A_b:     exec_opcode2s _sub C A b
_sub_D_C_b:     exec_opcode2s _sub D C b
_sub_B_A_b:     exec_opcode2s _sub B A b
_sub_C_B_b:     exec_opcode2s _sub C B b
_sub_A_C_b:     exec_opcode2s _sub A C b
_sub_C_D_b:     exec_opcode2s _sub C D b

_sub_b:
		test_decimal
		js   _subd_b
_subh_b:        mov  al,[bx]
		sub  [di],al
		update_carry
		NEXTI
_subd_b:	mov  al,[di]
		sub  al,[bx]
		das
		mov  [di],al
		update_carry
		NEXTI



_sub_w:         call maj2_w
		test_decimal
		js   _subd_w
_subh_w:        mov  ax,[si]
		sub  [di],ax
		mov  ax,[si+2]
		sbb  [di+2],ax
		mov  ax,[si+4]
		sbb  [di+4],ax
		mov  ax,[si+6]
		sbb  [di+6],ax
		update_carry
		maj_a
                pop  si
		NEXTI

_subd_w:
                mov  ax,[di]
		mov  bx,[si]
		sub  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di],ax
		mov  ax,[di+2]
		mov  bx,[si+2]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+2],ax
		mov  ax,[di+4]
		mov  bx,[si+4]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+4],ax
		mov  ax,[di+6]
		mov  bx,[si+6]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+6],ax
		update_carry
		maj_a
		pop  si
		NEXTI


                align 2
_sub_A_B_s:     exec_opcode2s _sub A B s
_sub_B_C_s:     exec_opcode2s _sub B C s
_sub_C_A_s:     exec_opcode2s _sub C A s
_sub_D_C_s:     exec_opcode2s _sub D C s
_sub_B_A_s:     exec_opcode2s _sub B A s
_sub_C_B_s:     exec_opcode2s _sub C B s
_sub_A_C_s:     exec_opcode2s _sub A C s
_sub_C_D_s:     exec_opcode2s _sub C D s

_sub_s:
		test_decimal
		js   _subd_s
_subh_s:    	mov  al,[bx+7]
		and  al,0f0h
		sub  [di+7],al
		update_carry
		NEXTI
_subd_s:	mov  al,[di+7]
		mov  ah,[bx+7]
		mov  dl,al      ; copie original destination
		and  ax,0f0f0h  ; masque   is it needed due to das
		sub  al,ah
		jnc  subd_s1
		das
subd_s1:	update_carry
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		NEXTI


; dfinitions du reste du code
		build_arith2 _sub



; ----------------------------------------



; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m, w>

_sub_A_B_&CHAMP&:     exec_opcode2 _sub A B CHAMP
_sub_B_C_&CHAMP&:     exec_opcode2 _sub B C CHAMP
_sub_C_A_&CHAMP&:     exec_opcode2 _sub C A CHAMP
_sub_D_C_&CHAMP&:     exec_opcode2 _sub D C CHAMP
_sub_B_A_&CHAMP&:     exec_opcode2 _sub B A CHAMP
_sub_C_B_&CHAMP&:     exec_opcode2 _sub C B CHAMP
_sub_A_C_&CHAMP&:     exec_opcode2 _sub A C CHAMP
_sub_C_D_&CHAMP&:     exec_opcode2 _sub C D CHAMP

		endm




; **************************************************************
; 		subr
; dest = src - dest
; **************************************************************

; fonctions de soustraction inverse hexa

subrh_byte      macro
		mov  al,[si]
		sbb  al,[di]
		mov  [di],al
		inc  si
		inc  di
		endm

_subrh_hinib:	;
		mov  al,[si]
		mov  ah,[di]
                mov  dh,ah
		and  ax,0f0f0h
                and  dh,00fh
                or   al,dh
		sub  al,ah
		mov  [di],al
		ret

_subrh_16:      subrh_byte
_subrh_14:      subrh_byte
_subrh_12:      subrh_byte
_subrh_10:      subrh_byte
_subrh_8:       subrh_byte
_subrh_6:       subrh_byte
_subrh_4:       subrh_byte
_subrh_2:       subrh_byte
		update_carry
		ret

_subrh_15:      subrh_byte
_subrh_13:      subrh_byte
_subrh_11:      subrh_byte
_subrh_9:       subrh_byte
_subrh_7:       subrh_byte
_subrh_5:       subrh_byte
_subrh_3:       subrh_byte
_subrh_1:
_subrh_lonib:
		mov  al,[si]
		mov  dh,[di]
		sbb  al,dh
		update_halfcarry
		mov  ah,dh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret



; fonctions de soustraction inverse dcimal

subrd_byte      macro
		mov  al,[si]
		sbb  al,[di]
		das
		mov  [di],al
		inc  si
		inc  di
		endm

_subrd_hinib:	;
		mov  al,[si]
		mov  ah,[di]
		mov  dl,ah      ; copie original destination
		and  ax,0f0f0h  ; masque
		sub  al,ah
		jnc  subrd_hi1
		das
subrd_hi1:	lahf
		and  dl,0fh
		or   al,dl
		sahf
		mov  [di],al
		ret

_subrd_16:      subrd_byte
_subrd_14:      subrd_byte
_subrd_12:      subrd_byte
_subrd_10:      subrd_byte
_subrd_8:       subrd_byte
_subrd_6:       subrd_byte
_subrd_4:       subrd_byte
_subrd_2:       subrd_byte
		update_carry
		ret

_subrd_15:      subrd_byte
_subrd_13:      subrd_byte
_subrd_11:      subrd_byte
_subrd_9:       subrd_byte
_subrd_7:       subrd_byte
_subrd_5:       subrd_byte
_subrd_3:       subrd_byte
_subrd_1:
_subrd_lonib:
		mov  al,[si]
		mov  dh,[di]
		sbb  al,dh
		lahf
		and  ah,16
		je   subrd_lo1
		sahf
		das
;		update_halfcarry
subrd_lo1:	sto_carry
		mov  ah,dh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


; ----------------------------------------

		align 2

_subr_A_B_a:     exec_opcode2s _subr A B a
_subr_B_C_a:     exec_opcode2s _subr B C a
_subr_C_A_a:     exec_opcode2s _subr C A a
_subr_D_C_a:     exec_opcode2s _subr D C a

_subr_a:
		test_decimal
                js   _subrd_a
_subrh_a:       mov  ax,[bx]
		sub  ax,[di]
		mov  [di],ax
		mov  al,[bx-1]
		sbb  al,[di-1]
		mov  [di-1],al
		update_halfcarry
		NEXT

_subrd_a:       mov  ax,[bx]
		mov  dx,[di]
		sub  al,dl
		das
		xchg al,ah
		sbb  al,dh
		das
		xchg al,ah
		mov  [di],ax
		mov  al,[bx-1]
		sbb  al,[di-1]
		das
		mov  [di-1],al
		update_halfcarry
		NEXT


		align 2

_subr_A_B_b:     exec_opcode2s _subr A B b
_subr_B_C_b:     exec_opcode2s _subr B C b
_subr_C_A_b:     exec_opcode2s _subr C A b
_subr_D_C_b:     exec_opcode2s _subr D C b

_subr_b:
		test_decimal
		js   _subrd_b
_subrh_b:	mov  al,[bx]
		sub  al,[di]
		mov  [di],al
		update_carry
		NEXTI
_subrd_b:	mov  al,[bx]
		sub  al,[di]
		das
		mov  [di],al
		update_carry
		NEXTI

_subr_w:        call maj2_w
		test_decimal
		js   _subrd_w
_subrh_w:       mov  ax,[si]
		sub  ax,[di]
		mov  [di],ax
		mov  ax,[si+2]
		sbb  ax,[di+2]
		mov  [di+2],ax
		mov  ax,[si+4]
		sbb  ax,[di+4]
		mov  [di+4],ax
		mov  ax,[si+6]
		sbb  ax,[di+6]
		mov  [di+6],ax
		update_carry
		maj_a
		pop  si
		NEXTI

_subrd_w:       mov  ax,[si]
		mov  bx,[di]
		sub  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di],ax
		mov  ax,[si+2]
		mov  bx,[di+2]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+2],ax
		mov  ax,[si+4]
		mov  bx,[di+4]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+4],ax
		mov  ax,[si+6]
		mov  bx,[di+6]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+6],ax
		update_carry
		maj_a
		pop  si
		NEXTI

		align 2

_subr_A_B_s:     exec_opcode2s _subr A B s
_subr_B_C_s:     exec_opcode2s _subr B C s
_subr_C_A_s:     exec_opcode2s _subr C A s
_subr_D_C_s:     exec_opcode2s _subr D C s

_subr_s:
		test_decimal
		js   _subrd_s
_subrh_s:
		mov  al,[bx+7]
		mov  ah,[di+7]
		mov  dh,ah
		and  ax,0f0f0h
		and  dh,00fh
		or   al,dh
		sub  al,ah
		mov  [di+7],al
		update_carry
		NEXTI
_subrd_s:	mov  al,[bx+7]
		mov  ah,[di+7]
		mov  dl,ah      ; copie original destination
		and  ax,0f0f0h  ; masque
		sub  al,ah
		jnc  _subrd_s1
		das
_subrd_s1:	update_carry
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		NEXTI


; dfinitions du reste du code
		build_arith2 _subr



; ----------------------------------------



; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m, w>

_subr_A_B_&CHAMP&:     exec_opcode2 _subr A B CHAMP
_subr_B_C_&CHAMP&:     exec_opcode2 _subr B C CHAMP
_subr_C_A_&CHAMP&:     exec_opcode2 _subr C A CHAMP
_subr_D_C_&CHAMP&:     exec_opcode2 _subr D C CHAMP

		endm






; **************************************************************
; 		add
; **************************************************************

; fonctions d'addition hexa


addh_byte       macro
		mov  al,[si]
		adc  [di],al
		inc  si
		inc  di
		endm

_addh_hinib:
		mov  al,[si]
		and  al,0f0h
		add  [di],al
		ret



_addh_16:       addh_byte
_addh_14:       addh_byte
_addh_12:       addh_byte
_addh_10:       addh_byte
_addh_8:        addh_byte
_addh_6:        addh_byte
_addh_4:        addh_byte
_addh_2:        addh_byte
		update_carry
		ret



_addh_15:       addh_byte
_addh_13:       addh_byte
_addh_11:       addh_byte
_addh_9:        addh_byte
_addh_7:        addh_byte
_addh_5:        addh_byte
_addh_3:        addh_byte
_addh_1:
_addh_lonib:
		mov  al,[di]
		mov  bh,al
		adc  al,[si]
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


; fonctions d'addition dcimal

addd_byte       macro
		mov  al,[di]
		adc  al,[si]
		daa
		mov  [di],al
		inc  si
		inc  di
		endm

_addd_hinib:	;
		mov  al,[di]
		mov  ah,[si]
                mov  dl,al      ; copie original destination
		and  ax,0f0f0h  ; masque
		add  al,ah
		daa
                lahf
		and  dl,0fh
		or   al,dl
		sahf
		mov  [di],al
		ret

_addd_16:       addd_byte
_addd_14:       addd_byte
_addd_12:       addd_byte
_addd_10:       addd_byte
_addd_8:        addd_byte
_addd_6:        addd_byte
_addd_4:        addd_byte
_addd_2:        addd_byte
		update_carry
		ret

_addd_15:       addd_byte
_addd_13:       addd_byte
_addd_11:       addd_byte
_addd_9:        addd_byte
_addd_7:        addd_byte
_addd_5:        addd_byte
_addd_3:        addd_byte
_addd_1:
_addd_lonib:
		mov  al,[di]
		mov  bh,al
		adc  al,[si]
		daa
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret

; ----------------------------------------

		align 2

_add_A_B_a:     exec_opcode2s _add A B a
_add_B_C_a:     exec_opcode2s _add B C a
_add_C_A_a:     exec_opcode2s _add C A a
_add_D_C_a:     exec_opcode2s _add D C a
_add_A_A_a:     exec_opcode2s _add A A a
_add_B_B_a:     exec_opcode2s _add B B a
_add_C_C_a:     exec_opcode2s _add C C a
_add_D_D_a:     exec_opcode2s _add D D a
_add_B_A_a:     exec_opcode2s _add B A a
_add_C_B_a:     exec_opcode2s _add C B a
_add_A_C_a:     exec_opcode2s _add A C a
_add_C_D_a:     exec_opcode2s _add C D a

_add_a:
		test_decimal
		js   _addd_a
_addh_a:        mov  ax,[bx]
		add  [di],ax
		mov  al,[bx-1]
		adc  [di-1],al
		update_halfcarry
		NEXT

_addd_a:        mov  ax,[di]
		mov  dx,[bx]
		add  al,dl
		daa
		xchg al,ah
		adc  al,dh
		daa
		xchg al,ah
		mov  [di],ax
		mov  al,[di-1]
		adc  al,[bx-1]
		daa
		mov  [di-1],al
		update_halfcarry
		NEXT



                align 2

_add_A_B_b:     exec_opcode2s _add A B b
_add_B_C_b:     exec_opcode2s _add B C b
_add_C_A_b:     exec_opcode2s _add C A b
_add_D_C_b:     exec_opcode2s _add D C b
_add_A_A_b:     exec_opcode2s _add A A b
_add_B_B_b:     exec_opcode2s _add B B b
_add_C_C_b:     exec_opcode2s _add C C b
_add_D_D_b:     exec_opcode2s _add D D b
_add_B_A_b:     exec_opcode2s _add B A b
_add_C_B_b:     exec_opcode2s _add C B b
_add_A_C_b:     exec_opcode2s _add A C b
_add_C_D_b:     exec_opcode2s _add C D b


_add_b:
		test_decimal
		js   _addd_b
_addh_b:	mov  al,[bx]
		add  [di],al
		update_carry
		NEXTI
_addd_b:	mov  al,[di]
		add  al,[bx]
		daa
		mov  [di],al
		update_carry
		NEXTI


_add_w:         call maj2_w
		test_decimal
		js   _addd_w
_addh_w:        mov  ax,[si]
		add  [di],ax
		mov  ax,[si+2]
		adc  [di+2],ax
		mov  ax,[si+4]
		adc  [di+4],ax
		mov  ax,[si+6]
		adc  [di+6],ax
		update_carry
		maj_a
		pop  si
		NEXTI

_addd_w:        mov  ax,[di]
		mov  bx,[si]
		add  al,bl
		daa
		xchg al,ah
		adc  al,bh
		daa
		xchg al,ah
		mov  [di],ax
		mov  ax,[di+2]
		mov  bx,[si+2]
		adc  al,bl
		daa
		xchg al,ah
		adc  al,bh
		daa
		xchg al,ah
		mov  [di+2],ax
		mov  ax,[di+4]
		mov  bx,[si+4]
		adc  al,bl
		daa
		xchg al,ah
		adc  al,bh
		daa
		xchg al,ah
		mov  [di+4],ax
		mov  ax,[di+6]
		mov  bx,[si+6]
		adc  al,bl
		daa
		xchg al,ah
		adc  al,bh
		daa
		xchg al,ah
		mov  [di+6],ax
		update_carry
		maj_a
		pop  si
		NEXTI

		align 2
_add_A_B_s:     exec_opcode2s _add A B s
_add_B_C_s:     exec_opcode2s _add B C s
_add_C_A_s:     exec_opcode2s _add C A s
_add_D_C_s:     exec_opcode2s _add D C s
_add_A_A_s:     exec_opcode2s _add A A s
_add_B_B_s:     exec_opcode2s _add B B s
_add_C_C_s:     exec_opcode2s _add C C s
_add_D_D_s:     exec_opcode2s _add D D s
_add_B_A_s:     exec_opcode2s _add B A s
_add_C_B_s:     exec_opcode2s _add C B s
_add_A_C_s:     exec_opcode2s _add A C s
_add_C_D_s:     exec_opcode2s _add C D s

_add_s:
		test_decimal
		js   _addd_s
_addh_s:    	mov  al,[bx+7]
		and  al,0f0h
		add  [di+7],al
		update_carry
		NEXTI
_addd_s:	mov  al,[di+7]
		mov  ah,[bx+7]
		mov  dl,al      ; copie original destination
		and  ax,0f0f0h  ; masque
		add  al,ah
		daa
                update_carry
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		NEXTI




; dfinitions du reste du code
		build_arith2 _add

; -----------------------------------------



; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m, w>

_add_A_B_&CHAMP&:     exec_opcode2 _add A B CHAMP
_add_B_C_&CHAMP&:     exec_opcode2 _add B C CHAMP
_add_C_A_&CHAMP&:     exec_opcode2 _add C A CHAMP
_add_D_C_&CHAMP&:     exec_opcode2 _add D C CHAMP
_add_A_A_&CHAMP&:     exec_opcode2 _add A A CHAMP
_add_B_B_&CHAMP&:     exec_opcode2 _add B B CHAMP
_add_C_C_&CHAMP&:     exec_opcode2 _add C C CHAMP
_add_D_D_&CHAMP&:     exec_opcode2 _add D D CHAMP
_add_B_A_&CHAMP&:     exec_opcode2 _add B A CHAMP
_add_C_B_&CHAMP&:     exec_opcode2 _add C B CHAMP
_add_A_C_&CHAMP&:     exec_opcode2 _add A C CHAMP
_add_C_D_&CHAMP&:     exec_opcode2 _add C D CHAMP

		endm












; ***************************************************************
; 		dfinitions des macros
; ***************************************************************

; -------------------------------------------------------------------
; build_arith
; -------------------------------------------------------------------
; macro de traitement des instructions arithmtiques  1 argument
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT = _inc, _dec, _neg
;          ADJUST_C = stc ou clc selon instruction
; entres: di pointe sur le registre oprande destination
; -------------------------------------------------------------------

build_arith	macro INSTRUCT, ADJUST_C
		local lbl_p1, lbl_p2, lbl_p3, lbl_p8, lbl_p9
		local lbl_wp1, lbl_wp2
		local tab_arith_wp
		local lbl_xs1, lbl_xs2
		local lbl_x1, lbl_x2
		local lbl_s1, lbl_s2
		local lbl_m1, lbl_m2
		local lbl_w1, lbl_w2

&INSTRUCT&_p:
		push  di
		call  maj_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    lbl_p2
		; paire, quartet infrieur
		add   di,ax
		test_decimal
		js    lbl_p1
		ADJUST_C
		call  &INSTRUCT&h_lonib
		jmp   short lbl_p9
lbl_p1:		ADJUST_C
		call  &INSTRUCT&d_lonib            ; !
		jmp   short lbl_p9
lbl_p2:         ; impaire, quartet suprieur
		add   di,ax
		test_decimal
		js    lbl_p3
		call  &INSTRUCT&h_hinib
		jmp   short lbl_p8
lbl_p3:		call  &INSTRUCT&d_hinib             ; !
lbl_p8:		update_carry
lbl_p9:         pop   di
		maj_a
		NEXTI


&INSTRUCT&_wp:
		push  di
		call  maj_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		test_decimal
		jns   lbl_wp1
		add   bx,32
lbl_wp1:	ADJUST_C
		call  cs:[bx+tab_arith_wp]
		pop   di
		maj_a
		NEXTI
tab_arith_wp	dw    &INSTRUCT&h_1,  &INSTRUCT&h_2,  &INSTRUCT&h_3,  &INSTRUCT&h_4
		dw    &INSTRUCT&h_5,  &INSTRUCT&h_6,  &INSTRUCT&h_7,  &INSTRUCT&h_8
		dw    &INSTRUCT&h_9,  &INSTRUCT&h_10, &INSTRUCT&h_11, &INSTRUCT&h_12
		dw    &INSTRUCT&h_13, &INSTRUCT&h_14, &INSTRUCT&h_15, &INSTRUCT&h_16
		dw    &INSTRUCT&d_1,  &INSTRUCT&d_2,  &INSTRUCT&d_3,  &INSTRUCT&d_4
		dw    &INSTRUCT&d_5,  &INSTRUCT&d_6,  &INSTRUCT&d_7,  &INSTRUCT&d_8
		dw    &INSTRUCT&d_9,  &INSTRUCT&d_10, &INSTRUCT&d_11, &INSTRUCT&d_12
		dw    &INSTRUCT&d_13, &INSTRUCT&d_14, &INSTRUCT&d_15, &INSTRUCT&d_16


&INSTRUCT&_xs:
		inc   di
		test_decimal
		js    lbl_xs1
		ADJUST_C
		call  &INSTRUCT&h_lonib
		NEXTI
lbl_xs1:        ADJUST_C
		call  &INSTRUCT&d_lonib
		NEXTI

&INSTRUCT&_x:
		test_decimal
		js    lbl_x1
		ADJUST_C
		call  &INSTRUCT&h_3
		NEXTI
lbl_x1:         ADJUST_C
		call  &INSTRUCT&d_3
		NEXTI

;&INSTRUCT&_s:
;		add   di,7
;		test_decimal
;		js    lbl_s1
;		call  &INSTRUCT&h_hinib
;		update_carry
;		NEXTI
;lbl_s1:         call  &INSTRUCT&d_hinib       ; !
;		update_carry
;		NEXTI

&INSTRUCT&_m:
		push  di
		call  maj_w
		inc   di
		test_decimal
		js    lbl_m1
		call  &INSTRUCT&h_hinib
		inc   di
		call  &INSTRUCT&h_11
		jmp   short lbl_m2
lbl_m1:         call  &INSTRUCT&d_hinib
		inc   di
		call  &INSTRUCT&d_11
lbl_m2:         pop   di
		maj_a
		NEXTI


		endm





; **************************************************************
; 		inc
; **************************************************************

; fonctions d'incrmentation hexa


inch_byte       macro
		adc  byte ptr [di],0
		inc  di
		endm

_inch_hinib:
		add  byte ptr [di],010h
		ret

_inch_16:       inch_byte
_inch_14:       inch_byte
_inch_12:       inch_byte
_inch_10:       inch_byte
_inch_8:        inch_byte
_inch_6:        inch_byte
_inch_4:        inch_byte
_inch_2:        inch_byte
		update_carry
		ret

_inch_15:       inch_byte
_inch_13:       inch_byte
_inch_11:       inch_byte
_inch_9:        inch_byte
_inch_7:        inch_byte
_inch_5:        inch_byte
_inch_3:        inch_byte
_inch_1:
_inch_lonib:
		mov  al,[di]
		mov  bh,al
		adc  al,0
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


; fonctions d'incrmentation dcimal


incd_byte       macro
		mov  al,[di]
		adc  al,0
		daa
		mov  [di],al
		inc  di
		endm

_incd_hinib:
		mov  al,[di]
		mov  dl,al
		and  al,0f0h
		add  al,010h
		daa
		lahf
		and  dl,0fh
		or   al,dl
		sahf
		mov  [di],al
		ret

_incd_16:       incd_byte
_incd_14:       incd_byte
_incd_12:       incd_byte
_incd_10:       incd_byte
_incd_8:        incd_byte
_incd_6:        incd_byte
_incd_4:        incd_byte
_incd_2:        incd_byte
		update_carry
		ret

_incd_15:       incd_byte
_incd_13:       incd_byte
_incd_11:       incd_byte
_incd_9:        incd_byte
_incd_7:        incd_byte
_incd_5:        incd_byte
_incd_3:        incd_byte
_incd_1:
_incd_lonib:
		mov  al,[di]
		mov  bh,al
		adc  al,0
		daa
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret

; ----------------------------------


exec_inc_a      macro REG
                local _incd, _inc2
                align 2
		test_decimal
		js   _incd
                add  word ptr _reg&REG&,1
                jc   _inc2
                clear_carry
		NEXT
_inc2:		adc  byte ptr _reg&REG&-1,0
		update_halfcarry
		NEXT
_incd:          mov  di, offset _reg&REG&
                jmp  short _incd_a
                endm

_inc_A_a:       exec_inc_a A
_inc_B_a:       exec_inc_a B
_inc_C_a:       exec_inc_a C

_incd_a:        mov  ax,[di]
		add  al,1
		daa
		xchg al,ah
		adc  al,0
		daa
		xchg al,ah
		mov  [di],ax
		mov  al,[di-1]
		adc  al,0
		daa
		mov  [di-1],al
		update_halfcarry
		NEXT

_inc_D_a:       exec_inc_a D


exec_inc_b      macro REG
                local _incd, _inc2
                align 2
		test_decimal
		js   _incd
                add  byte ptr _reg&REG&,1
                jc   _inc2
		clear_carry
		NEXTI
_inc2:		set_carry
		NEXTI
_incd:          mov  di, offset _reg&REG&
                jmp  short _incd_b
                endm

_inc_A_b:       exec_inc_b A
_inc_B_b:       exec_inc_b B
_inc_C_b:       exec_inc_b C
_inc_D_b:       exec_inc_b D

_incd_b:	mov  al,[di]
		add  al,1
		daa
		mov  [di],al
		update_carry
		NEXTI

                align 2
_inc_A_w:       exec_opcode1s _inc A w
_inc_B_w:       exec_opcode1s _inc B w
_inc_D_w:       exec_opcode1s _inc D w
_inc_C_w:       mov   di, offset _regC

_inc_w:         maj_ws
		test_decimal
		js   _incd_w
_inch_w:        add  word ptr [di],1
                jc   _inch_w1
                clear_carry
		NEXTI
_inch_w1:	adc  word ptr [di+2],0
		adc  word ptr [di+4],0
		adc  word ptr [di+6],0
         	update_carry
		maj_a
		NEXTI
_incd_w:        mov  ax,[di]
		add  al,1
		daa
		xchg al,ah
		adc  al,0
		daa
		xchg al,ah
		mov  [di],ax
		jnc  _incd_w1            ; oui ?
		mov  ax,[di+2]
		adc  al,0
		daa
		xchg al,ah
		adc  al,0
		daa
		xchg al,ah
		mov  [di+2],ax
		mov  ax,[di+4]
		adc  al,0
		daa
		xchg al,ah
		adc  al,0
		daa
		xchg al,ah
		mov  [di+4],ax
		mov  ax,[di+6]
		adc  al,0
		daa
		xchg al,ah
		adc  al,0
		daa
		xchg al,ah
		mov  [di+6],ax
_incd_w1:	update_carry
		maj_a
		NEXTI



exec_inc_s      macro REG
                local _incd, _inc2
                align 2
		test_decimal
		js   _incd
                add  byte ptr _reg&REG&+7,010h
                jc  _inc2
		clear_carry
		NEXTI
_inc2:		set_carry
		NEXTI
_incd:          mov  di, offset _reg&REG&
                jmp  short _incd_s
                endm


_inc_A_s:       exec_inc_s A
_inc_B_s:       exec_inc_s B
_inc_C_s:       exec_inc_s C
_inc_D_s:       exec_inc_s D

_incd_s:
		mov  al,[di+7]
		mov  dl,al
		and  al,0f0h
		add  al,010h
		daa
		update_carry
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		NEXTI

comment !
                align 2
_inc_A_p:       exec_opcode1s _inc A p
_inc_B_p:       exec_opcode1s _inc B p
_inc_D_p:       exec_opcode1s _inc D p
_inc_C_p:       mov   di, offset _regC

_inc_p: 	test_decimal
		js   _incd_p
                mov  al,_Pointer
                shr  al,1
                jc   _inch_p1
                cmp  al,2
                je   _inch_p4
                cbw
                add  di,ax
		mov  al,[di]
		mov  bh,al
		add  al,1
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		NEXTI
_inch_p1:       cbw
                add  di,ax
                add  byte ptr [di],010h
		update_carry
                NEXTI
_inch_p4:       add  byte ptr [di-1],1
                update_halfcarry
                NEXTI

_incd_p:        mov  al,_Pointer
                shr  al,1
                jc   _incd_p1
                cmp  al,2
                je   _incd_p4
                cbw
                add  di,ax
		mov  al,[di]
		mov  bh,al
		add  al,1
                daa
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		NEXTI
_incd_p1:       cbw
                add  di,ax
		mov  al,[di]
		mov  dl,al
		and  al,0f0h
		add  al,010h
		daa
		update_carry
		and  dl,0fh
		or   al,dl
		mov  [di],al
                NEXTI
_incd_p4:       mov  al,[di-1]
                add  al,1
                daa
                mov  [di-1],al
                update_halfcarry
                NEXTI
!

; dfinitions du reste du code
		build_arith _inc stc

; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m>

_inc_A_&CHAMP&:     exec_opcode _inc A CHAMP
_inc_B_&CHAMP&:     exec_opcode _inc B CHAMP
_inc_C_&CHAMP&:     exec_opcode _inc C CHAMP
_inc_D_&CHAMP&:     exec_opcode _inc D CHAMP

		endm






; -----------------------------------------


; **************************************************************
; 		dec
; **************************************************************

; fonctions de decrmentation hexa


dech_byte       macro
		sbb  byte ptr [di],0
		inc  di
		endm

_dech_hinib:
		sub  byte ptr [di],010h
		ret

_dech_16:       dech_byte
_dech_14:       dech_byte
_dech_12:       dech_byte
_dech_10:       dech_byte
_dech_8:        dech_byte
_dech_6:        dech_byte
_dech_4:        dech_byte
_dech_2:        dech_byte
		update_carry
		ret

_dech_15:       dech_byte
_dech_13:       dech_byte
_dech_11:       dech_byte
_dech_9:        dech_byte
_dech_7:        dech_byte
_dech_5:        dech_byte
_dech_3:        dech_byte
_dech_1:
_dech_lonib:
		mov  al,[di]
		mov  bh,al
		sbb  al,0
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret



; fonctions de decrmentation dcimal


decd_byte       macro
		mov  al,[di]
		sbb  al,0
		das
		mov  [di],al
		inc  di
		endm

_decd_hinib:
		mov  al,[di]
		mov  dl,al
		and  al,0f0h
		sub  al,010h
		jnc  decd_hi1     ; pour comportement avec argument > 9 !
		das
decd_hi1:	lahf
		and  dl,0fh
		or   al,dl
		sahf
		mov  [di],al
		ret

_decd_16:       decd_byte
_decd_14:       decd_byte
_decd_12:       decd_byte
_decd_10:       decd_byte
_decd_8:        decd_byte
_decd_6:        decd_byte
_decd_4:        decd_byte
_decd_2:        decd_byte
		update_carry
		ret

_decd_15:       decd_byte
_decd_13:       decd_byte
_decd_11:       decd_byte
_decd_9:        decd_byte
_decd_7:        decd_byte
_decd_5:        decd_byte
_decd_3:        decd_byte
_decd_1:
_decd_lonib:
		mov  al,[di]
		mov  bh,al
		sbb  al,0
		lahf
		and  ah,16
		je   decd_lo1     ; pour comportement avec argument > 9 !
		sahf
		das
decd_lo1:	sto_carry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


exec_dec_a      macro REG
                local _decd, _dec2
                align 2
		test_decimal
		js   _decd
                sub  word ptr _reg&REG&,1
                jc   _dec2
		clear_carry
		NEXT
_dec2:		sbb  byte ptr _reg&REG&-1,0
		update_halfcarry
		NEXT
_decd:          mov  di, offset _reg&REG&
                jmp  short _decd_a
                endm


_dec_A_a:       exec_dec_a A
_dec_B_a:       exec_dec_a B
_dec_C_a:       exec_dec_a C

_decd_a:        mov  ax,[di]
		sub  al,1
		das
		xchg al,ah
		sbb  al,0
		das
		xchg al,ah
		mov  [di],ax
		mov  al,[di-1]
		sbb  al,0
		das
		mov  [di-1],al
		update_halfcarry
		NEXT

_dec_D_a:       exec_dec_a D



                align 2

exec_dec_b      macro REG
                local _decd, _dec2
                align 2
		test_decimal
		js   _decd
                sub  byte ptr _reg&REG&,1
                jc   _dec2
		clear_carry
		NEXTI
_dec2:		set_carry
		NEXTI
_decd:          mov  di, offset _reg&REG&
                jmp  short _decd_b
                endm

_dec_A_b:       exec_dec_b A
_dec_B_b:       exec_dec_b B
_dec_C_b:       exec_dec_b C
_dec_D_b:       exec_dec_b D

_decd_b:	mov  al,[di]
                cmp  al,0a0h     ; pour comportement avec digits>9!
                jnc  _dech_b     ; JFG 02/03/03 - added back 18/07/07 !
		sub  al,1
		das
		mov  [di],al
		update_carry
		NEXTI
_dech_b:	sub  byte ptr [di],1
		update_carry
		NEXTI


                align 2
_dec_A_w:       exec_opcode1s _dec A w
_dec_B_w:       exec_opcode1s _dec B w
_dec_D_w:       exec_opcode1s _dec D w
_dec_C_w:       mov   di, offset _regC

_dec_w:         maj_ws
		test_decimal
		js   _decd_w
_dech_w:        sub  word ptr [di],1
                jc   _dech_w1
         	clear_carry
		NEXTI
_dech_w1:       sbb  word ptr [di+2],0
		sbb  word ptr [di+4],0
		sbb  word ptr [di+6],0
         	update_carry
		maj_a
		NEXTI
_decd_w:        mov  ax,[di]
		sub  al,1
		das
		xchg al,ah
		sbb  al,0
		das
		xchg al,ah
		mov  [di],ax
		jnc  _decd_w1		; oui ?
		mov  ax,[di+2]
		sbb  al,0
		das
		xchg al,ah
		sbb  al,0
		das
		xchg al,ah
		mov  [di+2],ax
		mov  ax,[di+4]
		sbb  al,0
		das
		xchg al,ah
		sbb  al,0
		das
		xchg al,ah
		mov  [di+4],ax
		mov  ax,[di+6]
		sbb  al,0
		das
		xchg al,ah
		sbb  al,0
		das
		xchg al,ah
		mov  [di+6],ax
_decd_w1:	update_carry
		maj_a
		NEXTI


exec_dec_s      macro REG
                local _decd, _dec2
                align 2
		test_decimal
		js   _decd
                sub  byte ptr _reg&REG&+7,010h
                jc   _dec2
		clear_carry
		NEXTI
_dec2:		set_carry
		NEXTI
_decd:          mov  di, offset _reg&REG&
                jmp  short _decd_s
                endm

_dec_A_s:       exec_dec_s A
_dec_B_s:       exec_dec_s B
_dec_C_s:       exec_dec_s C
_dec_D_s:       exec_dec_s D

_decd_s:	mov  al,[di+7]
		mov  dl,al
		and  al,0f0h
		sub  al,010h
		jnc  _decd_s1     ; pour comportement avec argument > 9 !
		das
_decd_s1:       update_carry
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		NEXTI


comment !
                align 2
_dec_A_p:       exec_opcode1s _dec A p
_dec_B_p:       exec_opcode1s _dec B p
_dec_D_p:       exec_opcode1s _dec D p
_dec_C_p:       mov   di, offset _regC

_dec_p: 	test_decimal
		js   _decd_p
                mov  al,_Pointer
                shr  al,1
                jc   _dech_p1
                cmp  al,2
                je   _dech_p4
                cbw
                add  di,ax
		mov  al,[di]
		mov  bh,al
		sub  al,1
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		NEXTI
_dech_p1:       cbw
                add  di,ax
                sub  byte ptr [di],010h
		update_carry
                NEXTI
_dech_p4:       sub  byte ptr [di-1],1
                update_halfcarry
                NEXTI

_decd_p:        mov  al,_Pointer
                shr  al,1
                jc   _decd_p1
                cmp  al,2
                je   _decd_p4
                cbw
                add  di,ax
		mov  al,[di]
		mov  bh,al
		sub  al,1
		lahf
		and  ah,16
		je   _decd_p2     ; pour comportement avec argument > 9
		sahf
		das
_decd_p2:	sto_carry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
        	NEXTI
_decd_p1:       cbw
                add  di,ax
		mov  al,[di]
		mov  dl,al
		and  al,0f0h
		sub  al,010h
		jnc  _decd_p3     ; pour comportement avec argument > 9
		das
_decd_p3:	update_carry
		and  dl,0fh
		or   al,dl
		mov  [di],al
                NEXTI
_decd_p4:       mov  al,[di-1]
                sub  al,1
		lahf
		and  ah,16
		je   decd_p5     ; pour comportement avec argument > 9
		sahf
                das
_decd_p5:	sto_carry
                mov  [di-1],al
                NEXTI
!




; dfinitions du reste du code
		build_arith _dec stc

; -----------------------------------------








; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m>

_dec_A_&CHAMP&:     exec_opcode _dec A CHAMP
_dec_B_&CHAMP&:     exec_opcode _dec B CHAMP
_dec_C_&CHAMP&:     exec_opcode _dec C CHAMP
_dec_D_&CHAMP&:     exec_opcode _dec D CHAMP

		endm






; **************************************************************
; 		neg
; dest = 0-dest
; **************************************************************

; fonctions de complmentation  2 hexa


negh_byte       macro
		mov  al,0
		sbb  al,[di]
		mov  [di],al
		inc  di
		endm

_negh_hinib:	;
		mov  al,[di]
		mov  ah,al
		and  ax,0f00fh
		sub  al,ah
		mov  [di],al
		ret


_negh_16:       negh_byte
_negh_14:       negh_byte
_negh_12:       negh_byte
_negh_10:       negh_byte
_negh_8:        negh_byte
_negh_6:        negh_byte
_negh_4:        negh_byte
_negh_2:        negh_byte
		update_carry
		ret



_negh_15:       negh_byte
_negh_13:       negh_byte
_negh_11:       negh_byte
_negh_9:        negh_byte
_negh_7:        negh_byte
_negh_5:        negh_byte
_negh_3:        negh_byte
_negh_1:
_negh_lonib:
		mov  al,0
		mov  dh,[di]
		sbb  al,dh
		update_halfcarry
		mov  ah,dh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


; fonctions de complmentation  2 dcimal


negd_byte       macro
		mov  al,0
		sbb  al,[di]
		das
		mov  [di],al
		inc  di
		endm

_negd_hinib:
		mov  al,0
		mov  ah,[di]
		mov  dl,ah      ; copie original destination
		and  ah,0f0h  ; masque
		sub  al,ah
		das
		lahf
		and  dl,0fh
		or   al,dl
		sahf
		mov  [di],al
		ret

_negd_16:       negd_byte
_negd_14:       negd_byte
_negd_12:       negd_byte
_negd_10:       negd_byte
_negd_8:        negd_byte
_negd_6:        negd_byte
_negd_4:        negd_byte
_negd_2:        negd_byte
		update_carry
		ret

_negd_15:       negd_byte
_negd_13:       negd_byte
_negd_11:       negd_byte
_negd_9:        negd_byte
_negd_7:        negd_byte
_negd_5:        negd_byte
_negd_3:        negd_byte
_negd_1:
_negd_lonib:
		mov  dh,[di]
		mov  al,0
		sbb  al,dh
		das
		update_halfcarry
		mov  ah,dh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


		align 2

_neg_A_a:     exec_opcode1s _neg A a
_neg_B_a:     exec_opcode1s _neg B a
_neg_D_a:     exec_opcode1s _neg D a
_neg_C_a:     mov   di, offset _regC

_neg_a:
		test_decimal
		js   _negd_a
_negh_a:        xor  ax,ax
		sub  ax,[di]		; neg word ptr [di] ?
		mov  [di],ax
		mov  al,0
		sbb  al,[di-1]
		mov  [di-1],al
		update_halfcarry
		NEXT
_negd_a:        xor  ax,ax
		mov  bx,[di]
		sub  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di],ax
		mov  al,0
		sbb  al,[di-1]
		das
		mov  [di-1],al
		update_halfcarry
		NEXT

		align 2
_neg_A_b:       exec_opcode1s _neg A b
_neg_B_b:       exec_opcode1s _neg B b
_neg_D_b:       exec_opcode1s _neg D b
_neg_C_b:       mov   di, offset _regC

_neg_b:
		test_decimal
		js   _negd_b
_negh_b:	xor  al,al           ; neg  byte ptr [di] ?
		sub  al,[di]
		mov  [di],al
		update_carry
		NEXTI
_negd_b:	xor  al,al
		sub  al,[di]
		das
		mov  [di],al
		update_carry
		NEXTI


		align 2
_neg_A_w:       exec_opcode1s _neg A w
_neg_B_w:       exec_opcode1s _neg B w
_neg_D_w:       exec_opcode1s _neg D w
_neg_C_w:       mov   di, offset _regC

_neg_w:         maj_ws
		test_decimal
		js   _negd_w
_negh_w:        xor  ax,ax
		sub  ax,[di]
		mov  [di],ax
		mov  ax,0
		sbb  ax,[di+2]
		mov  [di+2],ax
		mov  ax,0
		sbb  ax,[di+4]
		mov  [di+4],ax
		mov  ax,0
		sbb  ax,[di+6]
		mov  [di+6],ax
		update_carry
		maj_a
		NEXTI
_negd_w:        xor  ax,ax
		mov  bx,[di]
		sub  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di],ax
		mov  ax,0
		mov  bx,[di+2]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+2],ax
		mov  ax,0
		mov  bx,[di+4]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+4],ax
		mov  ax,0
		mov  bx,[di+6]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+6],ax
		update_carry
		maj_a
		NEXTI


		align 2
_neg_A_s:       exec_opcode1s _neg A s
_neg_B_s:       exec_opcode1s _neg B s
_neg_D_s:       exec_opcode1s _neg D s
_neg_C_s:       mov   di, offset _regC

_neg_s:
		test_decimal
		js   _negd_s
_negh_s:	mov  al,[di+7]
		mov  ah,al
		and  ax,0f00fh
		sub  al,ah
		mov  [di+7],al
		update_carry
		NEXTI
_negd_s:
		mov  al,0
		mov  ah,[di+7]
		mov  dl,ah      ; copie original destination
		and  ah,0f0h  ; masque
		sub  al,ah
		das
		update_carry
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		NEXTI


comment !
                align 2
_neg_A_p:       exec_opcode1s _neg A p
_neg_B_p:       exec_opcode1s _neg B p
_neg_D_p:       exec_opcode1s _neg D p
_neg_C_p:       mov   di, offset _regC

_neg_p: 	test_decimal
		js   _negd_p
                mov  al,_Pointer
                shr  al,1
                jc   _negh_p1
                cmp  al,2
                je   _negh_p4
                cbw
                add  di,ax
		mov  al,[di]
		mov  bh,al

		sub  al,1
		update_halfcarry


		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		NEXTI
_negh_p1:       cbw
                add  di,ax
                sub  byte ptr [di],010h
		update_carry
                NEXTI
_negh_p4:       sub  byte ptr [di-1],1
                update_halfcarry
                NEXTI

_negd_p:        mov  al,_Pointer
                shr  al,1
                jc   _negd_p1
                cmp  al,2
                je   _negd_p4
                cbw
                add  di,ax
		mov  al,[di]
		mov  bh,al
		sub  al,1
		lahf
		and  ah,16
		je   _negd_p2     ; pour comportement avec argument > 9
		sahf
		das
_negd_p2:	sto_carry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
        	NEXTI
_negd_p1:       cbw
                add  di,ax
		mov  al,[di]
		mov  dl,al
		and  al,0f0h
		sub  al,010h
		jnc  _negd_p3     ; pour comportement avec argument > 9
		das
_negd_p3:	update_carry
		and  dl,0fh
		or   al,dl
		mov  [di],al
                NEXTI
_negd_p4:       mov  al,[di-1]
                sub  al,1
		lahf
		and  ah,16
		je   negd_p5     ; pour comportement avec argument > 9
		sahf
                das
_negd_p5:	sto_carry
                mov  [di-1],al
                NEXTI
!




; dfinitions du reste du code
		build_arith _neg clc


; -----------------------------------------




; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m>

_neg_A_&CHAMP&:     exec_opcode _neg A CHAMP
_neg_B_&CHAMP&:     exec_opcode _neg B CHAMP
_neg_C_&CHAMP&:     exec_opcode _neg C CHAMP
_neg_D_&CHAMP&:     exec_opcode _neg D CHAMP

		endm





; **************************************************************
; 		not
; dest = -1-dest
; **************************************************************

; fonctions de complmentation  1



; -------------------------------------------------------------------
; build_not
; -------------------------------------------------------------------
; macro de traitement des instructions arithmtiques  1 argument
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT = _not
; entres: di pointe sur le registre oprande destination
; -------------------------------------------------------------------

build_not	macro INSTRUCT
		local lbl_p1, lbl_p2, lbl_p3, lbl_p8, lbl_p9
		local lbl_wp1, lbl_wp2
		local tab_arith_wp
		local lbl_xs1, lbl_xs2
		local lbl_x1, lbl_x2
		local lbl_s1, lbl_s2
		local lbl_m1, lbl_m2
		local lbl_w1, lbl_w2

&INSTRUCT&_p:
		push  di
		call  maj_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    lbl_p2
		; paire, quartet infrieur
		add   di,ax
		test_decimal
		js    lbl_p1
		call  &INSTRUCT&h_lonib
		jmp   short lbl_p9
lbl_p1:         call  &INSTRUCT&d_lonib
		jmp   short lbl_p9
lbl_p2:         ; impaire, quartet suprieur
		add   di,ax
		test_decimal
		js    lbl_p3
		call  &INSTRUCT&h_hinib
		jmp   short lbl_p8
lbl_p3:         call  &INSTRUCT&d_hinib
lbl_p8:		clear_carry
lbl_p9:         pop   di
		maj_a
		NEXTI


&INSTRUCT&_wp:
		push  di
		call  maj_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		test_decimal
		jns   lbl_wp1
		add   bx,32
lbl_wp1:	call  cs:[bx+tab_arith_wp]
		pop   di
		maj_a
		NEXTI
tab_arith_wp	dw    &INSTRUCT&h_1,  &INSTRUCT&h_2,  &INSTRUCT&h_3,  &INSTRUCT&h_4
		dw    &INSTRUCT&h_5,  &INSTRUCT&h_6,  &INSTRUCT&h_7,  &INSTRUCT&h_8
		dw    &INSTRUCT&h_9,  &INSTRUCT&h_10, &INSTRUCT&h_11, &INSTRUCT&h_12
		dw    &INSTRUCT&h_13, &INSTRUCT&h_14, &INSTRUCT&h_15, &INSTRUCT&h_16
		dw    &INSTRUCT&d_1,  &INSTRUCT&d_2,  &INSTRUCT&d_3,  &INSTRUCT&d_4
		dw    &INSTRUCT&d_5,  &INSTRUCT&d_6,  &INSTRUCT&d_7,  &INSTRUCT&d_8
		dw    &INSTRUCT&d_9,  &INSTRUCT&d_10, &INSTRUCT&d_11, &INSTRUCT&d_12
		dw    &INSTRUCT&d_13, &INSTRUCT&d_14, &INSTRUCT&d_15, &INSTRUCT&d_16


&INSTRUCT&_xs:
		inc   di
		test_decimal
		js    lbl_xs1
		call  &INSTRUCT&h_lonib
		NEXTI
lbl_xs1:        call  &INSTRUCT&d_lonib
		NEXTI

&INSTRUCT&_x:
		test_decimal
		js    lbl_x1
		call  &INSTRUCT&h_3
		NEXTI
lbl_x1:		call  &INSTRUCT&d_3
		NEXTI

comment !
&INSTRUCT&_s:
		add   di,7
		test_decimal
		js    lbl_s1
		call  &INSTRUCT&h_hinib
		clear_carry
		NEXTI
lbl_s1:         call  &INSTRUCT&d_hinib
		clear_carry
		NEXTI
!

&INSTRUCT&_m:
		push  di
		call  maj_w
		inc   di
		test_decimal
		js    lbl_m1
		call  &INSTRUCT&h_hinib
		inc   di
		call  &INSTRUCT&h_11
		jmp   short lbl_m2
lbl_m1:         stc
		call  &INSTRUCT&d_hinib
		inc   di
		call  &INSTRUCT&d_11
lbl_m2:         pop   di
		maj_a
		NEXTI

		endm



noth_byte       macro
		not  byte ptr [di]
		inc  di
		endm

_noth_hinib:	xor  byte ptr [di],0f0h
		ret

_noth_16:       noth_byte
_noth_14:       noth_byte
_noth_12:       noth_byte
_noth_10:       noth_byte
_noth_8:        noth_byte
_noth_6:        noth_byte
_noth_4:        noth_byte
_noth_2:        noth_byte
		clear_carry
		ret

_noth_15:       noth_byte
_noth_13:       noth_byte
_noth_11:       noth_byte
_noth_9:        noth_byte
_noth_7:        noth_byte
_noth_5:        noth_byte
_noth_3:        noth_byte
_noth_1:
_noth_lonib:	xor  byte ptr [di],00fh
		clear_carry
		ret


; fonctions de complmentation  1 dcimal


notd_byte       macro
		mov  al,099h
		sub  al,[di]
		das
		mov  [di],al
		inc  di
		endm

_notd_hinib:
		mov  al,090h
		mov  ah,[di]
		mov  dl,ah      ; copie original destination
		and  ah,0f0h  ; masque
		sub  al,ah
		das
		and  dl,0fh
		or   al,dl
		mov  [di],al
		ret

_notd_16:       notd_byte
_notd_14:       notd_byte
_notd_12:       notd_byte
_notd_10:       notd_byte
_notd_8:        notd_byte
_notd_6:        notd_byte
_notd_4:        notd_byte
_notd_2:        notd_byte
		clear_carry
		ret

_notd_15:       notd_byte
_notd_13:       notd_byte
_notd_11:       notd_byte
_notd_9:        notd_byte
_notd_7:        notd_byte
_notd_5:        notd_byte
_notd_3:        notd_byte
_notd_1:
_notd_lonib:	mov  dh,[di]
		mov  al,009h
		sub  al,dh
		das
		clear_carry
		mov  ah,dh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


		align 2

_not_A_a:       exec_opcode1s _not A a
_not_B_a:       exec_opcode1s _not B a
_not_D_a:       exec_opcode1s _not D a
_not_C_a:       mov   di, offset _regC

_not_a:
		test_decimal
		js   _notd_a
		not  word ptr [di]
		not  byte ptr [di-1]
		clear_carry
		NEXT
_notd_a:        xor  ax,ax
		stc
		mov  bx,[di]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di],ax
		mov  al,0
		sbb  al,[di-1]
		das
		mov  [di-1],al
		clear_carry
		NEXT


		align 2
_not_A_b:       exec_opcode1s _not A b
_not_B_b:       exec_opcode1s _not B b
_not_D_b:       exec_opcode1s _not D b
_not_C_b:       mov   di, offset _regC

_not_b:
		test_decimal
		js   _notd_b
		not  byte ptr [di]
		clear_carry
		NEXTI
_notd_b:        xor  ax,ax
		stc
		sbb  al,[di]
		das
		mov  [di],al
		clear_carry
		NEXTI


		align 2
_not_A_w:       exec_opcode1s _not A w
_not_B_w:       exec_opcode1s _not B w
_not_D_w:       exec_opcode1s _not D w
_not_C_w:       mov   di, offset _regC

_not_w:         maj_ws
		test_decimal
		js   _notd_w
		not  word ptr [di]
		not  word ptr [di+2]
		not  word ptr [di+4]
		not  word ptr [di+6]
		clear_carry
		maj_a
		NEXTI
_notd_w:        xor  ax,ax
		stc
		mov  bx,[di]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di],ax
		mov  ax,0
		mov  bx,[di+2]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+2],ax
		mov  ax,0
		mov  bx,[di+4]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+4],ax
		mov  ax,0
		mov  bx,[di+6]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+6],ax
		clear_carry
		maj_a
		NEXTI


		align 2
_not_A_s:       exec_opcode1s _not A s
_not_B_s:       exec_opcode1s _not B s
_not_D_s:       exec_opcode1s _not D s
_not_C_s:       mov   di, offset _regC

_not_s:
		test_decimal
		js    _notd_s
            	xor  byte ptr [di+7],0f0h
		clear_carry
		NEXTI
_notd_s:
		mov  al,090h
		mov  ah,[di+7]
		mov  dl,ah      ; copie original destination
		and  ah,0f0h  ; masque
		sub  al,ah
		das
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		clear_carry
		NEXTI



; dfinitions du reste du code
		build_not _not

; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m>

_not_A_&CHAMP&:     exec_opcode _not A CHAMP
_not_B_&CHAMP&:     exec_opcode _not B CHAMP
_not_C_&CHAMP&:     exec_opcode _not C CHAMP
_not_D_&CHAMP&:     exec_opcode _not D CHAMP

		endm


; ********************************************************************
;               mulation du processeur Saturn par le 8086
;                       J-F Garnier 1993,95
; --------------------------------------------------------------------
; Module sat_cmp.asm : mulation des instructions de comparaison
; --------------------------------------------------------------------
; Version 1 16/03/93
; Version 1.1 04/11/95 optimisation routage code
;   possibilit d'optimisation des instr. champ a, b, (w)  faire
; ********************************************************************


		; mulation du processeur Saturn par le 8086


; ==================================================================
; dfinitions des instructions de comparaison:
; ?x=y fs
; ?x#y fs
; ?x>y fs
; ?x>=y fs
; ?x<y fs
; ?x<=y fs
; ?x=0 fs
; ?x#0 fs
; ==================================================================





; ******************************************************************
; fonctions de comparaisons
; ******************************************************************


_cmp_p:		call  maj2_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    _cmp2_p_i
		; paire, quartet infrieur
		add   si,ax
		add   di,ax
		jmp   _cmp_lonib
_cmp2_p_i:      ; impaire, quartet suprieur
		add   si,ax
		add   di,ax
		jmp  _cmp_hinib

_cmp_wp:        call  maj2_w
		xor   ax,ax
		mov   al,_Pointer
		mov   bx,ax
		shr   al,1
		jc    _cmp_wp1
		add   si,ax
		add   di,ax
		call  _cmp_lonib
		jne   _cmp_wp9
		dec   si
		dec   di
		jmp   short _cmp_wp2
_cmp_wp1:	add   si,ax
		add   di,ax
		inc   bx
_cmp_wp2:	jmp   cs:[bx+tab_cmp_wp]
_cmp_wp9:	ret
tab_cmp_wp      dw    _cmp_0
		dw    _cmp_2,  _cmp_4,  _cmp_6,  _cmp_8
		dw    _cmp_10, _cmp_12, _cmp_14, _cmp_16

_cmp_xs:        inc   si
		inc   di
		jmp   _cmp_lonib

_cmp_x:         inc   si
		inc   di
		call  _cmp_lonib
		jne   _cmp_x1
		dec   si
		dec   di
		jmp   _cmp_2
_cmp_x1:        ret

;_cmp_s:		add   si,7
;		add   di,7
;		jmp   _cmp_hinib

_cmp_m:		call  maj2_w
		add   si,7
		add   di,7
		call  _cmp_lonib
		jne   _cmp_m1
		dec   si
		dec   di
		jmp   _cmp_11
_cmp_m1:        ret

;_cmp_w:
;               call  maj2_w
;               call  _cmp_16
;               ret

cmp_byte        macro
		mov  al,[si]
		cmp  [di],al
		jne  _cmp_end
		dec  si
		dec  di
		endm

_cmp_lonib:	mov  al,[si]
		mov  ah,[di]
		and  ax,0f0fh
		cmp  ah,al
		ret

_cmp_16:        cmp_byte
_cmp_14:        cmp_byte
_cmp_12:        cmp_byte
_cmp_10:        cmp_byte
_cmp_8:         cmp_byte
_cmp_6:         cmp_byte
_cmp_4:         cmp_byte
_cmp_2:         mov  al,[si]
		cmp  [di],al
_cmp_end:       ret
_cmp_0:		xor  ax,ax   ; Z=1
		ret


_cmp_15:        cmp_byte
_cmp_13:        cmp_byte
_cmp_11:        cmp_byte
_cmp_9:         cmp_byte
_cmp_7:         cmp_byte
_cmp_5:         cmp_byte
_cmp_3:         cmp_byte
_cmp_1:
_cmp_hinib:     mov  al,[si]
		mov  ah,[di]
		and  ax,0f0f0h
		cmp  ah,al
		ret

_cmp_a          macro
                local _cmp_a1
                mov  al,[bx-1]
		mov  ah,[di-1]
		and  ax,0f0fh
		cmp  ah,al
		jne  _cmp_a1
		mov  ax,[bx]
		cmp  [di],ax
_cmp_a1:
                endm

_cmp_b          macro
                mov  al,[bx]
		cmp  [di],al
		endm

_cmp_w:         mov  ax,[si+6]
		cmp  [di+6],ax
		jne  _cmp_w1
		mov  ax,[si+4]
		cmp  [di+4],ax
		jne  _cmp_w1
		call maj2_w        ; !!
		mov  ax,[si+2]
		cmp  [di+2],ax
		jne  _cmp_w1
		mov  ax,[si]
		cmp  [di],ax
_cmp_w1:        ret

_cmp_s          macro
		mov  al,[bx+7]
		mov  ah,[di+7]
		and  ax,0f0f0h
		cmp  ah,al
		endm


; ==============================================================

		align 2
condition_vrai:
		set_carry
                incr_pc
		mov   ax,es:[si]
		sar   ax,cl
		or    al,al
		je    cond_vrai2
		cbw
		jmp   dojump
cond_vrai2:     jmp   _rtn

; ==============================================================


; code d'excution d'instructions de comparaison de 2 arguments


		align 2
_eq_A_B_a:      exec_opcode2s _eq2 A B a
_eq_B_C_a:      exec_opcode2s _eq2 B C a
_eq_C_A_a:      exec_opcode2s _eq2 C A a
_eq_D_C_a:      exec_opcode2s _eq2 D C a

_eq2_a:         _cmp_a
		jne   faux_eq2
		jmp   condition_vrai

		align 2
_eq_A_B_b:      exec_opcode2s _eq2 A B b
_eq_B_C_b:      exec_opcode2s _eq2 B C b
_eq_C_A_b:      exec_opcode2s _eq2 C A b
_eq_D_C_b:      exec_opcode2s _eq2 D C b

_eq2_b:         _cmp_b
		jne   faux_eq2
		jmp   condition_vrai

		align 2
_eq_A_B_s:      exec_opcode2s _eq2 A B s
_eq_B_C_s:      exec_opcode2s _eq2 B C s
_eq_C_A_s:      exec_opcode2s _eq2 C A s
_eq_D_C_s:      exec_opcode2s _eq2 D C s

_eq2_s:         _cmp_s
		jne   faux_eq2
		jmp   condition_vrai

faux_eq2:       ; condition_fausse
		clear_carry
		inc   si
		NEXTI


		irp   CHAMP,<p, wp, xs, x, m, w>
		local faux
_eq2_&CHAMP&:   call  _cmp_&CHAMP&
		pop   si
		jne   faux_eq2
		jmp   condition_vrai
		endm


                align 2
_ne_A_B_a:      exec_opcode2s _ne2 A B a
_ne_B_C_a:      exec_opcode2s _ne2 B C a
_ne_C_A_a:      exec_opcode2s _ne2 C A a
_ne_D_C_a:      exec_opcode2s _ne2 D C a

_ne2_a:         _cmp_a
		je    faux_ne2
		jmp   condition_vrai

                align 2
_ne_A_B_b:      exec_opcode2s _ne2 A B b
_ne_B_C_b:      exec_opcode2s _ne2 B C b
_ne_C_A_b:      exec_opcode2s _ne2 C A b
_ne_D_C_b:      exec_opcode2s _ne2 D C b

_ne2_b:         _cmp_b
		je    faux_ne2
		jmp   condition_vrai

                align 2
_ne_A_B_s:      exec_opcode2s _ne2 A B s
_ne_B_C_s:      exec_opcode2s _ne2 B C s
_ne_C_A_s:      exec_opcode2s _ne2 C A s
_ne_D_C_s:      exec_opcode2s _ne2 D C s

_ne2_s:         _cmp_s
		je    faux_ne2
		jmp   condition_vrai

faux_ne2:       ; condition_fausse
		clear_carry
		inc   si
		NEXTI


		irp   CHAMP,<p, wp, xs, x, m, w>
		local faux
_ne2_&CHAMP&:   call  _cmp_&CHAMP&
		pop   si
		je    faux_ne2
		jmp   condition_vrai
		endm


                align 2
_gt_A_B_a:      exec_opcode2s _gt2 A B a
_gt_B_C_a:      exec_opcode2s _gt2 B C a
_gt_C_A_a:      exec_opcode2s _gt2 C A a
_gt_D_C_a:      exec_opcode2s _gt2 D C a
_gt_B_A_a:      exec_opcode2s _gt2 B A a
_gt_C_B_a:      exec_opcode2s _gt2 C B a
_gt_A_C_a:      exec_opcode2s _gt2 A C a
_gt_C_D_a:      exec_opcode2s _gt2 C D a

_gt2_a:         _cmp_a
		jbe   faux_gt2a
		jmp   condition_vrai
faux_gt2a:       ; condition_fausse
		clear_carry
		inc   si
		NEXTI

                align 2
_gt_A_B_b:      exec_opcode2s _gt2 A B b
_gt_B_C_b:      exec_opcode2s _gt2 B C b
_gt_C_A_b:      exec_opcode2s _gt2 C A b
_gt_D_C_b:      exec_opcode2s _gt2 D C b
_gt_B_A_b:      exec_opcode2s _gt2 B A b
_gt_C_B_b:      exec_opcode2s _gt2 C B b
_gt_A_C_b:      exec_opcode2s _gt2 A C b
_gt_C_D_b:      exec_opcode2s _gt2 C D b

_gt2_b:         _cmp_b
		jbe   faux_gt2b
		jmp   condition_vrai
faux_gt2b:       ; condition_fausse
		clear_carry
		inc   si
		NEXTI

                align 2
_gt_A_B_s:      exec_opcode2s _gt2 A B s
_gt_B_C_s:      exec_opcode2s _gt2 B C s
_gt_C_A_s:      exec_opcode2s _gt2 C A s
_gt_D_C_s:      exec_opcode2s _gt2 D C s
_gt_B_A_s:      exec_opcode2s _gt2 B A s
_gt_C_B_s:      exec_opcode2s _gt2 C B s
_gt_A_C_s:      exec_opcode2s _gt2 A C s
_gt_C_D_s:      exec_opcode2s _gt2 C D s

_gt2_s:         _cmp_s
		jbe   faux_gt2s
		jmp   condition_vrai
faux_gt2s:       ; condition_fausse
		clear_carry
		inc   si
		NEXTI

		irp   CHAMP,<p, wp, xs, x, m, w>
		local faux
_gt2_&CHAMP&:   call  _cmp_&CHAMP&
		pop   si
		jbe   faux_gt2
		jmp   condition_vrai
		endm
faux_gt2:       ; condition_fausse
		clear_carry
		inc   si
		NEXTI


                align 2
_ge_A_B_a:      exec_opcode2s _ge2 A B a
_ge_B_C_a:      exec_opcode2s _ge2 B C a
_ge_C_A_a:      exec_opcode2s _ge2 C A a
_ge_D_C_a:      exec_opcode2s _ge2 D C a
_ge_B_A_a:      exec_opcode2s _ge2 B A a
_ge_C_B_a:      exec_opcode2s _ge2 C B a
_ge_A_C_a:      exec_opcode2s _ge2 A C a
_ge_C_D_a:      exec_opcode2s _ge2 C D a

_ge2_a:         _cmp_a
		jb    faux_ge2a
		jmp   condition_vrai
faux_ge2a:       ; condition_fausse
		clear_carry
		inc   si
		NEXTI

                align 2
_ge_A_B_b:      exec_opcode2s _ge2 A B b
_ge_B_C_b:      exec_opcode2s _ge2 B C b
_ge_C_A_b:      exec_opcode2s _ge2 C A b
_ge_D_C_b:      exec_opcode2s _ge2 D C b
_ge_B_A_b:      exec_opcode2s _ge2 B A b
_ge_C_B_b:      exec_opcode2s _ge2 C B b
_ge_A_C_b:      exec_opcode2s _ge2 A C b
_ge_C_D_b:      exec_opcode2s _ge2 C D b

_ge2_b:         _cmp_b
		jb    faux_ge2b
		jmp   condition_vrai
faux_ge2b:       ; condition_fausse
		clear_carry
		inc   si
		NEXTI

                align 2
_ge_A_B_s:      exec_opcode2s _ge2 A B s
_ge_B_C_s:      exec_opcode2s _ge2 B C s
_ge_C_A_s:      exec_opcode2s _ge2 C A s
_ge_D_C_s:      exec_opcode2s _ge2 D C s
_ge_B_A_s:      exec_opcode2s _ge2 B A s
_ge_C_B_s:      exec_opcode2s _ge2 C B s
_ge_A_C_s:      exec_opcode2s _ge2 A C s
_ge_C_D_s:      exec_opcode2s _ge2 C D s

_ge2_s:         _cmp_s
		jb    faux_ge2s
		jmp   condition_vrai
faux_ge2s:       ; condition_fausse
		clear_carry
		inc   si
		NEXTI


		irp   CHAMP,<p, wp, xs, x, m, w>
		local faux
_ge2_&CHAMP&:   call  _cmp_&CHAMP&
		pop   si
		jb    faux_ge2
		jmp   condition_vrai
		endm
faux_ge2:       ; condition_fausse
		clear_carry
		inc   si
		NEXTI







; **************************************************************
;               eq2
; **************************************************************
; fonctions de comparaison


; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m, w>

_eq_A_B_&CHAMP&:     exec_opcode2 _eq2 A B CHAMP
_eq_B_C_&CHAMP&:     exec_opcode2 _eq2 B C CHAMP
_eq_C_A_&CHAMP&:     exec_opcode2 _eq2 C A CHAMP
_eq_D_C_&CHAMP&:     exec_opcode2 _eq2 D C CHAMP

		endm


; **************************************************************
;               ne2
; **************************************************************
; fonctions de comparaison


; dfinitions des points d'entres
		irp   CHAMP,<p, wp, xs, x, m, w>

_ne_A_B_&CHAMP&:     exec_opcode2 _ne2 A B CHAMP
_ne_B_C_&CHAMP&:     exec_opcode2 _ne2 B C CHAMP
_ne_C_A_&CHAMP&:     exec_opcode2 _ne2 C A CHAMP
_ne_D_C_&CHAMP&:     exec_opcode2 _ne2 D C CHAMP

		endm



; **************************************************************
;               gt2
; **************************************************************
; fonctions de comparaison


; dfinitions des points d'entres
		irp   CHAMP,<p, wp, xs, x, m, w>

_gt_A_B_&CHAMP&:     exec_opcode2 _gt2 A B CHAMP
_gt_B_C_&CHAMP&:     exec_opcode2 _gt2 B C CHAMP
_gt_C_A_&CHAMP&:     exec_opcode2 _gt2 C A CHAMP
_gt_D_C_&CHAMP&:     exec_opcode2 _gt2 D C CHAMP
_gt_B_A_&CHAMP&:     exec_opcode2 _gt2 B A CHAMP
_gt_C_B_&CHAMP&:     exec_opcode2 _gt2 C B CHAMP
_gt_A_C_&CHAMP&:     exec_opcode2 _gt2 A C CHAMP
_gt_C_D_&CHAMP&:     exec_opcode2 _gt2 C D CHAMP

		endm


; **************************************************************
;               ge2
; **************************************************************
; fonctions de comparaison


; dfinitions des points d'entres
		irp   CHAMP,<p, wp, xs, x, m, w>

_ge_A_B_&CHAMP&:     exec_opcode2 _ge2 A B CHAMP
_ge_B_C_&CHAMP&:     exec_opcode2 _ge2 B C CHAMP
_ge_C_A_&CHAMP&:     exec_opcode2 _ge2 C A CHAMP
_ge_D_C_&CHAMP&:     exec_opcode2 _ge2 D C CHAMP
_ge_B_A_&CHAMP&:     exec_opcode2 _ge2 B A CHAMP
_ge_C_B_&CHAMP&:     exec_opcode2 _ge2 C B CHAMP
_ge_A_C_&CHAMP&:     exec_opcode2 _ge2 A C CHAMP
_ge_C_D_&CHAMP&:     exec_opcode2 _ge2 C D CHAMP

		endm







; ******************************************************************
; fonctions de test
; ******************************************************************




_tst_p:
		call  maj_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    _tst_p_i
		; paire, quartet infrieur
		add   di,ax
		jmp   _tst_lonib
_tst_p_i:       ; impaire, quartet suprieur
		add   di,ax
		jmp   _tst_hinib

_tst_wp:	call  maj_w
		xor   ax,ax
		mov   al,_Pointer
		mov   bx,ax
		shr   al,1
		jc    _tst_wp1
		add   di,ax
		call  _tst_lonib
		jne   _tst_wp9
		dec   di
		jmp   short _tst_wp2
_tst_wp1:	add   di,ax
		inc   bx
_tst_wp2:	jmp   cs:[bx+tab_tst_wp]
_tst_wp9:	ret
tab_tst_wp      dw    _tst_0
		dw    _tst_2,  _tst_4,  _tst_6,  _tst_8
		dw    _tst_10, _tst_12, _tst_14, _tst_16

_tst_xs:	inc   di
		jmp   _tst_lonib

_tst_x:         inc   di
		call  _tst_lonib
		jne   _tst_x1
		dec   di
		jmp   _tst_2
_tst_x1:        ret

;_tst_s:		add   di,7
;		jmp   _tst_hinib

_tst_m:		call  maj_w
		add   di,7
		call  _tst_lonib
		jne   _tst_m1
		dec   di
		jmp   _tst_11
_tst_m1:        ret

;_tst_w:	call  maj_w
;		add   di,7
;		jmp   _tst_16





tst_byte        macro
		mov  al,[di]
		or   al,al
		jne  _tst_end
		dec  di
		endm

_tst_lonib:	mov  al,[di]
		and  al,0fh
		ret

_tst_16:        tst_byte
_tst_14:        tst_byte
_tst_12:        tst_byte
_tst_10:        tst_byte
_tst_8:         tst_byte
_tst_6:         tst_byte
_tst_4:         tst_byte
_tst_2:         mov  al,[di]
		or   al,al
_tst_end:       ret
_tst_0:		xor  ax,ax   ; Z=1
		ret

_tst_15:        tst_byte
_tst_13:        tst_byte
_tst_11:        tst_byte
_tst_9:         tst_byte
_tst_7:         tst_byte
_tst_5:         tst_byte
_tst_3:         tst_byte
_tst_1:
_tst_hinib:	mov  al,[di]
		and  al,0f0h
		ret


_tst_a          macro
                local _tst_a1
                mov  al,[di-1]
;		and  al,0fh
;		jne  _tst_a1
;		mov  ax,[di]
;		or   ax,ax
		and  ax,0fh        ; changed JFG, 20/07/07 in Arlington Hts, IL
		or   ax,[di]

_tst_a1:
                endm


_tst_b          macro
                mov  al,[di]
		or   al,al
		endm

_tst_w:		mov  ax,[di+6]
		or   ax,ax
		jne  _tst_w1
		mov  ax,[di+4]
		or   ax,ax
		jne  _tst_w1
		call maj_w           ; !!!
		mov  ax,[di+2]
		or   ax,ax
		jne  _tst_w1
		mov  ax,[di]
		or   ax,ax
_tst_w1:	ret

_tst_s          macro
		mov  al,[di+7]
		and  al,0f0h
		endm

                align 2
_eq_A_a:        exec_opcode1s _eq A a
_eq_B_a:        exec_opcode1s _eq B a
_eq_D_a:        exec_opcode1s _eq D a
_eq_C_a:        mov   di, offset _regC

_eq_a:          _tst_a
		jne   faux_eq
		jmp   condition_vrai

                align 2
_eq_A_b:        exec_opcode1s _eq A b
_eq_B_b:        exec_opcode1s _eq B b
_eq_D_b:        exec_opcode1s _eq D b
_eq_C_b:        mov   di, offset _regC

_eq_b:          _tst_b
		jne   faux_eq
		jmp   condition_vrai

                align 2
_eq_A_s:        exec_opcode1s _eq A s
_eq_B_s:        exec_opcode1s _eq B s
_eq_D_s:        exec_opcode1s _eq D s
_eq_C_s:        mov   di, offset _regC

_eq_s:          _tst_s
		jne   faux_eq
		jmp   condition_vrai

faux_eq:        ; condition_fausse
		clear_carry
		inc   si
		NEXTI

		align 2
		irp   CHAMP,<p, wp, xs, x, m, w>
		local  faux
_eq_&CHAMP&:    call  _tst_&CHAMP&
		jne   faux_eq
		jmp   condition_vrai
		endm


                align 2
_ne_A_a:        exec_opcode1s _ne A a
_ne_B_a:        exec_opcode1s _ne B a
_ne_D_a:        exec_opcode1s _ne D a
_ne_C_a:        mov   di, offset _regC

_ne_a:          _tst_a
		je    faux_ne
		jmp   condition_vrai

                align 2
_ne_A_b:        exec_opcode1s _ne A b
_ne_B_b:        exec_opcode1s _ne B b
_ne_D_b:        exec_opcode1s _ne D b
_ne_C_b:        mov   di, offset _regC

_ne_b:          _tst_b
		je    faux_ne
		jmp   condition_vrai

                align 2
_ne_A_s:        exec_opcode1s _ne A s
_ne_B_s:        exec_opcode1s _ne B s
_ne_D_s:        exec_opcode1s _ne D s
_ne_C_s:        mov   di, offset _regC

_ne_s:          _tst_s
		je    faux_ne
		jmp   condition_vrai

faux_ne:        ; condition_fausse
		clear_carry
		inc   si
		NEXTI

		align 2
		irp   CHAMP,<p, wp, xs, x, m, w>
		local  faux
_ne_&CHAMP&:    call  _tst_&CHAMP&
		je    faux_ne
		jmp   condition_vrai
		endm





; **************************************************************
;               eq
; **************************************************************
; fonctions de test


; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m, w>

_eq_A_&CHAMP&:     exec_opcode _eq A CHAMP
_eq_B_&CHAMP&:     exec_opcode _eq B CHAMP
_eq_C_&CHAMP&:     exec_opcode _eq C CHAMP
_eq_D_&CHAMP&:     exec_opcode _eq D CHAMP

		endm


; **************************************************************
;               ne
; **************************************************************
; fonctions de test

; dfinitions des points d'entres
		irp   CHAMP,<p, wp, xs, x, m, w>

_ne_A_&CHAMP&:     exec_opcode _ne A CHAMP
_ne_B_&CHAMP&:     exec_opcode _ne B CHAMP
_ne_C_&CHAMP&:     exec_opcode _ne C CHAMP
_ne_D_&CHAMP&:     exec_opcode _ne D CHAMP

		endm


; ********************************************************************
;               mulation du processeur Saturn par le 8086
;                       J-F Garnier 1993, 95
; --------------------------------------------------------------------
; Module sat_logi.asm : mulation des instructions logiques
; --------------------------------------------------------------------
; Version 1 16/03/93
;   prototype
; version 1.1 04/11/95
;   reste  faire _sr
; ********************************************************************


; routines logiques :
; x=x&y fs
; x=x!y fs

; x=0   fs      *
; x=y   fs      *
; xyex  fs      *

; Rn=x
; x=Rn
; xRnex

; xsl   fs      *
; xsr   fs      *

; xslc
; xsrc
; xsrb

; instructions marked as '*' actually are arith opcodes






; ***************************************************************
; 		dfinitions des macros
; ***************************************************************



; -------------------------------------------------------------------
; build_logic2
; -------------------------------------------------------------------
; macro de traitement des instructions arithmtiques  2 arguments
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT = _and, _or
; entres: si, di pointent sur les registres oprandes
;          si: source
;          di: destination
;          la sauvegarde de si (pointeur programme) est dans la pile
;          et doit tre restaur avant NEXT
; -------------------------------------------------------------------


build_logic2	macro INSTRUCT
		local logic2_p_i, logic2_p_x, tab_logic2_wp

&INSTRUCT&_p:   push  di
		call  maj2_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    logic2_p_i
		; paire, quartet infrieur
		add   si,ax
		add   di,ax
		call  &INSTRUCT&_lonib
		jmp   short logic2_p_x
logic2_p_i:	; impaire, quartet suprieur
		add   si,ax
		add   di,ax
		call  &INSTRUCT&_hinib
logic2_p_x:     pop   di
		maj_a
		pop   si
		NEXT


&INSTRUCT&_wp:  push  di
		call  maj2_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		call  cs:[bx+tab_logic2_wp]
                pop   di
		maj_a
		pop   si
		NEXT
tab_logic2_wp	dw    &INSTRUCT&_1,  &INSTRUCT&_2,  &INSTRUCT&_3,  &INSTRUCT&_4
		dw    &INSTRUCT&_5,  &INSTRUCT&_6,  &INSTRUCT&_7,  &INSTRUCT&_8
		dw    &INSTRUCT&_9,  &INSTRUCT&_10, &INSTRUCT&_11, &INSTRUCT&_12
		dw    &INSTRUCT&_13, &INSTRUCT&_14, &INSTRUCT&_15, &INSTRUCT&_16


&INSTRUCT&_xs:
		inc   si
		inc   di
		call  &INSTRUCT&_lonib
		pop   si
		NEXT

&INSTRUCT&_x:
		call  &INSTRUCT&_3
		pop   si
		NEXT

&INSTRUCT&_s:
		add   si,7
		add   di,7
		call  &INSTRUCT&_hinib
		pop   si
		NEXT

&INSTRUCT&_m:   push  di
		call  maj2_w
		inc   si
		inc   di
		call  &INSTRUCT&_hinib
		inc   si
		inc   di
		call  &INSTRUCT&_11
                pop   di
		maj_a
		pop   si
		NEXT


		endm








; **************************************************************
; 		and
; **************************************************************

; fonctions de and binaire


and_byte	macro
		mov  al,[si]
		and  [di],al
		inc  si
		inc  di
		endm

_and_hinib:     mov  al,[si]
		or   al,0fh
		and  [di],al
		ret

_and_16:        and_byte
_and_14:        and_byte
_and_12:        and_byte
_and_10:        and_byte
_and_8:         and_byte
_and_6:         and_byte
_and_4:         and_byte
_and_2:         and_byte
	        ret

_and_15:        and_byte
_and_13:        and_byte
_and_11:        and_byte
_and_9:         and_byte
_and_7:         and_byte
_and_5:         and_byte
_and_3:         and_byte
_and_1:
_and_lonib:     mov  al,[si]
	        or   al,0f0h
	        and  [di],al
	        ret

		align 2
_and_A_B_a:     exec_opcode2s _and A B a
_and_B_C_a:     exec_opcode2s _and B C a
_and_C_A_a:     exec_opcode2s _and C A a
_and_D_C_a:     exec_opcode2s _and D C a
_and_B_A_a:     exec_opcode2s _and B A a
_and_C_B_a:     exec_opcode2s _and C B a
_and_A_C_a:     exec_opcode2s _and A C a
_and_C_D_a:     exec_opcode2s _and C D a

_and_a:         mov  ax,[bx]
		and  [di],ax
		mov  al,[bx-1]
		and  [di-1],al
		NEXT

		align 2
_and_A_B_b:     exec_opcode2s _and A B b
_and_B_C_b:     exec_opcode2s _and B C b
_and_C_A_b:     exec_opcode2s _and C A b
_and_D_C_b:     exec_opcode2s _and D C b
_and_B_A_b:     exec_opcode2s _and B A b
_and_C_B_b:     exec_opcode2s _and C B b
_and_A_C_b:     exec_opcode2s _and A C b
_and_C_D_b:     exec_opcode2s _and C D b

_and_b:         mov  al,[bx]
		and  [di],al
		NEXT


_and_w:         call maj2_w
                mov  ax,[si]
                and  [di],ax
                mov  ax,[si+2]
                and  [di+2],ax
                mov  ax,[si+4]
                and  [di+4],ax
                mov  ax,[si+6]
                and  [di+6],ax
                maj_a
		pop  si
		NEXT

; dfinitions du reste du code
		build_logic2 _and

; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m, w>

_and_A_B_&CHAMP&:     exec_opcode2 _and A B CHAMP
_and_B_C_&CHAMP&:     exec_opcode2 _and B C CHAMP
_and_C_A_&CHAMP&:     exec_opcode2 _and C A CHAMP
_and_D_C_&CHAMP&:     exec_opcode2 _and D C CHAMP
_and_B_A_&CHAMP&:     exec_opcode2 _and B A CHAMP
_and_C_B_&CHAMP&:     exec_opcode2 _and C B CHAMP
_and_A_C_&CHAMP&:     exec_opcode2 _and A C CHAMP
_and_C_D_&CHAMP&:     exec_opcode2 _and C D CHAMP

		endm





; **************************************************************
; 		or
; **************************************************************

; fonctions de or binaire


or_byte		macro
		mov  al,[si]
		or   [di],al
		inc  si
		inc  di
		endm

_or_hinib:      mov  al,[si]
		and  al,0f0h
		or   [di],al
		ret

_or_16:         or_byte
_or_14:         or_byte
_or_12:         or_byte
_or_10:         or_byte
_or_8:          or_byte
_or_6:          or_byte
_or_4:          or_byte
_or_2:          or_byte
	        ret

_or_15:         or_byte
_or_13:         or_byte
_or_11:         or_byte
_or_9:          or_byte
_or_7:          or_byte
_or_5:          or_byte
_or_3:          or_byte
_or_1:
_or_lonib:      mov  al,[si]
		and  al,0fh
		or   [di],al
		ret


		align 2
_or_A_B_a:      exec_opcode2s _or A B a
_or_B_C_a:      exec_opcode2s _or B C a
_or_C_A_a:      exec_opcode2s _or C A a
_or_D_C_a:      exec_opcode2s _or D C a
_or_B_A_a:      exec_opcode2s _or B A a
_or_C_B_a:      exec_opcode2s _or C B a
_or_A_C_a:      exec_opcode2s _or A C a
_or_C_D_a:      exec_opcode2s _or C D a

_or_a:          mov  ax,[bx]
		or  [di],ax
		mov  al,[bx-1]
		or  [di-1],al
		NEXT

		align 2
_or_A_B_b:      exec_opcode2s _or A B b
_or_B_C_b:      exec_opcode2s _or B C b
_or_C_A_b:      exec_opcode2s _or C A b
_or_D_C_b:      exec_opcode2s _or D C b
_or_B_A_b:      exec_opcode2s _or B A b
_or_C_B_b:      exec_opcode2s _or C B b
_or_A_C_b:      exec_opcode2s _or A C b
_or_C_D_b:      exec_opcode2s _or C D b

_or_b:          mov  al,[bx]
		or  [di],al
		NEXT

_or_w:          call maj2_w
                mov  ax,[si]
                or   [di],ax
                mov  ax,[si+2]
                or   [di+2],ax
                mov  ax,[si+4]
                or   [di+4],ax
                mov  ax,[si+6]
                or   [di+6],ax
                maj_a
                pop  si
		NEXT

; dfinitions du reste du code
		build_logic2 _or

; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m, w>

_or_A_B_&CHAMP&:     exec_opcode2 _or A B CHAMP
_or_B_C_&CHAMP&:     exec_opcode2 _or B C CHAMP
_or_C_A_&CHAMP&:     exec_opcode2 _or C A CHAMP
_or_D_C_&CHAMP&:     exec_opcode2 _or D C CHAMP
_or_B_A_&CHAMP&:     exec_opcode2 _or B A CHAMP
_or_C_B_&CHAMP&:     exec_opcode2 _or C B CHAMP
_or_A_C_&CHAMP&:     exec_opcode2 _or A C CHAMP
_or_C_D_&CHAMP&:     exec_opcode2 _or C D CHAMP

		endm









; ========================================================
; copie, change de registres
; ========================================================


; -------------------------------------------------------------------
; build_mov
; -------------------------------------------------------------------
; macro de traitement des instructions arithmtiques  2 arguments
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT =  _mov
; entres: si, di pointent sur les registres oprandes
;          si: source
;          di: destination
;          la sauvegarde de si (pointeur programme) est dans la pile
;          et doit tre restaur avant NEXT
; -------------------------------------------------------------------


build_mov	macro INSTRUCT
		local logic2_p_i, logic2_p_x, tab_logic2_wp

&INSTRUCT&_p:   push  di
		call  maj2_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    logic2_p_i
		; paire, quartet infrieur
		add   si,ax
		add   di,ax
		call  &INSTRUCT&_lonib
		jmp   short logic2_p_x
logic2_p_i:	; impaire, quartet suprieur
		add   si,ax
		add   di,ax
		call  &INSTRUCT&_hinib
logic2_p_x:     pop   di
		maj_a
		pop   si
		NEXTI


&INSTRUCT&_wp:  push  di
		call  maj2_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		call  cs:[bx+tab_logic2_wp]
                pop   di
		maj_a
		pop   si
		NEXTI
tab_logic2_wp	dw    &INSTRUCT&_1,  &INSTRUCT&_2,  &INSTRUCT&_3,  &INSTRUCT&_4
		dw    &INSTRUCT&_5,  &INSTRUCT&_6,  &INSTRUCT&_7,  &INSTRUCT&_8
		dw    &INSTRUCT&_9,  &INSTRUCT&_10, &INSTRUCT&_11, &INSTRUCT&_12
		dw    &INSTRUCT&_13, &INSTRUCT&_14, &INSTRUCT&_15, &INSTRUCT&_16


&INSTRUCT&_xs:
		inc   si
		inc   di
		call  &INSTRUCT&_lonib
		pop   si
		NEXTI

&INSTRUCT&_x:
		call  &INSTRUCT&_3
		pop   si
		NEXTI

&INSTRUCT&_s:
		add   si,7
		add   di,7
		call  &INSTRUCT&_hinib
		pop   si
		NEXTI

&INSTRUCT&_m:   push  di
		call  maj2_w
		inc   si
		inc   di
		call  &INSTRUCT&_hinib
		inc   si
		inc   di
		call  &INSTRUCT&_11
                pop   di
		maj_a
		pop   si
		NEXTI


		endm





; **************************************************************
; 		mov
; **************************************************************

; fonctions de copie


mov_byte	macro
		mov  al,[si]
		mov  [di],al
		inc  si
		inc  di
		endm

_mov_hinib:     mov  al,[si]
		mov  ah,[di]
		and  ax,0ff0h
		or   al,ah
		mov  [di],al
		ret

_mov_16:        mov_byte
_mov_14:        mov_byte
_mov_12:        mov_byte
_mov_10:        mov_byte
_mov_8:         mov_byte
_mov_6:         mov_byte
_mov_4:         mov_byte
_mov_2:         mov_byte
	        ret

_mov_15:        mov_byte
_mov_13:        mov_byte
_mov_11:        mov_byte
_mov_9:         mov_byte
_mov_7:         mov_byte
_mov_5:         mov_byte
_mov_3:         mov_byte
_mov_1:
_mov_lonib:     mov  al,[si]
	        mov  ah,[di]
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


		align 2
_mov_A_B_a:     exec_opcode2s _mov A B a
_mov_B_C_a:     exec_opcode2s _mov B C a
_mov_C_A_a:     exec_opcode2s _mov C A a
_mov_D_C_a:     exec_opcode2s _mov D C a
_mov_B_A_a:     exec_opcode2s _mov B A a
_mov_C_B_a:     exec_opcode2s _mov C B a
_mov_A_C_a:     exec_opcode2s _mov A C a
_mov_C_D_a:     exec_opcode2s _mov C D a

_mov_a:         mov  ax,[bx]
		mov  [di],ax
		mov  al,[bx-1]
		mov  [di-1],al
		NEXT

		align 2
_mov_A_B_b:     exec_opcode2s _mov A B b
_mov_B_C_b:     exec_opcode2s _mov B C b
_mov_C_A_b:     exec_opcode2s _mov C A b
_mov_D_C_b:     exec_opcode2s _mov D C b
_mov_B_A_b:     exec_opcode2s _mov B A b
_mov_C_B_b:     exec_opcode2s _mov C B b
_mov_A_C_b:     exec_opcode2s _mov A C b
_mov_C_D_b:     exec_opcode2s _mov C D b

_mov_b:         mov  al,[bx]
		mov  [di],al
		NEXTI

		align 2
_mov_A_B_w:     exec_opcode2s _mov A B w
_mov_B_C_w:     exec_opcode2s _mov B C w
_mov_C_A_w:     exec_opcode2s _mov C A w
_mov_D_C_w:     exec_opcode2s _mov D C w
_mov_B_A_w:     exec_opcode2s _mov B A w
_mov_C_B_w:     exec_opcode2s _mov C B w
_mov_A_C_w:     exec_opcode2s _mov A C w
_mov_C_D_w:     exec_opcode2s _mov C D w

_mov_w:
; mise  jour source !!
                xchg di,bx
                call maj_w
                xchg di,bx
                mov  ax,[bx]
                mov  [di],ax
                mov  ax,[bx+2]
                mov  [di+2],ax
                mov  ax,[bx+4]
                mov  [di+4],ax
		mov  ax,[bx+6]
                mov  [di+6],ax
                maj_a
                NEXTI

; dfinitions du reste du code
		build_mov _mov


                align 2
		irp   REG,<R0, R1, R2, R3, R4, R5, R6, R7>
_mov_&REG&_A:	exec_opcode2s _mov1 REG A w
_mov_&REG&_C:	exec_opcode2s _mov1 REG C w
                endm

; copie d'un registre (A,B,C,D) dans Rx
_mov1_w:
; mise  jour source !!
                xchg bx,di
                call maj_w
                xchg bx,di
                ; copie
                mov  ax,[bx]
                mov  [di],ax
                mov  ax,[bx+2]
                mov  [di+2],ax
                mov  ax,[bx+4]
                mov  [di+4],ax
		mov  ax,[bx+6]
		mov  [di+6],ax
                NEXTI

		align 2
		irp   REG,<R0, R1, R2, R3, R4, R5, R6, R7>
_mov_A_&REG&:	exec_opcode2s _mov2 A REG w
_mov_C_&REG&:	exec_opcode2s _mov2 C REG w
                endm

; copie d'un registre Rx dans (A,B,C,D)
_mov2_w:
                ; copie
                mov  ax,[bx]
                mov  [di],ax
                mov  ax,[bx+2]
                mov  [di+2],ax
                mov  ax,[bx+4]
                mov  [di+4],ax
                mov  ax,[bx+6]
                mov  [di+6],ax
                maj_a
                NEXTI

; new move ?
;               mov  bx,ds
;               xchg es,bx
;               mov  dx,di
;               movsw
;               movsw
;               movsw
;               movsw
;               mov  es,bx
;               mov  di,dx



; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m>

_mov_A_B_&CHAMP&:     exec_opcode2 _mov A B CHAMP
_mov_B_C_&CHAMP&:     exec_opcode2 _mov B C CHAMP
_mov_C_A_&CHAMP&:     exec_opcode2 _mov C A CHAMP
_mov_D_C_&CHAMP&:     exec_opcode2 _mov D C CHAMP
_mov_B_A_&CHAMP&:     exec_opcode2 _mov B A CHAMP
_mov_C_B_&CHAMP&:     exec_opcode2 _mov C B CHAMP
_mov_A_C_&CHAMP&:     exec_opcode2 _mov A C CHAMP
_mov_C_D_&CHAMP&:     exec_opcode2 _mov C D CHAMP

		endm




; -------------------------------------------------------------------
; build_xch
; -------------------------------------------------------------------
; macro de traitement de l'instruction xch
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT = _xch
; entres: si, di pointent sur les registres oprandes
;          si: source
;          di: destination
;          la sauvegarde de si (pointeur programme) est dans la pile
;          et doit tre restaur avant NEXT
; -------------------------------------------------------------------

build_xch	macro INSTRUCT
		local logic2_p_i, logic2_p_x, tab_logic2_wp

&INSTRUCT&_p:   push  di
		push  si
		call  maj2_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    logic2_p_i
		; paire, quartet infrieur
		add   si,ax
		add   di,ax
		call  &INSTRUCT&_lonib
		jmp   short logic2_p_x
logic2_p_i:	; impaire, quartet suprieur
		add   si,ax
		add   di,ax
		call  &INSTRUCT&_hinib
logic2_p_x:     pop   di
		maj_a
		pop   di
		maj_a
		pop   si
		NEXTI


&INSTRUCT&_wp:  push  di
		push  si
		call  maj2_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		call  cs:[bx+tab_logic2_wp]
		pop   di
		maj_a
		pop   di
		maj_a
		pop   si
		NEXTI
tab_logic2_wp	dw    &INSTRUCT&_1,  &INSTRUCT&_2,  &INSTRUCT&_3,  &INSTRUCT&_4
		dw    &INSTRUCT&_5,  &INSTRUCT&_6,  &INSTRUCT&_7,  &INSTRUCT&_8
		dw    &INSTRUCT&_9,  &INSTRUCT&_10, &INSTRUCT&_11, &INSTRUCT&_12
		dw    &INSTRUCT&_13, &INSTRUCT&_14, &INSTRUCT&_15, &INSTRUCT&_16


&INSTRUCT&_xs:
		inc   si
		inc   di
		call  &INSTRUCT&_lonib
		pop   si
		NEXTI

&INSTRUCT&_x:
		call  &INSTRUCT&_3
		pop   si
		NEXTI

&INSTRUCT&_s:
		add   si,7
		add   di,7
		call  &INSTRUCT&_hinib
		pop   si
		NEXTI

&INSTRUCT&_m:   push  di
		push  si
		call  maj2_w
		inc   si
		inc   di
		call  &INSTRUCT&_hinib
		inc   si
		inc   di
		call  &INSTRUCT&_11
		pop   di
		maj_a
		pop   di
		maj_a
		pop   si
		NEXTI


		endm








; **************************************************************
; 		xch
; **************************************************************

; fonctions de copie


xch_byte	macro
		mov  al,[si]
		xchg al,[di]
		mov  [si],al
		inc  si
		inc  di
		endm

_xch_hinib:	;
		mov  al,[si]
		mov  ah,[di]
		mov  bx,ax
		and  ax,0ff0h
		or   al,ah
		mov  [di],al
		mov  ax,bx
		and  ax,0f00fh
		or   al,ah
		mov  [si],al
		ret

_xch_16:        xch_byte
_xch_14:        xch_byte
_xch_12:        xch_byte
_xch_10:        xch_byte
_xch_8:         xch_byte
_xch_6:         xch_byte
_xch_4:         xch_byte
_xch_2:         xch_byte
	        ret

_xch_15:        xch_byte
_xch_13:        xch_byte
_xch_11:        xch_byte
_xch_9:         xch_byte
_xch_7:         xch_byte
_xch_5:         xch_byte
_xch_3:         xch_byte
_xch_1:
_xch_lonib:	mov  al,[si]
		mov  ah,[di]
		mov  bx,ax
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		mov  ax,bx
		and  ax,0ff0h
		or   al,ah
		mov  [si],al
		ret

		align 2
_xch_A_B_a:     exec_opcode2s _xch A B a
_xch_B_C_a:     exec_opcode2s _xch B C a
_xch_A_C_a:     exec_opcode2s _xch A C a
_xch_C_D_a:     exec_opcode2s _xch C D a

_xch_a:         mov  ax,[bx]
		xchg ax,[di]
		mov  [bx],ax
		mov  al,[bx-1]
		xchg al,[di-1]
		mov  [bx-1],al
		NEXT

		align 2
_xch_A_B_b:     exec_opcode2s _xch A B b
_xch_B_C_b:     exec_opcode2s _xch B C b
_xch_A_C_b:     exec_opcode2s _xch A C b
_xch_C_D_b:     exec_opcode2s _xch C D b

_xch_b:         mov  al,[bx]
		xchg al,[di]
		mov  [bx],al
		NEXTI

_xch_w:         call maj2_w
                mov  ax,[si]
                xchg ax,[di]
                mov  [si],ax
                mov  ax,[si+2]
                xchg ax,[di+2]
                mov  [si+2],ax
                mov  ax,[si+4]
                xchg ax,[di+4]
                mov  [si+4],ax
                mov  ax,[si+6]
                xchg ax,[di+6]
		mov  [si+6],ax
; mise  jour des 2 registres !
		maj_a
		mov  di,si
		maj_a
		pop  si
		NEXTI


; dfinitions du reste du code
		build_xch _xch


                align 2
		irp   REG,<R0, R1, R2, R3, R4, R5, R6, R7>
_xch_A_&REG&:	exec_opcode2s _xch1 A REG w
_xch_C_&REG&:	exec_opcode2s _xch1 C REG w
                endm

; echange entre un registre (A,B,C,D) et Rx
_xch1_w:        call maj_w
		mov  ax,[bx]
		xchg ax,[di]
		mov  [bx],ax
		mov  ax,[bx+2]
		xchg ax,[di+2]
		mov  [bx+2],ax
		mov  ax,[bx+4]
		xchg ax,[di+4]
		mov  [bx+4],ax
		mov  ax,[bx+6]
		xchg ax,[di+6]
		mov  [bx+6],ax
		maj_a
		NEXTI



; -----------------------------------------



; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m, w>

_xch_A_B_&CHAMP&:     exec_opcode2 _xch A B CHAMP
_xch_B_C_&CHAMP&:     exec_opcode2 _xch B C CHAMP
_xch_A_C_&CHAMP&:     exec_opcode2 _xch A C CHAMP
_xch_C_D_&CHAMP&:     exec_opcode2 _xch C D CHAMP

		endm









; -------------------------------------------------------------------
; build_logic
; -------------------------------------------------------------------
; macro de traitement des instructions arithmtiques  1 argument
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT = _sl, _clr
; entres: di pointe sur le registre oprande destination
; -------------------------------------------------------------------

build_logic	macro INSTRUCT
		local logic_p_i, logic_p_x, tab_logic_wp

&INSTRUCT&_p:   push  di
		call  maj_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    logic_p_i
		; paire, quartet infrieur
		add   di,ax
		xor   ax,ax
		call  &INSTRUCT&_lonib
		jmp   short logic_p_x
logic_p_i:	; impaire, quartet suprieur
		add   di,ax
		xor   ax,ax
		call  &INSTRUCT&_hinib
logic_p_x:      pop   di
		maj_a
		NEXTI


&INSTRUCT&_wp:  push  di
		call  maj_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		xor   ax,ax
		call  cs:[bx+tab_logic_wp]
		pop   di
		maj_a
		NEXTI
tab_logic_wp	dw    &INSTRUCT&_1,  &INSTRUCT&_2,  &INSTRUCT&_3,  &INSTRUCT&_4
		dw    &INSTRUCT&_5,  &INSTRUCT&_6,  &INSTRUCT&_7,  &INSTRUCT&_8
		dw    &INSTRUCT&_9,  &INSTRUCT&_10, &INSTRUCT&_11, &INSTRUCT&_12
		dw    &INSTRUCT&_13, &INSTRUCT&_14, &INSTRUCT&_15, &INSTRUCT&_16

;&INSTRUCT&_n:   push  di
;		call  maj_w
;		add   bx,bx
;		xor   ax,ax
;		call  cs:[bx+tab_logic_wp]
;		pop   di
;		maj_a
;		NEXT

&INSTRUCT&_xs:  inc   di
		xor   ax,ax
		call  &INSTRUCT&_lonib
		NEXTI

&INSTRUCT&_x:   xor   ax,ax
		call  &INSTRUCT&_3
		NEXTI

&INSTRUCT&_s:   add   di,7
		xor   ax,ax
		call  &INSTRUCT&_hinib
		NEXTI

&INSTRUCT&_m:   push  di
		call  maj_w
		inc   di
                xor   ax,ax
		call  &INSTRUCT&_hinib
		inc   di
		call  &INSTRUCT&_11
                pop   di
		maj_a
		NEXTI

		endm





; **************************************************************
; 		clr
; **************************************************************

; fonctions de mise  0


clr_byte        macro
		mov  [di],al
		inc  di
		endm

_clr_hinib:     and  byte ptr [di],0fh
		ret

_clr_16:        clr_byte
_clr_14:        clr_byte
_clr_12:        clr_byte
_clr_10:        clr_byte
_clr_8:         clr_byte
_clr_6:         clr_byte
_clr_4:         clr_byte
_clr_2:         clr_byte
		ret

_clr_15:        clr_byte
_clr_13:        clr_byte
_clr_11:        clr_byte
_clr_9:         clr_byte
_clr_7:         clr_byte
_clr_5:         clr_byte
_clr_3:         clr_byte
_clr_1:
_clr_lonib:     and  byte ptr [di],0f0h
		ret


		align 2
_clr_A_a:       exec_opcode1s _clr A a
_clr_B_a:       exec_opcode1s _clr B a
_clr_D_a:       exec_opcode1s _clr D a
;_clr_C_a:       exec_opcode1s _clr C a
_clr_C_a:       mov   di, offset _regC

_clr_a:         xor  ax,ax
		mov  [di],ax
		mov  [di-1],al
		NEXT

		align 2
_clr_A_b:       exec_opcode1s _clr A b
_clr_B_b:       exec_opcode1s _clr B b
_clr_D_b:       exec_opcode1s _clr D b
_clr_C_b:       mov   di, offset _regC

_clr_b:		mov  byte ptr [di],0
		NEXTI

		align 2
_clr_A_w:       exec_opcode1s _clr A w
_clr_B_w:       exec_opcode1s _clr B w
_clr_D_w:       exec_opcode1s _clr D w
_clr_C_w:       mov   di, offset _regC

_clr_w:         xor  ax,ax
		mov  [di],ax
		mov  [di+2],ax
		mov  [di+4],ax
		mov  [di+6],ax
		maj_a
		NEXTI

; dfinitions du reste du code
		build_logic _clr

; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m>

_clr_A_&CHAMP&:     exec_opcode _clr A CHAMP
_clr_B_&CHAMP&:     exec_opcode _clr B CHAMP
_clr_C_&CHAMP&:     exec_opcode _clr C CHAMP
_clr_D_&CHAMP&:     exec_opcode _clr D CHAMP

		endm
















; **************************************************************
; 		sl
; **************************************************************

; fonctions de dcallage  gauche de 1 quartet
; SB inchang


; ah[7-4] doit contenir le quartet issu du decallage precedent
sl_byte         macro
		mov  al,[di]
		rol  ax,4
		mov  [di],al
		shl  ah,4   ; quartet decalle en ah[7-0]
		inc  di
		endm

_sl_hinib:      mov  al,[di]
		mov  ah,al
		and  al,0fh
		mov  [di],al
		ret

_sl_16:         sl_byte
_sl_14:         sl_byte
_sl_12:         sl_byte
_sl_10:         sl_byte
_sl_8:          sl_byte
_sl_6:          sl_byte
_sl_4:          sl_byte
_sl_2:          sl_byte
		ret

_sl_15:         sl_byte
_sl_13:         sl_byte
_sl_11:         sl_byte
_sl_9:          sl_byte
_sl_7:          sl_byte
_sl_5:          sl_byte
_sl_3:          sl_byte
_sl_1:
_sl_lonib:      mov  al,[di]
		shr  ah,4
		and  ax,0ff0h
		or   al,ah
		mov  [di],al
		ret


		align 2
_sl_A_a:        exec_opcode1s _sl A a
_sl_B_a:        exec_opcode1s _sl B a
_sl_D_a:        exec_opcode1s _sl D a
_sl_C_a:        mov   di, offset _regC

_sl_a:		mov  ax,[di]
		rol  ax,4
		mov  [di-1],al
		and  al,0f0h
		mov  [di],ax
		NEXT

		align 2
_sl_A_b:        exec_opcode1s _sl A b
_sl_B_b:        exec_opcode1s _sl B b
_sl_D_b:        exec_opcode1s _sl D b
_sl_C_b:        mov   di, offset _regC

_sl_b:          mov  al,[di]
		shl  al,4
		mov  [di],al
		NEXTI

		align 2
_sl_A_w:        exec_opcode1s _sl A w
_sl_B_w:        exec_opcode1s _sl B w
_sl_D_w:        exec_opcode1s _sl D w
_sl_C_w:        mov   di, offset _regC

_sl_w:          call maj_w
		push cx
		mov  ax,[di]
		mov  bx,[di+2]
		mov  cx,[di+4]
		mov  dx,[di+6]
		rept 4
		shl  ax,1
		rcl  bx,1
		rcl  cx,1
		rcl  dx,1
                endm
                mov  [di],ax
                mov  [di+2],bx
                mov  [di+4],cx
                mov  [di+6],dx
                pop  cx
		maj_a
		NEXTI



; dfinitions du reste du code
		build_logic _sl

; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m>

_sl_A_&CHAMP&:     exec_opcode _sl A CHAMP
_sl_B_&CHAMP&:     exec_opcode _sl B CHAMP
_sl_C_&CHAMP&:     exec_opcode _sl C CHAMP
_sl_D_&CHAMP&:     exec_opcode _sl D CHAMP

		endm








; -----------------------------------------

; -------------------------------------------------------------------
; build_logic0
; -------------------------------------------------------------------
; macro de traitement des instructions sr
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m, w
; argument INSTRUCT = _sr
; entres: di pointe sur le registre oprande destination
; -------------------------------------------------------------------

build_sr	macro INSTRUCT
		local logic_p_i, logic_p_x, tab_logic_wp

&INSTRUCT&_p:   push  di
		call  maj_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    logic_p_i
		; paire, quartet infrieur
		add   di,ax
		call  &INSTRUCT&_lonib
		call  &INSTRUCT&_0
		jmp   short logic_p_x
logic_p_i:	; impaire, quartet suprieur
		add   di,ax
		xor   ax,ax
		call  &INSTRUCT&_hinib
logic_p_x:      pop   di
		maj_a
		NEXTI


&INSTRUCT&_wp:  push  di
		call  maj_w
		xor   ax,ax
		mov   al,_Pointer
		mov   bx,ax
		shr   al,1
		jc    logic_wp1
		add   di,ax
		call  &INSTRUCT&_lonib
		dec   di
		; ici, bx paire, ok pour jmp
		jmp   short logic_wp2
logic_wp1:      add   di,ax
		xor   ax,ax
		; ajuste bx pour etre paire, et ok pour jmp
		inc   bx
logic_wp2:      call  cs:[bx+tab_logic_wp]
		pop   di
		maj_a
		NEXTI
tab_logic_wp	dw    &INSTRUCT&_0,  &INSTRUCT&_2,  &INSTRUCT&_4,  &INSTRUCT&_6
		dw    &INSTRUCT&_8,  &INSTRUCT&_10, &INSTRUCT&_12, &INSTRUCT&_14
		dw    &INSTRUCT&_16

&INSTRUCT&_xs:  inc   di
		call  &INSTRUCT&_lonib
		call  &INSTRUCT&_0
		NEXTI

&INSTRUCT&_x:   inc   di
		call  &INSTRUCT&_lonib
		dec   di
		call  &INSTRUCT&_2
		NEXTI

&INSTRUCT&_s:   add   di,7
		xor   ax,ax
		call  &INSTRUCT&_hinib
		NEXTI

&INSTRUCT&_m:   push  di
		call  maj_w
		add   di,7
		xor   ax,ax
		call  &INSTRUCT&_lonib
		dec   di
		call  &INSTRUCT&_11
		pop   di
		maj_a
		NEXTI

;&INSTRUCT&_w:   push  di
;		call  maj_w
;                xor   ax,ax
;		call  &INSTRUCT&_16
;                pop   di
;		maj_a
;		NEXTI

		endm





; **************************************************************
; 		sr
; **************************************************************

; fonctions de dcallage  droite de 1 quartet
; SB=1 si quartet perdu<>0

; ah[3-0] doit contenir le quartet issu du decallage precedent
sr_byte         macro
		mov  al,[di]
		ror  ax,4
		mov  [di],al
		dec  di
		shr  ah,4
		endm

_sr_lonib:      mov  al,[di]
		mov  ah,al
		and  al,0f0h
		mov  [di],al
		ret

_sr_16:         sr_byte
_sr_14:         sr_byte
_sr_12:         sr_byte
_sr_10:         sr_byte
_sr_8:          sr_byte
_sr_6:          sr_byte
_sr_4:          sr_byte
_sr_2:          sr_byte
_sr_0:          and  ah,0fh
		je   _sr_00
		or   _hwstatus,2       ; SB
_sr_00:         ret

_sr_15:         sr_byte
_sr_13:         sr_byte
_sr_11:         sr_byte
_sr_9:          sr_byte
_sr_7:          sr_byte
_sr_5:          sr_byte
_sr_3:          sr_byte
_sr_1:
_sr_hinib:      mov  al,[di]
		mov  bh,al
		ror  ax,4
		mov  bl,al
		and  bx,0ff0h
		or   bl,bh
		mov  [di],bl
		and  al,0fh
		je   _sr_h1
		or   _hwstatus,2       ; SB
_sr_h1:	        ret



		align 2
_sr_A_a:        exec_opcode1s _sr A a
_sr_B_a:        exec_opcode1s _sr B a
_sr_D_a:        exec_opcode1s _sr D a
_sr_C_a:        mov   di, offset _regC

_sr_a:          xor  bx,bx
		mov  ax,[di]         ; ax : [ 3 2 1 0 ]
		xchg bl,[di-1]       ; bl : [ x 4 ]
		mov  bh,al           ; bx : [ 1 0 x 4 ]
		and  bx,0f0fh        ; bx = [ - 0 - 4 ]
		or   bh,bh
		je   _sr_a1
		or   _hwstatus,2       ; SB
_sr_a1:         and  al,0f0h         ; ax : [ 3 2 1 - ]
		or   al,bl           ; ax : [ 3 2 1 4 ]
		ror  ax,4            ; ax : [ 4 3 2 1 ]
		mov  [di],ax
		NEXT

		align 2
_sr_A_b:        exec_opcode1s _sr A b
_sr_B_b:        exec_opcode1s _sr B b
_sr_D_b:        exec_opcode1s _sr D b
_sr_C_b:        mov   di, offset _regC

_sr_b:          xor  ax,ax
		mov  al,[di]
		ror  ax,4
		mov  [di],al
		or   ah,ah
		je   _sr_b1
		or   _hwstatus,2       ; SB
_sr_b1:		NEXTI

		align 2
_sr_A_w:        exec_opcode1s _sr A w
_sr_B_w:        exec_opcode1s _sr B w
_sr_D_w:        exec_opcode1s _sr D w
_sr_C_w:        mov   di, offset _regC

_sr_w:          call maj_w
		push cx
		mov  ax,[di]
		mov  bx,ax
		and  bl,0fh
		je   _sr_w1
		or   _hwstatus,2       ; SB
_sr_w1:         mov  bx,[di+2]
		mov  cx,[di+4]
		mov  dx,[di+6]
		rept 4
		shr  dx,1
		rcr  cx,1
		rcr  bx,1
		rcr  ax,1
		endm
		mov  [di],ax
		mov  [di+2],bx
		mov  [di+4],cx
		mov  [di+6],dx
		pop  cx
		maj_a
		NEXTI


; dfinitions du reste du code
		build_sr _sr

; -----------------------------------------







; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m>

_sr_A_&CHAMP&:     exec_opcode _sr A CHAMP
_sr_B_&CHAMP&:     exec_opcode _sr B CHAMP
_sr_C_&CHAMP&:     exec_opcode _sr C CHAMP
_sr_D_&CHAMP&:     exec_opcode _sr D CHAMP

		endm






; **************************************************************
; rotations
; _slc, _src, _srb
; **************************************************************


		align 2
_slc_A:         exec_opcode1s _slc A w
_slc_B:         exec_opcode1s _slc B w
_slc_D:         exec_opcode1s _slc D w

; rotation d'un quartet vers la gauche du champ w
; SB inchang
_slc_w:
		call maj_w
		push cx
		push si
		mov  dx,[di+6]
		mov  si,dx
		mov  cx,[di+4]
		mov  bx,[di+2]
		mov  ax,[di]
		rept 4
		shl  si,1
		rcl  ax,1
		rcl  bx,1
		rcl  cx,1
		rcl  dx,1
		endm
		mov  [di],ax
		mov  [di+2],bx
		mov  [di+4],cx
		mov  [di+6],dx
		pop  si
		pop  cx
		maj_a
		NEXTI


_slc_C:
		mov  di, offset _regC
                jmp  short _slc_w   ; to be removed
;  this is a code to execute 2 CSLC W consecutive opcodes in one go:
                incr_pc
                mov  ax,es:[si]
                sar  ax,cl
                and  ah,0fh
                cmp  ax,0218h   ; CSRC is 812
                jne  short _slc_w
; ok next opcode is a 2nd CSLC
                inc  si
                maj_ws
		mov  bl,[di+7]
                mov  ax,[di+5]
                mov  [di+6],ax
                mov  ax,[di+3]
                mov  [di+4],ax
                mov  ax,[di+1]
                mov  [di+2],ax
                mov  ah,[di]
                mov  al,bl
                mov  [di],ax
                maj_a
                NEXTI



		align 2
_src_A:         exec_opcode1s _src A w
_src_B:         exec_opcode1s _src B w
_src_D:         exec_opcode1s _src D w

; rotation d'un quartet vers la droite du champ w
; SB=1 si quartet report en poids fort <>0
_src_w:
                call maj_w
                push cx
		push si
		mov  ax,[di]
		mov  si,ax
		mov  bx,[di+2]
		mov  cx,[di+4]
		mov  dx,[di+6]
		rept 4
		shr  si,1
		rcr  dx,1
		rcr  cx,1
		rcr  bx,1
		rcr  ax,1
		endm
		mov  [di],ax
		mov  [di+2],bx
		mov  [di+4],cx
		mov  [di+6],dx
		pop  si
		pop  cx
		and  dh,0f0h
		je   _src_w1
		or   _hwstatus,2       ; SB
_src_w1:        maj_a
                NEXTI


_src_C:
		mov   di, offset _regC
                jmp   short _src_w   ; to be removed
;  this is a code to execute 2 CSRC W consecutive opcodes in one go:
                incr_pc
                mov ax,es:[si]
                sar ax,cl
                and ah,0fh
                cmp ax,0618h   ; CSRC is 816
                jne short _src_w
; ok next opcode is a 2nd CSRC
                inc si
                maj_ws
		mov  bl,[di]
		or   bl,bl
		je   _sr_w3
		or   _hwstatus,2       ; SB
_sr_w3:
                mov ax,[di+1]
                mov [di],ax
                mov ax,[di+3]
                mov [di+2],ax
                mov ax,[di+5]
                mov [di+4],ax
                mov al,[di+7]
                mov ah,bl
                mov [di+6],ax
                maj_a
                NEXTI





		align 2
_srb_A:         exec_opcode1s _srb A w
_srb_B:         exec_opcode1s _srb B w
_srb_D:         exec_opcode1s _srb D w

_srb_C:		mov   di, offset _regC

; dcallage d'un bit vers la droite du champ w
; SB=1 si carry
_srb_w:
		call maj_w
                push cx
		mov  ax,[di]
                mov  bx,[di+2]
                mov  cx,[di+4]
		mov  dx,[di+6]
		shr  dx,1
		rcr  cx,1
		rcr  bx,1
		rcr  ax,1
		mov  [di],ax
		mov  [di+2],bx
		mov  [di+4],cx
		mov  [di+6],dx
		pop  cx
		jnc  _srb_w1
		or   _hwstatus,2       ; SB
_srb_w1:        maj_a
		NEXTI








; ********************************************************************
;               mulation du processeur Saturn par le 8086
;                       J-F Garnier 1993, 2001
; --------------------------------------------------------------------
; Module sat_jmp.inc : mulation des instructions de saut
; --------------------------------------------------------------------
; Version 0.0  04/09/93
;   prototype, faisabilit
; Version 1.0  04/11/95
;   finalisation, mise au point
; Version 2.0  09/09/01
;   add opcodes for HP42S emulation
; ********************************************************************







; goc, gonc:
; de la forme 4aa ou 5aa
; saut en addr+aa, o addr est l'adresse du 2me quartet de l'instruction
; en entre, ah[4-0] contient le 2me quartet
; optimisation:
; goc xy, gonc xy : ici, ax = [ - x  y - ]


; goc xy, gonc xy : ici, ax = [ - y x 5 ]
_goc:           test_carry
                jne  jump
                NEXTI

_gonc:          test_carry
                je   jump
                NEXTI


jump:           shr  ax,4     ; ax = [ - - x y ]
		cbw
		dec  ax    ; dplacement par rapport au 2me quartet

dojump:         ; effectue l'addition entre si,cl et ax
		sar  ax,1
		jnc  dojump1
		xor  cl,4
		jne  dojump1
		inc  si
dojump1:        add  si,ax
		js   fixjump
		NEXT
fixjump:	; changement de segment
		or   ax,ax
		mov  al,_regPC+2
		js   fixminus
		inc  al
		jmp  short fixend
fixminus:	dec  al
fixend:         and  ax,0fh
		mov  _regPC+2,al
		add  ax,ax
		mov  bx,ax
		mov  ax,[_tab_seg+bx]
                or   ax,ax
                je   fixend0
		mov  es,ax
		and  si,07fffh
		NEXT
fixend0:        jmp  resetpc




_goto:
                ; al[7-4]=quartet faible
		mov  bx,es:[si]
		shr  bx,cl         ; bl=quartets moyen et fort
                mov  ah,bl
                sar  ax,4          ; ax: dplacement en quartets
                dec  ax            ; pour pointer sur 2me quartet
                jmp  short dojump


_golong:
		mov  ax,es:[si]
		sar  ax,cl         ; al=octet faible
		mov  bx,es:[si+1]
		sar  bx,cl         ; bl=octet fort
                mov  ah,bl
		jmp  short dojump



_govlng:        mov  ax,es:[si]
                shr  ax,cl         ; al= low nibbles
                inc  si
                mov  dx,es:[si]
                shr  dx,cl         ; dl= medium nibbles, dh=high nibble
		mov  bl,dh
		and  bx,0fh
		mov  _regPC+2,bl
		mov  ah,dl
		xchg ax,si      ; ici bx-si contiennent une adresse 20 bits
_govlng1:
; conversion en pointer es:si, cl
; TEST for jump to address 0 (includes RET with empty stack)
                mov   ax,bx
                or    ax,si
                je    _gobreak
		xor   cl,cl
		shr   si,1
		rcl   cl,3
		add   bx,bx
		mov   bx,[_tab_seg+bx]
                or    bx,bx
                je    _govlng0
		mov   es,bx
		NEXT
_govlng0:	jmp   resetpc
_gobreak:       jmp   invalid


_gosub:
		mov  bl,al         ; bl[7-4]=quartet faible
		fetch_byte         ; al=quartets moyen et fort
                mov  ah,al
                mov  al,bl
		sar  ax,4          ; ax: dplacement en quartets

do_jsr:
		xchg ax,dx         ; save displacement in dx
		; construit bx pointant dans pile retour
                ; build di = stack ptr, decr stack ptr
		mov  al,_retstkptr
		sub  al,4
		mov  _retstkptr,al
		xor  ah,ah
		add  ax,offset _retstk
		xchg ax,di
                ;
		; convert regPC, si, cl in  20 bits addr
		; and save PC
		mov  al,_regPC+2
		mov  [di+2],al
		mov  al,cl
		mov  bx,si
		shr  al,3   ; carry = bit 0
		rcl  bx,1
		mov  [di],bx
		xchg ax,dx         ; restore displacement
		jmp  dojump


_gosubl:
		fetch_byte
                mov  bl,al         ; octet faible
                fetch_byte
                mov  ah,al         ; octet fort
		mov  al,bl
                jmp  do_jsr


_gosbvl:
		; construit bx pointant dans pile retour
		mov  al,_retstkptr
		sub  al,4
		mov  _retstkptr,al
		xor  ah,ah
		add  ax,offset _retstk
		xchg ax,bx
		; convert regPC, si, cl en adr quartet 20 bits ah-di
		mov  al,cl
		mov  di,si
		shr  al,3   ; carry = bit 0
		rcl  di,1
                add  di,5
		mov  al,_regPC+2
                adc  al,0
		; sauve PC
		mov  [bx],di
		mov  [bx+2],al
                ;
                jmp  _govlng



; rtnc , rtnnc
_rtnc:          test_carry
                jne  _rtn
                NEXTI

_rtnnc:         test_carry
                je   _rtn
                NEXTI


_rtnsc:         set_carry
		jmp  short _rtn

_rtncc:         clear_carry
		jmp  short _rtn

_rti:           ; devrait autoriser interruptions
		jmp  short _rtn

_rtnsxm:
; change JFG 7/01/06: check if call to debugger or breakpoint,
; else just execute normal RTNSXM
		mov  bh,_regPC+2
		mov  bl,cl
		mov  ax,si
                dec  ax      ; pc -= 2
		shr  bl,3   ; carry = bit 0
		rcl  ax,1
                cmp  ax,010h
                jne  _rtnsxm1
                cmp  bh,0fh   ; call to debug?
                je   _rtnsxm8
_rtnsxm1:       cmp  ax,word ptr _lastaddr
                jne  _rtnsxm9
                cmp  bh,byte ptr _lastaddr+2  ; breakpoint?
                jne  _rtnsxm9
_rtnsxm8:       mov  ax,8
		jmp  retour
_rtnsxm9:	or   _hwstatus,1    ; XM
; end of change.
; fall into _rtn

_rtn:           mov  al,_retstkptr
                mov  bl,al
                xor  ah,ah
                add  ax, offset _retstk
		xchg ax,bx
                add  al,4
                mov  _retstkptr,al
                ;
                xor  si,si
		xchg si,[bx]
                xor  ax,ax
		xchg al,[bx+2]
		and  al,0fh
                mov  bx,ax
		mov  _regPC+2,bl
		; ici bx-si contiennent une adresse 20 bits
                jmp  _govlng1



_mov_PC_A:      mov  di,offset _regA
                jmp short _mov_PC_x
_mov_PC_C:      mov  di,offset _regC
_mov_PC_x:      mov  si,[di]
                mov  bl,[di-1]
                and  bx,0fh
		mov  _regPC+2,bl
                ; ici bx-si contiennent une adresse 20 bits
                jmp  _govlng1


_movi_PC_C:     mov  di,offset _regC
                jmp short _movi_PC_x
                align 2
_movi_PC_A:     mov  di,offset _regA
_movi_PC_x:     mov  si,[di]
                mov  bl,[di-1]
                and  bx,0fh
                ; ici bx-si contiennent une adresse 20 bits
                ; conversion en pointer es:si, cl
		xor   cl,cl
		shr   si,1
		rcl   cl,3
		add   bx,bx
		mov   bx,[_tab_seg+bx]
                or    bx,bx
                je    _movi_PC_0
		mov   es,bx
                jmp  _govlng
_movi_PC_0:     ; cas es=0
                jmp  resetPC


_mov_A_PC:      mov  di,offset _regA
                jmp short _mov_x_PC
_mov_C_PC:      mov  di,offset _regC
_mov_x_PC:
                jmp  invalid
		; convert regPC, si, cl en adr quartet 20 bits ah-bx
		mov  ah,_regPC+2
		mov  al,cl
		mov  bx,si
		shr  al,3   ; carry = bit 0
		rcl  bx,1
                mov  [di],bx
                mov  [di-1],ah
                NEXT



_xch_A_PC:      mov  di,offset _regA
                jmp short _xch_x_PC
_xch_C_PC:      mov  di,offset _regC
_xch_x_PC:
                jmp  invalid
		; convert regPC, si, cl en adr quartet 20 bits ah-bx
		mov  ah,_regPC+2
		mov  al,cl
		mov  bx,si
		shr  al,3   ; carry = bit 0
		rcl  bx,1
                xchg [di],bx
                xchg [di-1],ah
                mov  si,bx
                mov  bl,ah
                and  bx,0fh
                ; ici bx-si contiennent une adresse 20 bits
                jmp  _govlng1

; ********************************************************************
;               mulation du processeur Saturn par le 8086
;                       J-F Garnier 1993
; --------------------------------------------------------------------
; Module sat_ctrl.asm : mulation des instructions de controle
; --------------------------------------------------------------------
; Version 1.0 04/09/93
;   prototype
; Version 1.1 04/11/95
; Version 2 09/09/01 add some systems opcodes, and bit op for HP42S support
; ********************************************************************



; ------------------------------------------------
; modes decimal, binaire
; ------------------------------------------------

_sethex:        clr_decimal
		NEXT

_setdec:        set_decimal
		NEXT

; ------------------------------------------------
; pile
; ------------------------------------------------

_oldrstk_c:        ; construit bx pointant dans pile retour
		mov  al,_retstkptr
		dec  al
		and  ax,7
		mov  _retstkptr,al
		sal  ax,2
		add  ax,offset _retstk
		mov  bx,ax
                ;
		mov  ax,word ptr _regC
		mov  [bx],ax
		mov  al,_regC-1
		mov  [bx+2],al
		NEXT


_rstk_c:       ; construit bx pointant dans pile retour
		mov  al,_retstkptr
		sub  al,4
		mov  _retstkptr,al
                xor  ah,ah
                add  ax,offset _retstk
                xchg ax,bx
                ;
		mov  ax,word ptr _regC
		mov  [bx],ax
		mov  al,_regC-1
                and  al,0fh
		mov  [bx+2],al
		NEXT



_oldc_rstk:
		mov  al,_retstkptr
		xor  bh,bh
		mov  bl,al
		inc  al
		and  al,7
		mov  _retstkptr,al
		sal  bx,2
		add  bx, offset _retstk
		mov  ax,[bx]
		mov  word ptr _regC,ax
		mov  al,[bx+2]
		mov  _regC-1,al
		; remet  zro le niveau suprieur de la pile
		xor  ax,ax
		mov  [bx],ax
		mov  [bx+2],ax
		NEXT

_c_rstk:
		mov  al,_retstkptr
                mov  bl,al
                xor  ah,ah
                add  ax,offset _retstk
                xchg ax,bx
		add  al,4
		mov  _retstkptr,al
                ;
                xor  ax,ax
		xchg ax,[bx]
		mov  word ptr _regC,ax
                xor  al,al
		xchg  al,[bx+2]
		mov  _regC-1,al
                ;
		NEXT

; ------------------------------------------------
; LCHEX, LAHEX
; ------------------------------------------------


_lahex:         fetch_nib
                mov  di,offset _regA
		test_pointer
		jne  loadhex0
		; P=0
		or  al,al
		jne  load0hex2
		; lc(1)
		fetch_nib
                mov  ah,[di]
                and  ah,0f0h
                or   al,ah
		mov  [di],al
		NEXT
load0hex2:	cmp  al,1
		jne  load0hex5
		; lc(2)
		fetch_byte
		mov  [di],al
		NEXT
load0hex5:      cmp  al,4
		jne  load0hex4
		; lc(5)
		fetch_byte
		mov  [di],al
		fetch_3nib
		mov  [di+1],al
		mov  [di-1],ah
                set_majw
		NEXT
load0hex4:      cmp  al,3
                jne  loadhex0
		; lc(4)
		fetch_byte
		mov  [di],al
		fetch_byte
		mov  [di+1],al
		NEXT
loadhex0:       jmp  loadhex


; LC(1) ax=-x03
                align 2
_lchex1:        test_pointer
                jne  _lchex1x
                mov  al,_regC
                and  al,0f0h
                or   al,ah
                mov  _regC,al
                NEXTI
_lchex1x:       xor  bx,bx
                mov  bl,_Pointer
                shr  bl,1
                jnc  _lchex1c
                add  bx,offset _regC
                mov  al,[bx]
                and  al,0fh
                shl  ah,4
                or   al,ah
                mov  [bx],al
                NEXTI
_lchex1c:       cmp  bl,2
                je   _lchex14
                add  bx,offset _regC
                mov  al,[bx]
                and  al,0f0h
                or   al,ah
                mov  [bx],al
                NEXTI
_lchex14:       ; load 1 nibble in C[4]
                mov  _regC-1,ah
                set_majw
                NEXTI



; LC(3) ax=-x23
_lchex3:        test_pointer
                jne  _lchex44
                mov  di, offset _regC
		fetch_byte
                mov  [di],al
                mov  al,[di+1]
                and  ax,0ff0h
                or   al,ah
                mov  [di+1],al
                NEXTI

; LC(6) ax=-x53
_lchex6:        test_pointer
                jne  _lchex44
                mov  di, offset _regC
		fetch_byte
                mov  [di],al
		fetch_byte
                mov  [di+1],al
		fetch_byte
                mov  [di+2],al
                mov  [di-1],al
                clr_majw
                NEXT

; LC(4) ax=-x33
_lchex4:        test_pointer
_lchex44:       jne  _lchex
                mov  di, offset _regC
		fetch_byte
                mov  [di],al
		fetch_byte
                mov  [di+1],al
                NEXT


; LC(5) ax=-x43
                align 2
_lchex5:        test_pointer
                jne  _lchex
                mov  di, offset _regC
		fetch_byte
                mov  [di],al
		fetch_byte
                mov  [di+1],al
                mov  [di-1],ah
                set_majw
                NEXTI


; LC(2) ax=-x13
                align 2
_lchex2:        test_pointer
_lchex40:       jne  _lchex
		fetch_byte
                mov  _regC,al
                NEXT


                align 2
_lchex:         ; charge C  partir de la position P pour n+1 quartets
		; al[7-4] contient n
                shr  al,4
		mov  di,offset _regC

loadhex:        ; cas gnral
		; si P impaire:
		;   fetch_nib  --> hinib
		;   n--
		;   p++; si P>15 p=0
		; tant que n>0
		;   fetch_byte
		;   n-=2
		;   p+=2; si p>15 p=0
		; si n=0
		;   fetch_nib  --> lonib
		mov  dl,al        ; dl : n
		call maj_w
		push di
		xor  ax,ax
		mov  al,_Pointer
		mov  dh,al        ; dh : P
		shr  al,1
		add  di,ax        ; di = pointeur dans reg
		test dh,1
		je   loadhex1
		fetch_nib
		shl  al,4
		mov  ah,[di]
		and  ax,00ff0h
		or   al,ah
		mov  [di],al
		dec  dl
		js   loadhex9
		inc  di
		inc  dh
		cmp  dh,16
		jne  loadhex1
		sub  di,8
loadhex1:       or   dl,dl
		js   loadhex9
		je   loadhex2
		fetch_byte
		mov  [di],al
		inc  di
		dec  dl
		dec  dl
		inc  dh
		inc  dh
		cmp  dh,16
		jne  loadhex1
		sub  di,8
		jmp  loadhex1
loadhex2:       fetch_nib
		mov  ah,[di]
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
loadhex9:       pop  di
		maj_a
		NEXT

; ------------------------------------------------
; registre status
; ------------------------------------------------

_clrst:         and  _status,0f000h
		NEXT
_c_st:          mov  ax,_status
		mov  _regC,al
		mov  al,_regC+1
		and  ax,00ff0h
		or   al,ah
		mov  _regC+1,al
		NEXT
_st_c:          mov  ax,word ptr _regC
		mov  byte ptr _status,al
		mov  al,byte ptr _status+1
		and  ax,00ff0h
		or   al,ah
		mov  byte ptr _status+1,al
		NEXT
_cstex:         mov  al,_regC
		xchg al,byte ptr _status
		mov  _regC,al
		mov  al,byte ptr _regC+1
		mov  ah,byte ptr _status+1
		mov  bx,ax
		and  ax,0f00fh
		or   al,ah
		mov  byte ptr _status+1,al
		mov  ax,bx
		and  ax,00ff0h
		or   al,ah
		mov  _regC+1,al
		NEXT

_clr_st:        ; ah = mask
		mov  bl,cl
		mov  cl,ah
		mov  ax,1
		sal  ax,cl
		mov  cl,bl
		not  ax
		and  _status,ax
		NEXTI

_set_st:
		mov  bl,cl
		mov  cl,ah
		mov  ax,1
		sal  ax,cl
		mov  cl,bl
		or   _status,ax
		NEXTI

_tst_st0:       incr_pc
		mov  bl,cl
		mov  cl,ah
		mov  ax,1
		sal  ax,cl
		mov  cl,bl
		and  ax,_status
		je   _tst_stx
		clear_carry
		inc  si
		NEXT

_tst_st1:       incr_pc
		mov  bl,cl
		mov  cl,ah
		mov  ax,1
		sal  ax,cl
		mov  cl,bl
		and  ax,_status
		jne  _tst_stx
		clear_carry
		inc  si
		NEXT

; ------------------------------------------------
; hardware status
; ------------------------------------------------

_clrhst:        mov  al,ah
		not  al
		and  _hwstatus,al
		NEXTI

_tsthst:
                incr_pc
                mov  al,ah
		and  al,_hwstatus
		je   _tst_stx
		clear_carry
		inc  si
		NEXT
_tst_stx:       set_carry
		mov  ax,es:[si]
		sar  ax,cl
		or   al,al
		je   _tst_stx1
		cbw
		jmp  dojump
_tst_stx1:	jmp  _rtn

; ------------------------------------------------
; pointer P
; ------------------------------------------------


_tst_p0:        ; test si P # n ?
		incr_pc
		cmp  ah,_Pointer
		jne  _tst_stx
		clear_carry
		inc  si
		NEXT

_tst_p1:        ; test si P = n ?
		incr_pc
		cmp  ah,_Pointer
		je   _tst_stx
		clear_carry
		inc  si
		NEXT

_incr_p:        mov  al,_Pointer
		add  al,1
		update_halfcarry
		and  al,15
		mov  _Pointer,al
                je   _p_00
                set_pointer
		NEXT

_decr_p:        mov  al,_Pointer
		sub  al,1
		update_halfcarry
		and  al,15
		mov  _Pointer,al
                je   _p_00
                set_pointer
		NEXT


_p_0:           mov  _Pointer,0
_p_00:          clr_pointer
                NEXT

_p_x:           shr  al,4
		mov  _Pointer,al
                set_pointer
		NEXT



_add_C_P_1:     ; C(A)=C+P+1
		mov  al,_Pointer
		inc  al
		xor  ah,ah
		add  word ptr _regC,ax
		adc  _regC-1,0
		update_halfcarry
		NEXTI

_mov_C_P_n:     ; C=P  n
		mov  di,offset _regC
                fetch_byte
                shr  al,5
		jc   _mov_c_p_i
		; paire, quartet infrieur
                cmp  al,2
                je   _mov_c_p_4
                cbw
		add  di,ax
		mov  al,[di]
		and  al,0f0h
		or   al,_Pointer
                mov  [di],al
		NEXT
_mov_c_p_i:     ; impaire, quartet suprieur
                cbw
		add  di,ax
		mov  al,[di]
		and  al,00fh
		mov  ah,_Pointer
		sal  ah,4
		or   al,ah
                mov  [di],al
		NEXT
_mov_c_p_4:     mov  al, _Pointer
                mov  [di-1],al
                NEXT


_mov_P_C_n:     ; P=C  n
                fetch_byte
                shr  al,4
                mov  bl,al
		mov  di,offset _regC
		call maj_w
		mov  al,bl
		and  ax,15
		shr  al,1
		jc   _mov_p_c_i
		; paire, quartet infrieur
		add  di,ax
		mov  al,[di]
		jmp  short _mov_p_c_x
_mov_p_c_i:     ; impaire, quartet suprieur
		add  di,ax
		mov  al,[di]
		sar  al,4
_mov_p_c_x:     and  al,00fh
		mov  _Pointer,al
                set_pointer
		NEXT

_xch_C_P_n:     fetch_byte
                shr  al,4
                mov  bl,al
		mov  di,offset _regC
                push di
		call maj_w
		mov  al,bl
		and  ax,15
		shr  al,1
		jc   _xch_c_p_i
		; paire, quartet infrieur
		add  di,ax
		mov  al,[di]
		mov  ah,al
		and  ax,00ff0h
		or   al,_Pointer
		jmp  short _xch_c_p_x
_xch_c_p_i:     ; impaire, quartet suprieur
		add  di,ax
		mov  al,[di]
		ror  al,4
		mov  ah,al
		and  ax,00ff0h
		or   al,_Pointer
		rol  al,4
_xch_c_p_x:     mov  _Pointer,ah
		mov  [di],al
                pop  di
		maj_a
                set_pointer
		NEXT

; ------------------------------------------------
; registres input/output
; ------------------------------------------------

_mov_OUT_C:     mov  ax,word ptr _regC
		and  ax,00fffh
		mov  _output,ax
		NEXTI
_mov_OUT_CS:    mov  al,_regC
		mov  ah,byte ptr _output
		and  ax,0f00fh
		or   al,ah
		mov  byte ptr _output,al
		NEXTI
_mov_A_IN:      mov  ax,_input
		mov  word ptr _regA,ax
		NEXTI
_mov_C_IN:      mov  ax,_input
		mov  word ptr _regC,ax
		NEXTI

; ------------------------------------------------
; instructions systme
; ------------------------------------------------

chktimer:
                mov  bx,es
                mov  ax,word ptr [_tab_seg+4]
                mov  es,ax

                mov  di,070feh
                mov  ax,es:[di]
                and  al,ah
                and  al,080h
                jne  chtim1

                add  di,080h
                mov  ax,es:[di]
                and  al,ah
                and  al,080h
                jne  chtim1

                add  di,080h
                mov  ax,es:[di]
                and  al,ah
                and  al,080h
;                jne  chtim1

chtim1:         mov  es,bx
                ret



_uncnfg:        mov  ax,1
		jmp  retouri
_config:        mov  ax,2
		jmp  retouri
_mov_C_ID:      mov  ax,3
		jmp  retouri
_shutdn:        mov  ax,4
		jmp  retouri
_reset:         mov  ax,5
		jmp  retouri
_buscc:         mov  ax,6
		jmp  retouri

_buscb:
_buscd:         mov  ax,6
		jmp  retour

_sreq:
;                mov  ax,7
;		jmp  retouri
		and  _regC,0f0h
                call chktimer
                or   al,al
		je   sreq1
		or   _regC,1
		or   _hwstatus,4      ; SR
;		mov  _timer_to,0
sreq1:		NEXTI

; (code 8 = RTNSXM)

; added 25/03/06:
_rsi:           ; RSI is 80810
                fetch_nib
_inton:
_intoff:
                NEXT




; ------------------------------------------------
; bit operations on A and C
; ------------------------------------------------

_clrb_A_n:      mov  di, offset _regA
                jmp  short _clrbit_n
_clrb_C_n:      mov  di, offset _regC
_clrbit_n:      fetch_nib
                mov  dx,1
                xchg al,cl
                shl  dx,cl
                mov  cl,al
                not  dx
                and  [di],dx
                NEXT


_setb_A_n:      mov  di, offset _regA
                jmp  short _setbit_n
_setb_C_n:      mov  di, offset _regC
_setbit_n:      fetch_nib
                mov  dx,1
                xchg al,cl
                shl  dx,cl
                mov  cl,al
                or   [di],dx
                NEXT


_isset_A_n:     mov  di, offset _regA
                jmp  short _issetbit_n
_isset_C_n:     mov  di, offset _regC
_issetbit_n:    fetch_nib
                mov  dx,1
                xchg al,cl
                shl  dx,cl
                mov  cl,al
                and  dx,[di]
                je   _isset2
		jmp  _tst_stx
_isset2:        clear_carry
                inc  si
                NEXT


_isclr_A_n:     mov  di, offset _regA
                jmp  short _isclrbit_n
_isclr_C_n:     mov  di, offset _regC
_isclrbit_n:    fetch_nib
                mov  dx,1
                xchg al,cl
                shl  dx,cl
                mov  cl,al
                and  dx,[di]
                jne  _isset2
		jmp  _tst_stx




; ------------------------------------------------
; registres D0, D1
; ------------------------------------------------

_ld_D0_2:       fetch_byte
		mov  _regD0,al
		NEXT
_ld_D0_4:       fetch_byte
		mov  _regD0,al
		fetch_byte
		mov  _regD0+1,al
		NEXT
_ld_D0_5:       fetch_byte
		mov  _regD0,al
		fetch_byte
                and  ah,0fh
		mov  word ptr _regD0+1,ax
		NEXTI
_ld_D1_2:       fetch_byte
		mov  _regD1,al
		NEXT
_ld_D1_4:       fetch_byte
		mov  _regD1,al
		fetch_byte
		mov  _regD1+1,al
		NEXT
_ld_D1_5:       fetch_byte
		mov  _regD1,al
		fetch_byte
                and  ah,0fh
		mov  word ptr _regD1+1,ax
		NEXTI

; inc Dx, dec Dx : ax = [ - n x x ]
_inc_D0:	mov  al,ah
		inc  al
                cbw
                add  word ptr _regD0,ax
                jc   _inc_D0x
		clear_carry
		NEXTI
_inc_D0x:	adc  _regD0+2,0
		update_halfcarry
		NEXTI

_inc_D1:	mov  al,ah
		inc  al
                cbw
                add  word ptr _regD1,ax
                jc   _inc_D1x
                clear_carry
                NEXTI
_inc_D1x:       adc  _regD1+2,0
		update_halfcarry
		NEXTI

_dec_D0:	mov  al,ah
		inc  al
                cbw
                sub  word ptr _regD0,ax
                jc   _dec_D0x
                clear_carry
                NEXTI
_dec_D0x:       sbb  _regD0+2,0
		update_halfcarry
		NEXTI
_dec_D1:	mov  al,ah
		inc  al
                cbw
                sub  word ptr _regD1,ax
                jc   _dec_D1x
                clear_carry
                NEXTI
_dec_D1x:       sbb  _regD1+2,0
		update_halfcarry
		NEXTI


mov_Dx_as       macro POINTER, REGISTER
                align 2
_mov_&POINTER&_&REGISTER&:
                mov   al,_reg&REGISTER&-1
                and   al,0fh
		mov   _reg&POINTER&+2,al
_mov_&POINTER&_&REGISTER&_s:
                mov   ax,word ptr _reg&REGISTER&
		mov   word ptr _reg&POINTER&,ax
		NEXTI
                endm

                mov_Dx_as D0 A
                mov_Dx_as D1 A
                mov_Dx_as D0 C
                mov_Dx_as D1 C





xch_Dx_as       macro POINTER, REGISTER
                align 2
_xch_&REGISTER&_&POINTER&:
                mov   al,_reg&REGISTER&-1
                and   al,0fh
		xchg  al,_reg&POINTER&+2
                mov   _reg&REGISTER-1,al
_xch_&REGISTER&_&POINTER&_s:
                mov   ax,word ptr _reg&REGISTER&
		xchg  ax,word ptr _reg&POINTER&
                mov   word ptr _reg&REGISTER&,ax
		NEXTI
                endm

                xch_Dx_as D0 A
                xch_Dx_as D1 A
                xch_Dx_as D0 C
                xch_Dx_as D1 C



_con_Reg_n:
                ; al[3-0]=field, al[5-4]=reg, al[7-6]=operation, ah[3-0]=n
                mov  dl,al
                call _con0
                ; di=destination, bh=length, bl=start, ah=n
                or   bx,bx
                je   _con16
                mov  al,dl
                shr  al,6    ; rol al,2 ??
                and  al,3
                je   _add_Reg_n
_con15:         sub  al,2
                je   _sub_Reg_n
_con16:         jmp  invalid

                ; di=destination, bh=length, bl=start, ah=n
_add_Reg_n:     mov  dl,ah
                call maj_w
                push di
                xor  ax,ax
                mov  al,bl
                shr  al,1
                add  di,ax
                dec  bh
                shr  bl,1
                jc   _add1
                or   bh,bh
                stc
                jmp  short _add4
                ;
; add to high nibble
_add1:          shl  dl,4
                add  dl,00fh
                stc
                adc  [di],dl
                mov  dl,0
                inc  di
                dec  bh
                jmp  short _add4
; add to byte(s)
_add2:          adc  [di],dl
                mov  dl,0
                inc  di
                dec  bh
                dec  bh
_add4:          js   _add5
                jne  _add2
; add to low nibble
                mov  al,[di]
                mov  bh,al
                adc  al,dl
_add4a:         update_halfcarry
                mov  ah,bh
                and  ax,0f00fh
                or   al,ah
                mov  [di],al
                jmp  short _add6
_add5:          update_carry
_add6:          pop  di
                maj_a
                NEXT

                ; di=destination, bh=length, bl=start, ah=n
_sub_Reg_n:     mov  dl,ah
                call maj_w
                push di
                xor  ax,ax
                mov  al,bl
                shr  al,1
                add  di,ax
                dec  bh
                shr  bl,1
                jc   _sub1
                or   bh,bh
                stc
                jmp  short _sub4
                ;
; sub from high nibble
_sub1:          shl  dl,4
                add  dl,00fh
                stc
                sbb  [di],dl
                mov  dl,0
                inc  di
                dec  bh
                jmp  short _sub4
; sub from byte(s)
_sub2:          sbb  [di],dl
                mov  dl,0
                inc  di
                dec  bh
                dec  bh
_sub4:          js   _add5      ; _sub5
                jne  _sub2
; sub from low nibble
                mov  al,[di]
                mov  bh,al
                sbb  al,dl
                jmp  short _add4a

comment !
_sub4a:         update_halfcarry
                mov  ah,bh
                and  ax,0f00fh
                or   al,ah
                mov  [di],al
                jmp  short _sub6
_sub5:          update_carry
_sub6:          pop  di
                maj_a
                NEXT
 !


_srb_Reg_n:
                ; al[3-0]=field, al[5-4]=reg, al[7-6]=operation
                mov  dl,al
                call _con0
                ; di=destination, bh=length, bl=start
                or   bx,bx
                je   _srb16
                mov  al,dl
                shr  al,6      ; rol al,2 ??
                and  al,3
                je   _srb_Reg
_srb16:         jmp  invalid


                ; di=destination, bh=length, bl=start
_srb_Reg:
                call maj_w
                push di
                xor  ax,ax
                mov  al,bl
                add  al,bh
                dec  al
                mov  bl,al
                shr  al,1
                add  di,ax
                dec  bh
                shr  bl,1
                jnc  _srb1
                or   bh,bh
                jmp  short _srb4

; srb low nibble
_srb1:          mov  al,[di]
                mov  ah,al
                and  ax,0f00fh
                mov  dl,ah
                shr  al,1
                lahf
                or   al,dl
                mov  [di],al
                sahf
                dec  di
                dec  bh
                jmp  short _srb4
; srb byte(s)
_srb2:          rcr  byte ptr [di],1
                dec  di
                dec  bh
                dec  bh
_srb4:          js   _srb5
                jne  _srb2
; srb high nibble
                mov  al,[di]
                mov  dl,al
                rcr  al,1
                ; update SB
                mov  ah,al
                and  ah,8
                je   _srb4a
		or   _hwstatus,2       ; SB
_srb4a:         mov  ah,dl
                and  ax,00ff0h
                or   al,ah
                mov  [di],al
                jmp  short _srb6
_srb5:          ; update SB
                jnc  _srb6
		or   _hwstatus,2       ; SB
_srb6:          pop  di
                maj_a
                NEXT






_con0:
                ; inputs:
                ; al[3-0]=field, al[5-4]=reg, al[7-6]=operation, ah[3-0]=n
                ; output:
                ; di: point to destination reg
                ; bh: field length, bl: start of field
                shr  al,4
                and  al,3
                jne  _con1
                mov  di,offset _regA
                jmp  short _con4
_con1:          dec  al
                jne  _con2
                mov  di,offset _regB
                jmp  short _con4
_con2:          dec  al
                jne  _con3
                mov  di,offset _regC
                jmp  short _con4
_con3:          mov  di,offset _regD
_con4:          mov  al,dl
                and  al,0fh
                cmp  al,0fh
                je   _con13
                or   al,al
                jne  _con5
                ; P field
                mov  bl,_Pointer
                mov  bh,1
                jmp  short _con14
_con5:          dec  al
                jne  _con6
                ; WP
                mov  bx,0100h
                add  bh,_Pointer
                jmp  short _con14
_con6:          dec  al
                jne  _con7
                ; XS
                mov  bx,0102h
                jmp  short _con14
_con7:          dec  al
                jne  _con8
                ; X
                mov  bx,0300h
                jmp  short _con14
_con8:          dec  al
                jne  _con9
                ; S
                mov  bx,010fh
                jmp  short _con14
_con9:          dec  al
                jne  _con10
                ; M
                mov  bx,0c03h
                jmp  short _con14
_con10:         dec  al
                jne  _con11
                ; B
                mov  bx,0200h
                jmp  short _con14
_con11:         dec  al
                jne  _con12
                ; W
                mov  bx,1000h
                jmp  short _con14
_con12:         sub  al,8
                je   _con13
                xor  bx,bx
                jmp  short _con14
_con13:         ; A
                mov  bx,0500h
_con14:         ret



; ********************************************************************
;               mulation du processeur Saturn par le 8086
;                       J-F Garnier 1993, 95
; --------------------------------------------------------------------
; Module sat_mem.asm : mulation des instructions d'accs mmoire
; --------------------------------------------------------------------
; Version 1.0 06/09/93
;   prototype
; Version 1.1 04/11/95
; ********************************************************************



; routines accs mmoire :
; A=DAT0 fs
; C=DAT0 fs
; A=DAT1 fs
; C=DAT1 fs
; DAT0=A fs
; DAT0=C fs
; DAT1=A fs
; DAT1=C fs



exec_mem        macro INSTRUCT, REGIS, POINTER, CHAMP
		mov   bx, offset _reg&POINTER&
		mov   di, offset _reg&REGIS&
		jmp   exec_&INSTRUCT&_&CHAMP&
		endm

exec_mems       macro INSTRUCT, REGIS, POINTER, CHAMP
		mov   bx, offset _reg&POINTER&
		mov   di, offset _reg&REGIS&
		jmp   short exec_&INSTRUCT&_&CHAMP&
		endm







build_load      macro INSTRUCT
		local mem_p_i, mem_p_x, tab_mem_wp

&INSTRUCT&_p:   push  di
		call  maj_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    mem_p_i
		; paire, quartet infrieur
		add   di,ax
		call  &INSTRUCT&_lonib
		jmp   short mem_p_x
mem_p_i:        ; impaire, quartet suprieur
		add   di,ax
		call  &INSTRUCT&_hinib
mem_p_x:        pop   di
		maj_a
		pop   cx
		pop   si
		pop   es
		NEXT

&INSTRUCT&_wp:  push  di
		call  maj_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		call  cs:[bx+tab_mem_wp]
                pop   di
		maj_a
                pop   cx
		pop   si
		pop   es
		NEXT
tab_mem_wp      dw    &INSTRUCT&_1,  &INSTRUCT&_2,  &INSTRUCT&_3,  &INSTRUCT&_4
		dw    &INSTRUCT&_5,  &INSTRUCT&_6,  &INSTRUCT&_7,  &INSTRUCT&_8
		dw    &INSTRUCT&_9,  &INSTRUCT&_10, &INSTRUCT&_11, &INSTRUCT&_12
		dw    &INSTRUCT&_13, &INSTRUCT&_14, &INSTRUCT&_15, &INSTRUCT&_16



&INSTRUCT&_n:   push  di
		call  maj_w
		add   bx,bx
		call  cs:[bx+tab_mem_wp]
		pop   di
		maj_a
		pop   cx
		pop   si
		pop   es
		NEXT

&INSTRUCT&_xs:
		inc   di
		call  &INSTRUCT&_lonib
		pop   cx
		pop   si
		pop   es
		NEXT

&INSTRUCT&_x:
		call  &INSTRUCT&_3
		pop   cx
		pop   si
		pop   es
		NEXT

&INSTRUCT&_s:
		add   di,7
		call  &INSTRUCT&_hinib
                pop   cx
		pop   si
		pop   es
		NEXT

&INSTRUCT&_m:   push  di
		call  maj_w
		inc   di
		call  &INSTRUCT&_hinib
		inc   di
		call  &INSTRUCT&_11
		pop   di
		maj_a
		pop   cx
		pop   si
		pop   es
		NEXT

		endm







; code d'excution d'instructions accs mmoire

; cr un pointeur mmoire  partir du registre D0 ou D1 point par bx
get_pointer:
		mov   si,[bx]
		mov   bl,[bx+2]
		and   bx,0fh
; cr un pointeur mmoire  partir d'une adresse bx-si
bld_pointer:
		xor   cl,cl
		shr   si,1
		rcl   cl,3
		add   bx,bx
comment !
                cmp   bx,10
                jne   bp1
                cmp   si,02000h
                jc    bp1
    		mov   bx,[_tab_seg+bx]
		mov   es,bx
                stc
                ret
bp1:
!
		cmp   bx,4      ; pour test criture ROMs internes !!!
    		mov   bx,[_tab_seg+bx]
		mov   es,bx
		ret





		irp   CHAMP,<p, wp, xs, x, m, w>
		local load00
exec_load_&CHAMP&:
		push  es
		push  si
		push  cx
		call  get_pointer
		; es:si, cl : pointeur mmoire source
		; ds:di : pointeur sur registre destination
		or    bx,bx
		je    load00
		jmp   load_&CHAMP&
load00:         jmp   load00_&CHAMP&
		endm


                align 2
_lod_A_D0_a:    exec_mems load A D0 a
_lod_C_D0_a:    exec_mems load C D0 a
_lod_A_D1_a:    exec_mems load A D1 a
_lod_C_D1_a:    exec_mems load C D1 a

exec_load_a:
		mov   ax,[bx]
		mov   bl,[bx+2]
		and   bx,0fh
; cr un pointeur mmoire  partir d'une adresse bl-ax
		add   bx,bx
		mov   bx,[_tab_seg+bx]
                xchg  ax,bx
                or    ax,ax
                je    exec_load_a0
                mov   dx,es
                mov   es,ax
		shr   bx,1
                mov   ax,es:[bx]
                mov   bx,es:[bx+1]
                jnc   exec_load_a1
                shr   ax,4
                shr   bx,4
                mov   ah,bl
exec_load_a1:   mov   [di],ax
                mov   [di-1],bh
                mov   es,dx
                NEXTI
exec_load_a0:   xor   ax,ax
                mov   [di],ax
                mov   [di-1],al
                NEXTI



                align 2
_lod_A_D0_b:    exec_mems load A D0 b
_lod_C_D0_b:    exec_mems load C D0 b
_lod_A_D1_b:    exec_mems load A D1 b
_lod_C_D1_b:    exec_mems load C D1 b

exec_load_b:
		mov   ax,[bx]
		mov   bl,[bx+2]
		and   bx,0fh
; cr un pointeur mmoire  partir d'une adresse bl-ax
		add   bx,bx
		mov   bx,[_tab_seg+bx]
                xchg  ax,bx
                or    ax,ax
                je    exec_load_b0
                mov   dx,es
                mov   es,ax
		shr   bx,1
                mov   ax,es:[bx]
                jnc   exec_load_b1
                shr   ax,4
exec_load_b1:   mov   [di],al
                mov   es,dx
                NEXTI
exec_load_b0:   xor   ax,ax
                mov   [di],al
                NEXTI


                align 2
_lod_A_D0_s:    exec_mems load A D0 s
_lod_C_D0_s:    exec_mems load C D0 s
_lod_A_D1_s:    exec_mems load A D1 s
_lod_C_D1_s:    exec_mems load C D1 s

exec_load_s:
		mov   ax,[bx]
		mov   bl,[bx+2]
		and   bx,0fh
; cr un pointeur mmoire  partir d'une adresse bl-ax
		add   bx,bx
		mov   bx,[_tab_seg+bx]
                xchg  ax,bx
                or    ax,ax
                je    exec_load_s0
                mov   dx,es
                mov   es,ax
		shr   bx,1
                mov   al,es:[bx]
                jc    exec_load_s1
            	shl   al,4
exec_load_s1:   mov   ah,[di+7]
		and   ax,0ff0h
		or    al,ah
                mov   [di+7],al
                mov   es,dx
                NEXT
exec_load_s0:   and   byte ptr [di+7],0fh
                NEXT





; load_n et store_n :
exec_load_n:
                fetch_byte
                shr   al,4
                and   ax,15
; al contient d-1, d tant le nombre d'octets  transfrer
                push  es
		push  si
		push  cx
		push  ax
		call  get_pointer
		or    bx,bx
		je    exec_load0
		pop   bx
		; es:si, cl : pointeur mmoire,
		; ds:di : pointeur sur registre (source ou destination)
		; bx: nbre d'octets-1
		jmp   load_n
exec_load0:     pop   bx
		jmp   load00_n




; support for HPIL mailbox

; DAT0=C 8 in HPIL mailbox
_stoil_C_D0_8:
		mov   bx, offset _regD0
		mov   di, offset _regC
                push  es
		push  si
		push  cx
		call  get_pointer
		jc    _stoil0     ; pour test criture ROMs !!!
		or    bx,bx
		je    _stoil0
		; es:si, cl : pointeur mmoire,
		; ds:di : pointeur sur registre (source ou destination)
                call  maj_w
                or    cl,cl
		je    _stoil1
		; cl#0, il faut ecrire @di[3-0] dans hinib
		call  store_hinib
                call  store_7
                jmp   short _stoil2
_stoil1:        call  store_8
_stoil2:        pop   cx
		pop   si
		pop   es
		mov   ax,10       ; WRIL
                jmp   retour
_stoil0:        pop   cx
		pop   si
		pop   es
		NEXT



; C=DAT0 7 from HPIL mailbox
_lodil_C_D0_7:
		mov   bx, offset _regD0
		mov   di, offset _regC
                push  es
		push  si
		push  cx
		call  get_pointer
		or    bx,bx
		je    _lodil0
		; es:si, cl : pointeur mmoire,
		; ds:di : pointeur sur registre (source ou destination)
                push  di
		call  maj_w
                call  load_7
		pop   di
		maj_a
		pop   cx
		pop   si
		pop   es
		mov   ax,11     ; RDIL
                jmp   retour
_lodil0:        pop   cx
		pop   si
		pop   es
                mov   bx,6
		jmp   load00_n   ; was _clr_n







; -------------------------------------------------------------------
; chargement d'un registre depuis la mmoire
; -------------------------------------------------------------------



load_byte       macro
		mov   ax,es:[si]
		shr   ax,cl
		mov   [di],al
		inc   si
		inc   di
		endm

; charge un quartet es:si, cl dans hinib di[7-4]
; si cl=0, es:si[3-0], sinon es:si[7-4]
; incremente es:si,cl
load_hinib:     mov   al,es:[si]
		xor   cl,4
		jne   load_hinib1
		inc   si
load_hinib1:	shl   al,cl
		mov   ah,[di]
		and   ax,0ff0h
		or    al,ah
		mov   [di],al
		ret



load_16:        load_byte
load_14:        load_byte
load_12:        load_byte
load_10:        load_byte
load_8:         load_byte
load_6:         load_byte
load_4:         load_byte
load_2:         load_byte
		ret

load_15:        load_byte
load_13:        load_byte
load_11:        load_byte
load_9:         load_byte
load_7:         load_byte
load_5:         load_byte
load_3:         load_byte
load_1:
; charge un quartet es:si, cl dans lonib di[3-0]
; si cl=0, es:si[3-0], sinon es:si[7-4]
load_lonib:	mov   al,es:[si]
		sar   al,cl
		mov   ah,[di]
		and   ax,0f00fh
		or    al,ah
		mov   [di],al
		ret

load_a:         mov   ax,es:[si]
		sar   ax,cl
		mov   bx,es:[si+1]
		sar   bx,cl
		mov   ah,bl
		mov   [di],ax
		mov   [di-1],bh
		pop   cx
		pop   si
		pop   es
		NEXTI

load_b:
		mov   ax,es:[si]
		sar   ax,cl
		mov   [di],al
                pop   cx
		pop   si
		pop   es
		NEXTI


load_w:         mov   ax,es:[si]
		sar   ax,cl
		mov   bx,es:[si+1]
		sar   bx,cl
		mov   ah,bl
		mov   [di],ax

		mov   ax,es:[si+2]
		sar   ax,cl
		mov   bx,es:[si+3]
		sar   bx,cl
		mov   ah,bl
		mov   [di+2],ax

		mov   ax,es:[si+4]
		sar   ax,cl
		mov   bx,es:[si+5]
		sar   bx,cl
		mov   ah,bl
		mov   [di+4],ax

		mov   ax,es:[si+6]
		sar   ax,cl
		mov   bx,es:[si+7]
		sar   bx,cl
		mov   ah,bl
		mov   [di+6],ax

		pop   cx
		pop   si
		pop   es
		maj_a
		NEXT

comment !
; version optimis de load_w,  mettre au point.
new_load_w:     or    cl,cl
		jne   load_w1
		mov   ax,es:[si]
		mov   [di],ax
		mov   ax,es:[si+2]
		mov   [di+2],ax
		mov   ax,es:[si+4]
		mov   [di+4],ax
		mov   ax,es:[si+6]
		mov   [di+6],ax
		pop   cx
		pop   si
		pop   es
		maj_a
		NEXT
load_w1:
		mov   al,es:[si]        [ x x 0 x ]

		mov   bx,es:[si+1]     [ 4 3 2 1 ]
		mov   dx,es:[si+3]     [ 8 7 6 5 ]



; j'ai droit  8 instructions:
sol1:		rol   bx,cl            [ 3 2 1 4 ]
		rol   dx,cl            [ 7 6 5 8 ]
		mov   ah,bl            [ 1 4 0 -]
		and   ax,0f00fh
		or    al,ah            [ x x  1 0 ]
		xchg  al,bl            [ x x 1 4 ]
		mov   ah,dl            [ 5 8 1 4 ]
		and   ax,0f00fh
		or    al,ah            [ x x 5 4 ]
		xchg  al,dl            [ x x 5 8 ]

sol2:		mov   ah,bl            [ 2 1 0 x ]
		shr   ax,cl            [ - 2 1 0 ]
		rol   bx,cl            [ 3 2 1 4 ]
		xchg  al,bl        ax: [ - 2 1 4 ]    bx : [ 3 2 1 0 ]
		rol   dx,cl            [ 7 6 5 8 ]

sol3:           mov   ah,bh        ax: [ 4 3 0 x ]
                and   ax,00ff0h        [ - 3 0 - ]
                or    al,ah            [ - 3 0 3 ]
                xchg  al,bh        bx: [ 0 3 2 1 ]   al [ 4 3 ]
                mov   ah,dh        ax: [ 8 7 4 3 ]
                and   ax,00ff0h        [ - 7 4 - ]
                or    al,ah            [ 4 7]
                mov  dh,al             [ 4 7 6 5 ]


		mov   ax,es:[si]       [ 2 1 0 x ]
		mov   bx,es:[si+2]     [ 6 5 4 3 ]



		mov   [di],bx
		mov   [di+2],dx


		pop   cx
		pop   si
		pop   es
		maj_a
		NEXT
!

; dfinitions du reste du code
		build_load load




load00_p:       mov  bl,_Pointer
                mov  bh,bl
                jmp  short load_n0
load00_wp:      mov  bh,_Pointer
                mov  bl,0
                jmp  short load_n0
load00_xs:      mov  bx,0202h
                jmp  short load_n0
load00_x:       mov  bx,0200h
                jmp  short load_n0
load00_s:       mov  bx,0f0fh
                jmp  short load_n0
load00_m:       mov  bx,0e03h
                jmp  short load_n0
load00_w:       mov  bx,0f00h
                jmp  short load_n0
load00_n:       xchg bl,bh
load_n0:        push di
                call maj_w
                mov  ax,bx
                shr  bl,1
                xor  bh,bh
                add  di,bx
                sub  ah,al
                shr  al,1
                jnc  load_n1
                and  byte ptr [di],0fh
                inc  di
                dec  ah
load_n1:        xor  al,al
                or   ah,ah
load_n2:        js   load_n4
                je   load_n3
                mov  [di],al
                inc  di
                sub  ah,2
                jmp  load_n2
load_n3:        and  byte ptr [di],0f0h
load_n4:        pop   di
                maj_a
                pop   cx
		pop   si
		pop   es
                NEXT


load00_b:       xor  al,al
                mov  [di],al
		jmp  short load00_ab
load00_a:       xor  ax,ax
                mov  [di],ax
                mov  [di-1],al
load00_ab:	pop   cx
		pop   si
		pop   es
                NEXTI


; ======================================================================


; il faut inverser la signification de cl:
; c'est maintenant le dcallage  faire subir  la source (registre)


build_store     macro INSTRUCT
		local mem_p_i, mem_p_x, mem_p_1, mem_p_2
		local tab_mem_wp, mem_wp_0
		local mem_xs_1, mem_xs_2
		local mem_x_1, mem_x_2
		local mem_s_1, mem_s_2
		local mem_m_1, mem_m_2

&INSTRUCT&_p:   call  maj_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    mem_p_i
		; paire, quartet infrieur
		add   di,ax
		or    cl,cl
		jne   mem_p_1
		; cl=0, il faut ecrire @di[3-0] dans lonib
		call  &INSTRUCT&_lonib
		jmp   short mem_p_x
mem_p_1:        ; cl#0, il faut ecrire @di[3-0] dans hinib
		call  &INSTRUCT&_hinib
		jmp   short mem_p_x
mem_p_i:        ; impaire, quartet suprieur
		add   di,ax
		or    cl,cl
		jne   mem_p_2
		; cl=0, il faut ecrire @di[7-4] dans lonib
		mov   cl,4
		call  &INSTRUCT&_lonib
		jmp   short mem_p_x
mem_p_2:	; cl#0, il faut ecrire @di[7-4] dans hinib
		mov   cl,0
		call  &INSTRUCT&_hinib
mem_p_x:        pop   cx
		pop   si
		pop   es
		NEXT

&INSTRUCT&_n:   call  maj_w
		jmp   short mem_wp_0

&INSTRUCT&_wp:  call  maj_w
		xor   bx,bx
		mov   bl,_Pointer
mem_wp_0:       or    cl,cl
		je    mem_wp_1
		; cl#0, il faut ecrire @di[3-0] dans hinib
		call  &INSTRUCT&_hinib
		dec   bx
		js    mem_wp_2    ; il n'y avait qu'un quartet
		; pointe sur quartet destination suivant
		inc   si
		; cl reste #0 pour ecriture octet source [11-4]
mem_wp_1:       add   bx,bx
		call  cs:[bx+tab_mem_wp]
mem_wp_2:       pop   cx
		pop   si
		pop   es
		NEXT
tab_mem_wp      dw    &INSTRUCT&_1,  &INSTRUCT&_2,  &INSTRUCT&_3,  &INSTRUCT&_4
		dw    &INSTRUCT&_5,  &INSTRUCT&_6,  &INSTRUCT&_7,  &INSTRUCT&_8
		dw    &INSTRUCT&_9,  &INSTRUCT&_10, &INSTRUCT&_11, &INSTRUCT&_12
		dw    &INSTRUCT&_13, &INSTRUCT&_14, &INSTRUCT&_15, &INSTRUCT&_16

&INSTRUCT&_xs:  inc   di
		or    cl,cl
		jne   mem_xs_1
		; cl=0, il faut ecrire @di[3-0] dans lonib
		call  &INSTRUCT&_lonib
		jmp   short mem_xs_2
mem_xs_1:       ; cl#0, il faut ecrire @di[3-0] dans hinib
		call  &INSTRUCT&_hinib
mem_xs_2:       pop   cx
		pop   si
		pop   es
		NEXT

&INSTRUCT&_x:   or    cl,cl
		jne   mem_x_1
		call  &INSTRUCT&_3
		jmp   short mem_x_2
mem_x_1:        ; cl#0, il faut ecrire @di[3-0] dans hinib
		call  &INSTRUCT&_hinib
		; pointe sur quartet destination suivant
		inc   si
		; cl reste #0 pour ecriture octet source [11-4]
		call  &INSTRUCT&_2
mem_x_2:        pop   cx
		pop   si
		pop   es
		NEXT

&INSTRUCT&_s:   add   di,7
		or    cl,cl
		jne   mem_s_1
		; cl=0, il faut ecrire @di[7-4] dans lonib
		mov   cl,4
		call  &INSTRUCT&_lonib
		jmp   short mem_s_2
mem_s_1:	; cl#0, il faut ecrire @di[7-4] dans hinib
		mov   cl,0
		call  &INSTRUCT&_hinib
mem_s_2:        pop   cx
		pop   si
		pop   es
		NEXT

&INSTRUCT&_m:   call  maj_w
		inc   di
		or    cl,cl
		jne   mem_m_1
		; cl=0, il faut ecrire @di[11-4] dans octets destination
		mov   cl,4
		call  &INSTRUCT&_12
		jmp   short mem_m_2
mem_m_1:        ; cl#0, il faut ecrire @di[7-4] dans hinib
		mov   cl,0
		call  &INSTRUCT&_hinib
		inc   si
		inc   di
		; puis ecrire 5 octets @di[7-0] dans octets destination
		; puis ecrire @di[3-0] dans lonib
		call  &INSTRUCT&_11
mem_m_2:        pop   cx
		pop   si
		pop   es
		NEXT


		endm





; -------------------------------------------------------------------
; criture d'un registre dans la mmoire
; -------------------------------------------------------------------



; ecrit un octet dans l'octet de la RAM pointee par es:si
; si cl=0, ecrit @di[7-0], sinon ecrit @di[11-4]
store_byte      macro
		mov   ax,[di]
		shr   ax,cl
		mov   es:[si],al
		inc   si
		inc   di
		endm

; ecrit un quartet dans le hinib de la RAM pointee par es:si
; si cl=0, ecrit @di[7-4], sinon ecrit @di[3-0]
store_hinib:	mov   al,[di]
		shl   al,cl
		mov   ah,es:[si]
		and   ax,0ff0h
		or    al,ah
		mov   es:[si],al
		ret

store_16:       store_byte
store_14:       store_byte
store_12:       store_byte
store_10:       store_byte
store_8:        store_byte
store_6:        store_byte
store_4:        store_byte
store_2:        store_byte
		ret

store_15:       store_byte
store_13:       store_byte
store_11:       store_byte
store_9:        store_byte
store_7:        store_byte
store_5:        store_byte
store_3:        store_byte
store_1:
; ecrit un quartet dans le lonib de la RAM pointee par es:si
; si cl=0, ecrit @di[3-0], sinon ecrit @di[7-4]
store_lonib:
		mov   al,[di]
		shr   al,cl
		mov   ah,es:[si]
		and   ax,0f00fh
		or    al,ah
		mov   es:[si],al
		ret

store_a:
; si cl=0, crire di en es:si, et di-1[3-0] en es:si+2[3-0]
; sinon, di[3-0] en es:si[7-4], di[7-0] en es:si+1[3-0], ...
		or    cl,cl
		jne   store_a1
		mov   ax,[di]
		mov   es:[si],ax    ; quartets 0, 1, 2, 3
		mov   al,[di-1]
		mov   ah,es:[si+2]
		and   ax,0f00fh
		or    al,ah
		mov   es:[si+2],al  ; quartet 4, -
		pop   cx
		pop   si
		pop   es
		NEXTI
store_a1:       mov   ax,[di]
		mov   bh,ah
		shl   ax,cl
		mov   bl,es:[si]
		and   bl,00fh
		or    al,bl
		mov   es:[si],ax    ; quartets -, 0, 1, 2
		mov   al,bh
		mov   ah,[di-1]
		shr   ax,cl
		mov   es:[si+2],al  ; quartets 3, 4
		pop   cx
		pop   si
		pop   es
		NEXTI



store_b:
; si cl=0, crire en es:si
; sinon, di[3-0] en es:si[7-4], et di[7-4] en es:si+1[3-0]
		mov   al,[di]
		or    cl,cl
		jne   store_b1
		mov   es:[si],al
		pop   cx
		pop   si
		pop   es
		NEXTI
store_b1:       xor   ah,ah
		shl   ax,4
		mov   bx,es:[si]
		and   bx,0f00fh
		or    ax,bx
		mov   es:[si],ax
		pop   cx
		pop   si
		pop   es
		NEXTI



store_w:
; mise  jour registre
		call  maj_w
; si cl=0, crire di en es:si, et di-1[3-0] en es:si+2[3-0]
; sinon, di[3-0] en es:si[7-4], di[7-0] en es:si+1[3-0], ...
		or    cl,cl
		jne   store_w1
		mov   ax,[di]
		mov   es:[si],ax
		mov   ax,[di+2]
		mov   es:[si+2],ax
		mov   ax,[di+4]
		mov   es:[si+4],ax
		mov   ax,[di+6]
		mov   es:[si+6],ax
		pop   cx
		pop   si
		pop   es
		NEXT
store_w1:       mov   ax,[di]        ; ax = [ 3 2 1 0 ]
		mov   bh,ah          ; bh = [ 3 2 ]
		shl   ax,4           ; ax = [ 2 1 0 - ]
		mov   bl,es:[si]     ; bl = [ x x ]
		and   bl,00fh        ; bl = [ - x ]
		or    al,bl          ; ax = [ 2 1 0 x ]
		mov   es:[si],ax     ; quartets 0, 1, 2
		; ici, bh[7-4] contient le prochain quartet (3)

		mov   al,bh          ; al = [ 3 2 ]
		mov   bx,[di+2]      ; bx = [ 7 6 5 4 ]
		mov   ah,bl          ; ah = [ 5 4 ]
		shr   ax,4           ; ax = [ - 5 4 3 ]
		mov   dx,bx          ; dx = [ 7 6 5 4 ]
		shl   dx,4           ; dx = [ 6 5 4 - ]
		mov   ah,dh          ; ax = [ 6 5 4 3 ]
		mov   es:[si+2],ax   ; quartets 3, 4, 5, 6

		mov   al,bh
		mov   bx,[di+4]
		mov   ah,bl
		shr   ax,4
		mov   dx,bx
		shl   dx,4
		mov   ah,dh
		mov   es:[si+4],ax   ; quartets 7, 8, 9, 10

		mov   al,bh
		mov   bx,[di+6]
		mov   ah,bl
		shr   ax,4
		mov   dx,bx
		shl   dx,4
		mov   ah,dh
		mov   es:[si+6],ax   ; quartets 11, 12, 13, 14

		shr   bh,4           ; bh = [ - 15 ]
		mov   al,es:[si+8]   ; al = [ x x]
		and   al,0f0h        ; al = [ x - ]
		or    al,bh          ; al = [ x 15]
		mov   es:[si+8],al   ; quartet 15

		pop   cx
		pop   si
		pop   es
		NEXT



		irp   CHAMP,<p, wp, xs, x, m, w>
		local store00
exec_store_&CHAMP&:
		push  es
		push  si
		push  cx
		call  get_pointer
		jc    store00      ; pour test criture ROMs !!!
		; es:si, cl : pointeur mmoire destination
		; ds:di : pointeur sur registre source
		or    bx,bx
		je    store00
		jmp   store_&CHAMP&
store00:        jmp   exec_store00
		endm



                align 2
_sto_A_D0_a:    exec_mems store A D0 a
_sto_C_D0_a:    exec_mems store C D0 a
_sto_A_D1_a:    exec_mems store A D1 a
_sto_C_D1_a:    exec_mems store C D1 a

exec_store_a:
		mov   ax,[bx]
		mov   bl,[bx+2]
		and   bx,0fh
; cr un pointeur mmoire  partir d'une adresse bl-ax
		add   bx,bx
comment !
                cmp   bx,10
                jne   ea1
                cmp   ax,04000h
                jnc   exec_store_a0
ea1:
!
                cmp   bx,4
                jc    exec_store_a0
		mov   bx,[_tab_seg+bx]
                xchg  ax,bx
                or    ax,ax
                je    exec_store_a0
                mov   dx,es
                mov   es,ax
		shr   bx,1
                xchg  bx,di
                jnc   exec_store_a1
; bx[3-0] en es:di[7-4], bx[7-4] en es:di+1[3-0], ...
                mov   ax,[bx]       ; 3 2 1 0
                mov   bh,[bx-1]     ; - 4
                mov   bl,ah         ; - 4 3 2
		shl   ax,4          ;2 1 0 -
                shl   bx,4          ; 4 3 2 -
		mov   bl,es:[di]
		and   bl,00fh
		or    al,bl
		mov   es:[di],ax    ; quartets -, 0, 1, 2
		mov   es:[di+2],bh  ; quartets 3, 4
                mov   es,dx
                NEXTI
exec_store_a1:
; crire bx en es:di, et bx-1[3-0] en es:di+2[3-0]
                mov   ax,[bx]
                mov   es:[di],ax
                mov   al,[bx-1]
                mov   ah,es:[di+2]
		and   ax,0f00fh
		or    al,ah
                mov   es:[di+2],al
                mov   es,dx
exec_store_a0:  NEXTI


                align 2
_sto_A_D0_b:    exec_mems store A D0 b
_sto_C_D0_b:    exec_mems store C D0 b
_sto_A_D1_b:    exec_mems store A D1 b
_sto_C_D1_b:    exec_mems store C D1 b

exec_store_b:
		mov   ax,[bx]
		mov   bl,[bx+2]
		and   bx,0fh
; cr un pointeur mmoire  partir d'une adresse bl-ax
		add   bx,bx
comment !
                cmp   bx,10
                jne   ea2
                cmp   ax,04000h
                jnc   exec_store_b0
ea2:
!
                cmp   bx,4
		mov   bx,[_tab_seg+bx]
                jc    exec_store_b0
                xchg  ax,bx
                or    ax,ax
                je    exec_store_b0
                mov   dx,es
                mov   es,ax
		shr   bx,1
                xchg  bx,di
                jnc   exec_store_b1
; ecrire bx[3-0] en es:di[7-4], et bx[7-4] en es:di+1[3-0]
                mov   al,[bx]       ; 1 0
                xor   ah,ah
		shl   ax,4          ; - 1 0 -
		mov   bx,es:[di]
		and   bx,0f00fh
		or    ax,bx
		mov   es:[di],ax
                mov   es,dx
                NEXTI
exec_store_b1:
; crire bx en es:di
                mov   al,[bx]
                mov   es:[di],al
                mov   es,dx
exec_store_b0:  NEXTI


                align 2
_sto_A_D0_s:    exec_mems store A D0 s
_sto_C_D0_s:    exec_mems store C D0 s
_sto_A_D1_s:    exec_mems store A D1 s
_sto_C_D1_s:    exec_mems store C D1 s

exec_store_s:
		mov   ax,[bx]
		mov   bl,[bx+2]
		and   bx,0fh
; cr un pointeur mmoire  partir d'une adresse bl-ax
		add   bx,bx
comment !
                cmp   bx,10
                jne   ea3
                cmp   ax,04000h
                jnc   exec_store_s0
ea3:
!
                cmp   bx,4
		mov   bx,[_tab_seg+bx]
                jc    exec_store_s0
                xchg  ax,bx
                or    ax,ax
                je    exec_store_s0
                mov   dx,es
                mov   es,ax
		shr   bx,1
                xchg  bx,di
                mov   al,[bx+7]
                jc    exec_store_s1
; ecrire al[7-4] en es:di[3-0]
                shr   al,4
		mov   ah,es:[di]
		and   ax,0f00fh
		or    al,ah
		mov   es:[di],al
                mov   es,dx
                NEXT
exec_store_s1:
; ecrire al [7-4] en es:di[7-4]
                mov   ah,es:[di]
                and   ax,0ff0h
                or    al,ah
                mov   es:[di],al
                mov   es,dx
exec_store_s0:  NEXT




exec_store_n:
                fetch_byte
                shr   al,4
                and   ax,15
                push  es
		push  si
		push  cx
		push  ax
		call  get_pointer
		jc    exec_store0     ; pour test criture ROMs !!!
		or    bx,bx
		je    exec_store0
		pop   bx
		; es:si, cl : pointeur mmoire,
		; ds:di : pointeur sur registre (source ou destination)
		; bx: nbre d'octets-1
		jmp   store_n
exec_store0:    pop   ax
exec_store00:   pop   cx
		pop   si
		pop   es
		NEXT

exec_store00ab: pop   cx
		pop   si
		pop   es
		NEXTI

err_store0:     pop   ax
err_store00:    pop   cx
		pop   si
		pop   es
		mov   ax,-2
		jmp   retour







; dfinitions du reste du code
		build_store store

; -----------------------------------------







; dfinitions des points d'entres
		irp   CHAMP,<p, wp, xs, x, m, w, n>

_lod_A_D0_&CHAMP&:      exec_mem load A D0 CHAMP
_lod_C_D0_&CHAMP&:      exec_mem load C D0 CHAMP
_sto_A_D0_&CHAMP&:      exec_mem store A D0 CHAMP
_sto_C_D0_&CHAMP&:      exec_mem store C D0 CHAMP
_lod_A_D1_&CHAMP&:      exec_mem load A D1 CHAMP
_lod_C_D1_&CHAMP&:      exec_mem load C D1 CHAMP
_sto_A_D1_&CHAMP&:      exec_mem store A D1 CHAMP
_sto_C_D1_&CHAMP&:      exec_mem store C D1 CHAMP

		endm


		end

